package org.artifactory.hqc.service;

import com.google.common.annotations.VisibleForTesting;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.ha.HaCommonAddon;
import org.artifactory.common.ConstantValues;
import org.artifactory.hqc.cleanup.HeavyQueryCacheCleanupJob;
import org.artifactory.hqc.config.HeavyQueryCacheConfig;
import org.artifactory.hqc.enums.HeavyQueryCacheType;
import org.artifactory.hqc.exception.HeavyQueryCacheException;
import org.artifactory.schedule.Task;
import org.artifactory.schedule.TaskService;
import org.artifactory.schedule.TaskUtils;
import org.artifactory.spring.ContextCreationListener;
import org.artifactory.storage.StorageException;
import org.artifactory.storage.db.DbService;
import org.artifactory.storage.db.hqc.dao.HeavyQueryCacheDao;
import org.artifactory.storage.db.hqc.entity.HeavyQueryCacheEntry;
import org.jfrog.common.JsonUtils;
import org.jfrog.common.concurrency.ConflictGuard;
import org.jfrog.common.concurrency.ConflictsGuard;
import org.joda.time.DateTime;
import org.joda.time.Days;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import javax.annotation.Nullable;
import java.sql.SQLException;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.TimeUnit;

import static org.artifactory.common.ConstantValues.hqcCleanupJobCron;
import static org.artifactory.common.ConstantValues.hqcEnabled;
import static org.artifactory.common.ConstantValues.hqcRequestTimeoutSecs;
import static org.artifactory.cron.CronUtils.getNextExecution;
import static org.artifactory.cron.CronUtils.getNextExecutionAfter;
import static org.artifactory.cron.CronUtils.isValid;

/**
 * @author Tomer Slutsky
 * @author Andrey Tashlitsky
 */
@Slf4j
@Service
public class HeavyQueryCacheServiceImpl implements HeavyQueryCacheService, ContextCreationListener {

    @VisibleForTesting
    static final String CLEANUP_JOB_DEFAULT_CRON = "0 0 0 ? * * *";
    private static final int CLEANUP_JOB_MIN_DAYS = 1;
    private static final int CLEANUP_JOB_MAX_DAYS = 7;
    private static final int SIZE_KB = 1024;
    private final HeavyQueryCacheDao heavyQueryCacheDao;
    private final DbService dbService;
    private final AddonsManager addonsManager;
    private final TaskService taskService;

    @Autowired
    public HeavyQueryCacheServiceImpl(HeavyQueryCacheDao heavyQueryCacheDao, DbService dbService,
            AddonsManager addonsManager, TaskService taskService) {
        this.heavyQueryCacheDao = heavyQueryCacheDao;
        this.dbService = dbService;
        this.addonsManager = addonsManager;
        this.taskService = taskService;
    }

    @Override
    public <T> T get(@NonNull Callable<T> callable, @NonNull Class<T> clazz, @NonNull String name,
            @NonNull HeavyQueryCacheType hqcType, @NonNull HeavyQueryCacheConfig hqcConfig) {
        if (!hqcEnabled.getBoolean()) {
            try {
                log.debug("Heavy query cache is disabled");
                return callable.call();
            } catch (Exception e) {
                log.error("Callable failed", e);
                return null;
            }
        }
        ConflictGuard lock = null;
        String lockName = null;
        try {
            long startTime = System.currentTimeMillis();
            log.debug("Started processing the request at: {}", startTime);
            lockName = hqcType + (StringUtils.isNotEmpty(name) ? "_" + name : "");
            T result = readFromHeavyQueryCache(lockName, startTime, clazz);
            if (result != null) {
                log.debug("Found cached result, returning it");
                return result;
            }
            ConflictsGuard<String> conflictsGuard =
                    addonsManager.addonByType(HaCommonAddon.class).getConflictsGuard(lockName);
            lock = conflictsGuard.getLock(lockName);
            if (lock.tryToLock(hqcRequestTimeoutSecs.getInt(), TimeUnit.SECONDS)) {
                log.debug("Acquired hqc lock for: {}", lockName);
                result = readFromHeavyQueryCache(lockName, startTime, clazz);
                return result != null ? result : getCallableQueryResult(callable, lockName, hqcConfig);
            }
            throw new HeavyQueryCacheException("The query is in progress," +
                    " but a timeout has occurred for the current request");
        } catch (InterruptedException e) {
            log.error("Failed to acquire hqc lock for: {}", lock, e);
            Thread.currentThread().interrupt();
            return null;
        } finally {
            if (lock != null) {
                lock.unlock();
                log.debug("Released hqc lock for: {}", lockName);
            }
        }
    }

    @Override
    public int runCleanup() {
        try {
            long currentTimeMillis = System.currentTimeMillis();
            int count = 0;
            while (true) {
                List<HeavyQueryCacheEntry> expiredEntries = heavyQueryCacheDao.getExpiredEntries(currentTimeMillis);
                if (expiredEntries.isEmpty()) {
                    log.trace("Finished running cleanup");
                    break;
                }
                heavyQueryCacheDao.delete(expiredEntries);
                log.trace("Deleted all expired entries: {}", expiredEntries);
                count += expiredEntries.size();
            }
            log.info("Cleanup deleted {} expired entries from heavy query cache", count);
            return count;
        } catch (SQLException e) {
            throw new StorageException("Failed to run cleanup: " + e.getMessage(), e);
        }
    }

    @Override
    public void onContextCreated() {
        String cronExpression = getCronExpression();
        Task t = TaskUtils.createCronTask(HeavyQueryCacheCleanupJob.class, cronExpression);
        log.debug("Scheduling HeavyQueryCacheCleanupJob to run with Cron expression: {}", cronExpression);
        taskService.startTask(t, false);
        log.debug("Scheduled HeavyQueryCacheCleanupJob to run with Cron expression: {}", cronExpression);
    }

    /**
     * Try to find a cached BLOB from the heavy_query_cache table.
     *
     * @param cacheKey Entry key to look for.
     * @param time     Entries with expiry_time that are less than this value will be considered expired.
     * @param clazz    Object class to convert the result.
     * @return The cached BLOB if exists. If there is no cached entry, will return null.
     */
    @Nullable
    private <T> T readFromHeavyQueryCache(String cacheKey, long time, Class<T> clazz) {
        try {
            HeavyQueryCacheEntry hqcEntry = heavyQueryCacheDao.getCachedResponse(cacheKey, time);
            log.debug("Reading from the cache, found: {} ", hqcEntry);
            return convertToObject(hqcEntry, clazz);
        } catch (SQLException e) {
            throw new StorageException("An error has occurred when reading from cache '" + cacheKey + "': " +
                    e.getMessage(), e);
        }
    }

    /**
     * Run the query directly without looking in the cache first.
     *
     * @param callable  The callable to execute.
     * @param cacheKey  Entry key to cache for later.
     * @param hqcConfig Configuration values.
     * @return The query response.
     */
    private <T> T getCallableQueryResult(Callable<T> callable, String cacheKey, HeavyQueryCacheConfig hqcConfig) {
        try {
            long startTime = System.currentTimeMillis();
            T result = callable.call();
            long callableDuration = System.currentTimeMillis() - startTime + 1; // adding 1 to avoid 0 value
            log.debug("Callable duration: {} ms", callableDuration);
            byte[] data = convertToBytes(result);
            log.debug("Data length: {} bytes", data.length);

            if (shouldCache(callableDuration, data.length, hqcConfig)) {
                long id = dbService.nextId();
                long cacheTime = System.currentTimeMillis();
                log.debug("Cache time: {}", cacheTime);
                long expiryTime = getExpiryTime(cacheTime, callableDuration, hqcConfig);
                log.debug("Calculated expiry time: {}", expiryTime);
                HeavyQueryCacheEntry cacheEntry = HeavyQueryCacheEntry.builder()
                        .id(id)
                        .cacheKey(cacheKey)
                        .cacheTime(cacheTime)
                        .expiryTime(expiryTime)
                        .data(data)
                        .build();
                heavyQueryCacheDao.delete(cacheEntry);
                log.debug("Deleted old cache entries for: {}", cacheKey);
                heavyQueryCacheDao.create(cacheEntry);
                log.debug("Created cache entry: {}", cacheEntry);
            }
            return result;
        } catch (Exception e) {
            throw new StorageException("Failed to fetch response for '" + cacheKey + "': " + e.getMessage(), e);
        }
    }

    /**
     * Should cache or not, depending on the following logic:
     * Slower than {@link ConstantValues#hqcMinCallableDurationMillis} and smaller than
     * {@link ConstantValues#hqcMaxBlobSizeKb} in size.
     *
     * @param callableDuration Callback execution time.
     * @param blobLength       Size of the BLOB.
     * @param hqcConfig        Configuration values.
     * @return True if caching is needed. Otherwise, false.
     */
    @VisibleForTesting
    boolean shouldCache(long callableDuration, int blobLength, HeavyQueryCacheConfig hqcConfig) {
        boolean shouldCache = callableDuration > hqcConfig.getMinCallableDurationMillis()
                && (double) blobLength / SIZE_KB < hqcConfig.getMaxBlobSizeKb();
        log.debug("Should cache: {}", shouldCache);
        return shouldCache;
    }

    /**
     * Calculate the expiry time for a cache entry.
     *
     * @param cacheTime        Entry time of caching.
     * @param callableDuration Duration of the callable.
     * @param hqcConfig        Config values.
     * @return Adjusted expiry time.
     */
    @VisibleForTesting
    long getExpiryTime(long cacheTime, long callableDuration, HeavyQueryCacheConfig hqcConfig) {
        long cachePeriod = callableDuration * hqcConfig.getExpiryFactor();
        if (cachePeriod > hqcConfig.getExpiryUpperBound()) {
            return cacheTime + hqcConfig.getExpiryUpperBound();
        }
        if (cachePeriod < hqcConfig.getExpiryLowerBound()) {
            return cacheTime + hqcConfig.getExpiryLowerBound();
        }
        return cacheTime + cachePeriod;
    }

    /**
     * Convert an entry to actual requested object.
     *
     * @param hqcEntry Given cached entry.
     * @param clazz    Object class.
     * @return Converted object.
     */
    private <T> T convertToObject(HeavyQueryCacheEntry hqcEntry, Class<T> clazz) {
        log.trace("Entry: {}, Class: {}", hqcEntry, clazz);
        if (hqcEntry != null) {
            try {
                T value = JsonUtils.getInstance().readValue(hqcEntry.getData(), clazz);
                log.trace("Converted {} to {} object: {}", hqcEntry.getData(), clazz.getName(), value);
                return value;
            } catch (Exception e) {
                log.error("Failed to convert data to {} object.", clazz.getName(), e);
                log.trace("Data: {}", hqcEntry.getData());
                return null;
            }
        }
        return null;
    }

    /**
     * Convert an object to byte array.
     *
     * @param data Some object to convert.
     * @return Byte array representation of the object.
     */
    private <T> byte[] convertToBytes(T data) {
        log.trace("Data: {}", data);
        if (data != null) {
            byte[] bytes = JsonUtils.getInstance().valueToByteArray(data);
            log.trace("Converted to bytes: {}", bytes);
            return bytes;
        }
        return new byte[0];
    }

    /**
     * Get the cron expression from {@link ConstantValues#hqcCleanupJobCron} and validates it:
     * The cron must be in the range of {@link HeavyQueryCacheServiceImpl#CLEANUP_JOB_MIN_DAYS} and
     * {@link HeavyQueryCacheServiceImpl#CLEANUP_JOB_MAX_DAYS}. Otherwise,
     * {@link HeavyQueryCacheServiceImpl#CLEANUP_JOB_DEFAULT_CRON} is returned.
     *
     * @return A valid cron expression for the cleanup job.
     */
    @VisibleForTesting
    @NonNull
    String getCronExpression() {
        String cronExpression = hqcCleanupJobCron.getString();
        if (!isValid(cronExpression)) {
            log.debug("The cron {} is invalid, setting default {}.", cronExpression, CLEANUP_JOB_DEFAULT_CRON);
            return CLEANUP_JOB_DEFAULT_CRON;
        }
        DateTime nextExecution = new DateTime(getNextExecution(cronExpression));
        DateTime nextExecutionAfter = new DateTime(getNextExecutionAfter(cronExpression, nextExecution.getMillis()));
        Days intervalDays = Days.daysBetween(nextExecution, nextExecutionAfter);
        if (intervalDays.isLessThan(Days.days(CLEANUP_JOB_MIN_DAYS)) ||
                intervalDays.isGreaterThan(Days.days(CLEANUP_JOB_MAX_DAYS))) {
            log.debug("The cron {} is out of bounds. Allowed interval is {}-{} days, setting default {}.",
                    cronExpression, CLEANUP_JOB_MIN_DAYS, CLEANUP_JOB_MAX_DAYS, CLEANUP_JOB_DEFAULT_CRON);
            return CLEANUP_JOB_DEFAULT_CRON;
        }
        return cronExpression;
    }
}
