package org.artifactory.ph;

import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpRequestBase;
import org.artifactory.api.request.DownloadService;
import org.artifactory.api.request.InternalArtifactoryRequest;
import org.artifactory.api.request.JerseyArtifactoryResponse;
import org.artifactory.common.repo.PackageType;
import org.artifactory.common.resource.BaseResource;
import org.artifactory.core.api.ph.exception.PackageException;
import org.artifactory.core.api.ph.exception.PackageInternalServerError;
import org.artifactory.core.api.ph.exception.PackageNotFoundException;
import org.artifactory.core.api.ph.model.PackageArtifact;
import org.artifactory.core.api.ph.model.PackageDownloadContext;
import org.artifactory.core.api.ph.service.PackageArtifactService;
import org.artifactory.core.api.ph.virtual.resolution.PackageRequestContext;
import org.artifactory.fs.RepoResource;
import org.artifactory.repo.LocalCacheRepo;
import org.artifactory.repo.LocalRepo;
import org.artifactory.repo.RemoteRepo;
import org.artifactory.repo.Repo;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.repo.virtual.VirtualRepo;
import org.artifactory.request.DownloadRequestContext;
import org.artifactory.resource.ExpiredRepoResource;
import org.artifactory.rest.exception.RestException;
import org.artifactory.sapi.fs.VfsFile;

import javax.ws.rs.core.Response;
import java.io.IOException;
import java.io.InputStream;

import static org.apache.http.HttpStatus.SC_BAD_REQUEST;
import static org.apache.http.HttpStatus.SC_NOT_FOUND;
import static org.artifactory.repo.db.DbLocalRepo.BYPASS_UNSCANNED_BLOCKING_HEADER;

@Slf4j
public class PackageArtifactServiceImpl implements PackageArtifactService {

    private static final String REMOTE_CACHE_SUFFIX = "-cache";
    private final DownloadService downloadService;
    private final InternalRepositoryService repositoryService;

    public PackageArtifactServiceImpl(DownloadService downloadService, InternalRepositoryService repositoryService) {
        this.downloadService = downloadService;
        this.repositoryService = repositoryService;
    }

    @Override
    public Response download(PackageDownloadContext downloadContext) throws PackageException {
        InternalArtifactoryRequest request = fromPackageDownloadContext(downloadContext);
        return download(request);
    }

    @Override
    public Response downloadFromRemoteCache(PackageDownloadContext downloadContext) throws PackageException {
        String remoteRepoKey = downloadContext.getRepoKey();
        if (remoteRepoKey.endsWith(REMOTE_CACHE_SUFFIX)) {
            throw new PackageException(SC_BAD_REQUEST, "Bad request",
                    "Download request from remote-cache repo should not include '-cache' in the repo key");
        }
        LocalRepo<?> cachedRepo = repositoryService.localFederatedOrCachedRepositoryByKey(remoteRepoKey);
        if (!(cachedRepo instanceof LocalCacheRepo)) {
            String error = "Repo '" + remoteRepoKey + "' does not exist or storing artifacts locally is disabled";
            log.error(error);
            throw new PackageException(SC_BAD_REQUEST, error, error);
        }
        InternalArtifactoryRequest internalArtifactoryRequest =
                fromPackageDownloadContext(downloadContext, cachedRepo.getKey());
        return download(internalArtifactoryRequest);
    }

    private Response download(InternalArtifactoryRequest request) throws PackageException {
        try {
            JerseyArtifactoryResponse response = new JerseyArtifactoryResponse();
            downloadService.process(request, response);
            return response.build();
        } catch (RestException e) {
            throw new PackageException(e.getStatusCode(), e.getMessage(), e.getMessage());
        } catch (Exception e) {
            String message = String.format("Failed to process download for path:%s in repo:%s", request.getPath(),
                    request.getRepoKey());
            log.error(message);
            throw new PackageInternalServerError(e.getMessage());
        }
    }

    @Override
    public PackageArtifact getArtifact(String repoKey, String path) throws PackageException {
        return getArtifact(new PackageDownloadContext(repoKey, path));
    }

    @Override
    public PackageArtifact getArtifact(PackageDownloadContext downloadContext) throws PackageException {
        Repo<?> repo = repositoryService.repositoryByKey(downloadContext.getRepoKey());
        if (repo == null) {
            throw repoNotFoundException(downloadContext.getRepoKey(), downloadContext.getPath());
        }
        VfsFile<?> vfsFile = getVfsFile(downloadContext);
        PackageType packageType = PackageType.valueOf(repo.getDescriptor().getType().name().toUpperCase());
        return PackageArtifactMapper.artifact(vfsFile, packageType);
    }

    @Override
    public InputStream getStream(String repoKey, String path) throws PackageException {
        return getStream(new PackageDownloadContext(repoKey, path));
    }

    @Override
    public InputStream getStream(PackageDownloadContext downloadContext) throws PackageException {
        VfsFile<?> vfsFile = getVfsFile(downloadContext);
        return vfsFile.getStream();
    }

    @Override
    public BaseResource getResource(String repoKey, String path) throws PackageException {
        return getResource(new PackageDownloadContext(repoKey, path));
    }

    @Override
    public BaseResource getResource(PackageRequestContext packageRequestContext) throws PackageException {
        String repoKey = packageRequestContext.getRepoKey();
        String path = packageRequestContext.getPath();

        Repo<?> repo = repositoryService.repositoryByKey(repoKey);
        if (repo == null || !repo.isReal()) {
            throw repoNotFoundException(repoKey, path);
        }
        RepoPath repoPath = RepoPathFactory.create(repoKey, path);
        InternalArtifactoryRequest request = new InternalArtifactoryRequest(repoPath, false);
        updateRequestFromRequestContext(request, packageRequestContext);
        return repo.getInfo(new DownloadRequestContext(request));
    }

    @Override
    public BaseResource getResourceFromVirtualCache(String repoKey, String path) throws PackageException {
        RepoPath repoPath = RepoPathFactory.create(repoKey, path);
        VirtualRepo virtualRepo = repositoryService.virtualRepositoryByKey(repoKey);
        if (virtualRepo == null) {
            throw repoNotFoundException(repoKey, path);
        }
        InternalArtifactoryRequest request = new InternalArtifactoryRequest(repoPath, false);
        RepoResource cachedResource = virtualRepo.getInfoFromVirtualCache(new DownloadRequestContext(request));
        if (cachedResource.isFound() && virtualRepo.isCachedResourceExpired(cachedResource)) {
            cachedResource = new ExpiredRepoResource(cachedResource);
        }
        return cachedResource;
    }

    @Override
    public CloseableHttpResponse executeRemoteHttpRequest(String remoteRepoKey, HttpRequestBase httpRequest)
            throws PackageException {
        RemoteRepo<?> remoteRepo = repositoryService.remoteRepositoryByKey(remoteRepoKey);
        if (remoteRepo == null) {
            String message = "Remote repo " + remoteRepoKey + " does not exist";
            throw new PackageException(SC_NOT_FOUND, message, message);
        }
        if (remoteRepo.isOffline() || remoteRepo.isAssumedOffline() || remoteRepo.isBlackedOut()) {
            String message = "Remote repo " + remoteRepoKey + " is offline";
            throw new PackageException(SC_NOT_FOUND, message, message);
        }
        try {
            return remoteRepo.getHttpClient().execute(httpRequest);
        } catch (IOException e) {
            log.error("An error occurred while executing http request in {}: {}", remoteRepoKey, e.getMessage());
            log.debug("", e);
            String message = "Failed to execute http request with repo " + remoteRepoKey;
            throw new PackageException(SC_BAD_REQUEST, message, message, e);
        }
    }

    private static PackageException repoNotFoundException(String repoKey, String path) {
        String message = String.format("Failed to find repository in repo:%s, path:%s", repoKey, path);
        return new PackageNotFoundException(message);
    }

    private InternalArtifactoryRequest fromPackageDownloadContext(PackageDownloadContext downloadContext) {
        return fromPackageDownloadContext(downloadContext, null);
    }

    private InternalArtifactoryRequest fromPackageDownloadContext(PackageDownloadContext downloadContext,
            String overrideRepoKey) {
        final String repoKey = overrideRepoKey == null ? downloadContext.getRepoKey() : overrideRepoKey;
        RepoPath repoPath = RepoPathFactory.create(repoKey, downloadContext.getPath());
        InternalArtifactoryRequest request =
                new InternalArtifactoryRequest(repoPath, downloadContext.isDisableRedirect());
        downloadContext.getDelegationAllowedHeaders().forEach(request::setHeaderEligibleForDelegation);
        request.setZipResourcePath(downloadContext.getZipResourcePath());
        request.setReplaceHeadRequestWithGet(downloadContext.isReplaceHeadRequestWithGet());
        updateRequestFromRequestContext(request, downloadContext);
        request.setAlternateRemoteSiteUrl(downloadContext.getAlternateRemoteBaseUrl());
        request.setRequestMethod(downloadContext.getHttpMethod());
        request.setSkipEncoding(downloadContext.isSkipEncoding());
        request.setReplicationOriginatedDownloadRequest(downloadContext.isReplicationOriginatedDownloadRequest());

        if (downloadContext.getOriginalRepoKey() != null) {
            RepoPath originalRepoPath =
                    RepoPathFactory.create(downloadContext.getOriginalRepoKey(), downloadContext.getPath());
            request.setOriginalRepoPath(originalRepoPath);
        }
        return request;
    }

    private void updateRequestFromRequestContext(InternalArtifactoryRequest internalArtifactoryRequest,
            PackageRequestContext downloadContext) {
        if (StringUtils.isNotBlank(downloadContext.getAlternativeUrl())) {
            internalArtifactoryRequest.setAlternativeRemoteDownloadUrl(downloadContext.getAlternativeUrl());
        }
        internalArtifactoryRequest.addHeaders(downloadContext.getRequestHeaders());
        internalArtifactoryRequest.setServletContextUrl(downloadContext.getServletContextUrl());
    }

    @NonNull
    private VfsFile<?> getVfsFile(PackageDownloadContext downloadContext) throws PackageException {
        Repo<?> repo = repositoryService.repositoryByKey(downloadContext.getRepoKey());
        if (repo == null) {
            throw repoNotFoundException(downloadContext.getRepoKey(), downloadContext.getPath());
        }
        InternalArtifactoryRequest request = fromPackageDownloadContext(downloadContext);
        request.addHeader(BYPASS_UNSCANNED_BLOCKING_HEADER, Boolean.TRUE.toString());
        return ArtifactExtractor.locateFile(new DownloadRequestContext(request), repositoryService);
    }
}
