package org.artifactory.ph;

import lombok.extern.slf4j.Slf4j;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.authentication.ArtifactoryAuthenticationContextFilter;
import org.artifactory.authentication.ArtifactoryBasicAuthenticationFilter;
import org.artifactory.authentication.Context.AuthenticationContext;
import org.artifactory.core.api.ph.model.PackageAuthenticationRequestContext;
import org.artifactory.core.api.ph.service.PackageAuthenticationAdapter;
import org.artifactory.webapp.servlet.RequestUtils;

import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.ws.rs.core.Response;
import java.io.IOException;

import static javax.ws.rs.core.HttpHeaders.WWW_AUTHENTICATE;

/**
 * @author shahafg
 */
@Slf4j
public class PackageAuthenticationFilterImpl implements ArtifactoryAuthenticationContextFilter {

    private final PackageAuthenticationAdapter packageAuthenticationAdapter;

    private ArtifactoryBasicAuthenticationFilter basicFilter;

    public PackageAuthenticationFilterImpl(
            PackageAuthenticationAdapter packageAuthenticationAdapter) {
        this.packageAuthenticationAdapter = packageAuthenticationAdapter;
    }

    public void init(FilterConfig filterConfig) throws ServletException {
        ServletContext servletContext = filterConfig.getServletContext();
        ArtifactoryContext context = RequestUtils.getArtifactoryContext(servletContext);
        basicFilter = context.beanForType(ArtifactoryBasicAuthenticationFilter.class);
        log.debug("Initializing finished");
    }

    @Override
    public void doFilter(ServletRequest servletRequest, ServletResponse servletResponse, FilterChain filterChain)
            throws IOException, ServletException {
        basicFilter.doFilter(servletRequest, servletResponse, filterChain);
    }

    @Override
    public AuthenticationContext getAuthenticationContext(ServletRequest request, ServletResponse response)
            throws IOException {
        PackageAuthenticationRequestContext context =
                this.packageAuthenticationAdapter.getContext((HttpServletRequest) request);
        sendResponseErrorIfNeeded((HttpServletResponse) response, context.getResponseError());
        if (context.isAuthenticationAllowed()) {
            return AuthenticationContext.builder().authenticated(true)
                    .cacheKey(getCacheKey(request, context))
                    .loginIdentifier(getLoginIdentifier(request, context))
                    .extractedAuthorizationToken(context.getAuthorizationToken()).build();
        }
        return AuthenticationContext.builder().anonymousAllowed(context.isAnonymousRequestAllowed())
                .build();
    }

    private void sendResponseErrorIfNeeded(HttpServletResponse response, Response responseError) throws IOException {
        if (responseError != null) {
            response.setStatus(responseError.getStatus());

            if (responseError.getMetadata() != null && responseError.getMetadata().get(WWW_AUTHENTICATE) != null) {
                String authValue = responseError.getMetadata().get(WWW_AUTHENTICATE).get(0).toString();
                response.addHeader(WWW_AUTHENTICATE, authValue);
            }

            if (responseError.getEntity() != null) {
                response.getWriter().write(responseError.getEntity().toString() + "\n");
            }
        }
    }

    private String getCacheKey(ServletRequest request, PackageAuthenticationRequestContext context) {
        if (context.getAuthorizationToken() != null) {
            return context.getAuthorizationToken();
        }
        return this.basicFilter.getCacheKey(request);
    }

    private String getLoginIdentifier(ServletRequest request, PackageAuthenticationRequestContext context) {
        if (context.getPrincipleUserName() != null) {
            return context.getPrincipleUserName();
        }
        return this.basicFilter.getLoginIdentifier(request);
    }
}