package org.artifactory.ph;

import com.google.common.collect.HashMultimap;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.aql.AqlConverts;
import org.artifactory.aql.AqlService;
import org.artifactory.aql.api.domain.sensitive.AqlApiItem;
import org.artifactory.aql.api.internal.AqlBase;
import org.artifactory.aql.model.AqlComparatorEnum;
import org.artifactory.aql.model.AqlItemTypeEnum;
import org.artifactory.aql.result.AqlEagerResult;
import org.artifactory.aql.result.AqlLazyResult;
import org.artifactory.aql.result.rows.AqlItem;
import org.artifactory.aql.result.rows.AqlLazyObjectResultStreamer;
import org.artifactory.aql.result.rows.FileInfoWithStatisticsItemRow;
import org.artifactory.core.api.ph.exception.PackageException;
import org.artifactory.core.api.ph.exception.PackageInternalServerError;
import org.artifactory.core.api.ph.model.FilteringProperty;
import org.artifactory.core.api.ph.model.PackageArtifact;
import org.artifactory.core.api.ph.model.PackageSearchContext;
import org.artifactory.core.api.ph.service.PackageSearchArtifactService;
import org.jfrog.client.util.PathUtils;

import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static org.artifactory.aql.api.domain.sensitive.AqlApiItem.create;
import static org.artifactory.aql.api.domain.sensitive.AqlApiItem.createWithEmptyResults;
import static org.artifactory.aql.api.domain.sensitive.AqlApiItem.depth;
import static org.artifactory.aql.api.domain.sensitive.AqlApiItem.name;
import static org.artifactory.aql.api.domain.sensitive.AqlApiItem.path;
import static org.artifactory.aql.api.domain.sensitive.AqlApiItem.property;
import static org.artifactory.aql.api.domain.sensitive.AqlApiItem.repo;
import static org.artifactory.aql.api.domain.sensitive.AqlApiItem.statistic;
import static org.artifactory.aql.api.domain.sensitive.AqlApiItem.type;
import static org.artifactory.aql.api.internal.AqlBase.and;

@Slf4j
public class PackageSearchArtifactServiceImpl implements PackageSearchArtifactService {

    public static final String PARAMETER = "Parameter '";
    private final AqlService aqlService;
    private final AuthorizationService authorizationService;

    public PackageSearchArtifactServiceImpl(AqlService aqlService,
            AuthorizationService authorizationService) {
        this.aqlService = aqlService;
        this.authorizationService = authorizationService;
    }

    @Override
    public Stream<PackageArtifact> findArtifactsByExtension(String repoKey, String path, String... extensions) {
        path = StringUtils.isBlank(path) ? null : path.trim();
        assertValidRepoInput(repoKey);
        assertNoAstrix(path);
        assertValidExtensions(extensions);
        // Creating the query with no extra fields
        AqlApiItem query = AqlApiItem.createWithEmptyResults();
        // Adding the fields we need
        buildIncludeQuery(query);
        // Preparing the query filters
        AqlBase.AndClause<AqlApiItem> and = AqlBase.and();
        and.append(AqlApiItem.repo().equal(repoKey));
        filterByPathEqualOrGreater(and, path);
        filterByExtension(and, extensions);
        filterByDepth(and, path);
        and.append(AqlApiItem.type().equal(AqlItemTypeEnum.file.signature));
        // Adding the AND filter to the query
        query.filter(and);
        return executeAqlQueryAndMapResults(query);
    }

    @Override
    public Stream<PackageArtifact> findArtifactsChildren(String repoKey, String path) {
        path = StringUtils.isBlank(path) ? null : path.trim();
        assertValidRepoInput(repoKey);
        assertNoAstrix(path);
        // Creating the query with no extra fields
        AqlApiItem query = AqlApiItem.createWithEmptyResults();
        // Adding the fields we need
        buildIncludeQuery(query);
        // Preparing the query filters
        AqlBase.AndClause<AqlApiItem> and = AqlBase.and();
        and.append(AqlApiItem.repo().equal(repoKey));
        filterByPathEqualOrGreater(and, path);
        filterByDepth(and, path);
        and.append(AqlApiItem.type().equal(AqlItemTypeEnum.file.signature));
        // Adding the AND filter to the query
        query.filter(and);
        return executeAqlQueryAndMapResults(query);
    }

    @Override
    public Stream<PackageArtifact> findArtifactsChildrenStartWithName(String repoKey, String path, String name) {
        path = StringUtils.isBlank(path) ? null : path.trim();
        name = StringUtils.isBlank(name) ? null : name.trim();
        assertValidRepoInput(repoKey);
        assertNoAstrix(path);
        assertInputNotEmpty(name);
        assertNoAstrix(name);
        // Creating the query with no extra fields
        AqlApiItem query = AqlApiItem.createWithEmptyResults();
        // Adding the fields we need
        buildIncludeQuery(query);
        // Preparing the query filters
        AqlBase.AndClause<AqlApiItem> and = AqlBase.and();
        and.append(AqlApiItem.repo().equal(repoKey));
        filterByPathEqualOrGreater(and, path);
        filterByName(and, name + "*");
        filterByDepth(and, path);
        and.append(AqlApiItem.type().equal(AqlItemTypeEnum.file.signature));
        // Adding the AND filter to the query
        query.filter(and);
        return executeAqlQueryAndMapResults(query);
    }

    @Override
    public Stream<PackageArtifact> findArtifactsByExactName(String repoKey, String name) {
        name = StringUtils.isBlank(name) ? null : name.trim();
        assertValidRepoInput(repoKey);
        assertInputNotEmpty(name);
        assertNoAstrix(name);
        // Creating the query with no extra fields
        AqlApiItem query = AqlApiItem.createWithEmptyResults();
        // Adding the fields we need
        buildIncludeQuery(query);
        // Preparing the query filters
        AqlBase.AndClause<AqlApiItem> and = AqlBase.and();
        and.append(AqlApiItem.repo().equal(repoKey));
        filterByName(and, name);
        filterByDepth(and, null);
        and.append(AqlApiItem.type().equal(AqlItemTypeEnum.file.signature));
        // Adding the AND filter to the query
        query.filter(and);
        return executeAqlQueryAndMapResults(query);
    }

    @Override
    public Optional<PackageArtifact> findArtifactByExactPathAndName(String repoKey, String path, String name) {
        path = StringUtils.isBlank(path) ? null : path.trim();
        name = StringUtils.isBlank(name) ? null : name.trim();
        assertValidRepoInput(repoKey);
        assertInputNotEmpty(name);
        assertNoAstrix(name);
        assertNoAstrix(path);
        // Creating the query with no extra fields
        AqlApiItem query = AqlApiItem.createWithEmptyResults();
        // Adding the fields we need
        buildIncludeQuery(query);
        // Preparing the query filters
        AqlBase.AndClause<AqlApiItem> and = AqlBase.and();
        and.append(AqlApiItem.repo().equal(repoKey));
        filterByPath(and, path);
        filterByName(and, name);
        filterByDepth(and, path);
        and.append(AqlApiItem.type().equal(AqlItemTypeEnum.file.signature));
        // Adding the AND filter to the query
        query.filter(and);
        return executeAqlQueryAndMapResults(query).findFirst();
    }

    @Override
    public Stream<PackageArtifact> findArtifactsByPropKeys(String repoKey, String path, String... propKeys) {
        path = StringUtils.isBlank(path) ? null : path.trim();
        assertValidRepoInput(repoKey);
        assertNoAstrix(path);
        assertValidPropKeys(propKeys);
        // Creating the query with no extra fields
        AqlApiItem query = AqlApiItem.createWithEmptyResults();
        // Adding the fields we need
        buildIncludeQuery(query);
        // Preparing the query filters
        AqlBase.AndClause<AqlApiItem> and = AqlBase.and();
        and.append(AqlApiItem.repo().equal(repoKey));
        filterByPath(and, path);
        filterByDepth(and, path);
        filterByPropKeys(and, propKeys);
        and.append(AqlApiItem.type().equal(AqlItemTypeEnum.file.signature));
        // Adding the AND filter to the query
        query.filter(and);
        return executeAqlQueryAndMapResults(query);
    }

    @Override
    public Stream<PackageArtifact> findArtifactsWithProps(String repoKey, String path,
            List<FilteringProperty> filterProperties) throws PackageException {
        path = StringUtils.isBlank(path) ? null : path.trim();
        assertValidRepoInput(repoKey);
        assertNoAstrix(path);

        AqlApiItem aqlQuery = create().distinct(false)
                .include(statistic().downloads(), property().key(), property().value());

        // Preparing the query filters
        AqlBase.AndClause<AqlApiItem> and = AqlBase.and();
        and.append(AqlApiItem.repo().equal(repoKey));
        filterByPath(and, path);
        filterByDepth(and, path);
        for (FilteringProperty filterProperty : filterProperties) {
            final AqlBase.CriteriaClause<AqlApiItem> propertyValueCriteria;
            // if property should be searched ignore case.
            if (filterProperty.isCaseSensitive()) {
                propertyValueCriteria = property().value().equal(filterProperty.getPropertyValue());
            } else {
                propertyValueCriteria =
                        property().value().equalIgnoreCase(filterProperty.getPropertyValue().toUpperCase());

            }
            and.append(and(property().key().equal(filterProperty.getPropertyName()), propertyValueCriteria));
        }
        aqlQuery.filter(and);

        return populateRowsToSearchResult(aqlQuery, repoKey).stream();
    }

    @NonNull
    @Override
    public List<String> findArtifactsByPattern(@NonNull String repoKey, @NonNull String filename, String pathPattern) {
        if (StringUtils.isEmpty(pathPattern) || !pathPattern.contains("*") || !pathPattern.contains("/")) {
            throw new IllegalStateException("Search path is either empty, or doesn't contain '*' or '/'");
        }
        AqlApiItem query = createWithEmptyResults().filter(
                and(
                        repo().equal(repoKey),
                        name().equal(filename),
                        path().matches(pathPattern)

                )
        ).include(name(), repo(), path(), type(), depth());

        return aqlService.executeQueryEager(query).getResults().stream()
                .map(item -> String.join("/", item.getPath(), item.getName())).collect(Collectors.toList());
    }

    private Collection<PackageArtifact> populateRowsToSearchResult(AqlApiItem aqlQuery, String repoKey)
            throws PackageException {
        FileInfoWithStatisticsItemRow row;
        long queryStart = System.currentTimeMillis();
        Map<Long, PackageArtifact> searchResults = new HashMap<>();
        try (AqlLazyResult<AqlItem> result = aqlService.executeQueryLazy(aqlQuery)) {
            AqlLazyObjectResultStreamer<FileInfoWithStatisticsItemRow> streamer =
                    new AqlLazyObjectResultStreamer<>(result, FileInfoWithStatisticsItemRow.class);
            while ((row = streamer.getRow()) != null) {
                accumulateRowToCollection(row, searchResults);
            }
        } catch (Exception e) {
            log.error("Failed to map AQL results for populateRowsToSearchResult: " + e.getMessage(), e);
            log.debug("Query was: {}", aqlQuery.toNative(2));
            throw new PackageInternalServerError("Failed to execute populateRowsToSearchResult", e);
        }
        long queryEnd = System.currentTimeMillis();
        log.debug("Aql query with reducing properties took: {} ms, Number of elements:{}, Repo:{}",
                (queryEnd - queryStart), searchResults.values().size(), repoKey);

        return searchResults.values();
    }

    private void accumulateRowToCollection(FileInfoWithStatisticsItemRow row,
            Map<Long, PackageArtifact> searchResults) {
        long currentPackageId = row.getItemId();
        PackageArtifact packageWithProps = searchResults
                .computeIfAbsent(currentPackageId, key -> toPackageArtifact(row));
        String propKey = row.getPropKey();
        if (propKey != null) {
            //Add all properties
            String propVal = row.getPropVal();
            packageWithProps.getProperties().put(propKey, propVal);
            if (log.isTraceEnabled()) {
                log.trace("Added property {} : {} to PackageArtifact for path {}", propKey, propVal,
                        packageWithProps.getPath());
            }
        }
    }

    @NonNull
    private PackageArtifact toPackageArtifact(final FileInfoWithStatisticsItemRow row) {
        final PackageArtifact.PackageArtifactBuilder builder = PackageArtifact.builder();
        final HashMultimap<String, String> properties = HashMultimap.create();
        builder.properties(properties);

        builder.path(row.getPath());
        builder.created(row.getCreated());
        builder.name(row.getName());
        builder.lastUpdated(row.toFileInfo().getLastUpdated());
        builder.lastModified(row.getModified());
        builder.length(row.getSize());
        builder.downloadCount(row.getStatDownloaded());
        return builder.build();
    }

    @Override
    public Stream<PackageArtifact> findArtifactsByPropKeyAndVal(PackageSearchContext packageSearchContext,
            Map<String, String> properties) {
        String repoKey = packageSearchContext.getRepoKey();
        String path =
                StringUtils.isBlank(packageSearchContext.getPath()) ? null : packageSearchContext.getPath().trim();
        assertValidRepoInput(repoKey);
        assertNoAstrix(path);
        assertValidProperties(properties);
        // Creating the query with no extra fields
        AqlApiItem query = AqlApiItem.createWithEmptyResults();
        // Adding the fields we need
        buildIncludeQuery(query);
        // Preparing the query filters
        AqlBase.AndClause<AqlApiItem> and = AqlBase.and();
        and.append(AqlApiItem.repo().equal(repoKey));
        filterByPath(and, path);
        filterByDepth(and, path);
        filterByProperties(and, properties);
        and.append(AqlApiItem.type().equal(AqlItemTypeEnum.file.signature));
        if (packageSearchContext.getLimit() > 0) {
            query.limit(packageSearchContext.getLimit());
        }
        query.offset(packageSearchContext.getOffset());
        // Adding the AND filter to the query
        query.filter(and);
        return executeAqlQueryAndMapResults(query);
    }

    private void filterByProperties(AqlBase.AndClause<AqlApiItem> and, Map<String, String> propertiesToMatch) {
        for (Map.Entry<String, String> entry : propertiesToMatch.entrySet()) {
            String value = entry.getValue();
            if (value == null) {
                and.append(property().property(entry.getKey(), AqlComparatorEnum.equals, null));
            } else if (value.equals("*")) {
                throw new IllegalArgumentException(PARAMETER + "*' is not a legal property value");
            } else {
                and.append(property().property(
                        entry.getKey(),
                        value.contains("*") ? AqlComparatorEnum.matches : AqlComparatorEnum.equals,
                        value));
            }
        }
    }

    private void buildIncludeQuery(AqlApiItem query) {
        query.include(AqlApiItem.repo(), AqlApiItem.itemId(), AqlApiItem.path(), AqlApiItem.name(), AqlApiItem.type(),
                AqlApiItem.sha1Actual(), AqlApiItem.md5Actual(), AqlApiItem.sha2(), AqlApiItem.size(),
                AqlApiItem.created(), AqlApiItem.modified(), AqlApiItem.createdBy(), AqlApiItem.modifiedBy(),
                AqlApiItem.updated());
    }

    private void assertNoAstrix(String param) {
        if (StringUtils.isNotEmpty(param) && param.contains("*")) {
            throw new IllegalArgumentException(PARAMETER + param + "' cannot include '*'");
        }
    }

    private void assertValidRepoInput(String repoKey) {
        if (StringUtils.isEmpty(repoKey)) {
            throw new IllegalArgumentException("RepoKey cannot be empty");
        }
    }

    private void assertInputNotEmpty(String input) {
        if (StringUtils.isEmpty(input)) {
            throw new IllegalArgumentException("Input cannot be empty");
        }
    }

    private void assertValidPropKeys(String[] propKeys) {
        if (propKeys == null || propKeys.length == 0) {
            throw new IllegalArgumentException("No propKeys has been provided");
        }
        Arrays.stream(propKeys).iterator().forEachRemaining(extension -> {
            if (StringUtils.isEmpty(extension)) {
                throw new IllegalArgumentException("PropKey cannot be empty");
            }
            assertNoAstrix(extension);
        });
    }

    private void assertValidProperties(Map<String, String> properties) {
        if (properties == null || properties.size() == 0) {
            throw new IllegalArgumentException("No properties has been provided");
        }
        properties.forEach((k, v) -> {
            if (StringUtils.isEmpty(k)) {
                throw new IllegalArgumentException("PropKey cannot be empty");
            }
            assertNoAstrix(k);
        });
    }

    private void assertValidExtensions(String... extensions) {
        if (extensions == null || extensions.length == 0) {
            throw new IllegalArgumentException("No extension has been provided");
        }
        Arrays.stream(extensions).iterator().forEachRemaining(extension -> {
            if (StringUtils.isEmpty(extension)) {
                throw new IllegalArgumentException("Extension cannot be empty");
            }
            assertNoAstrix(extension);
        });
    }

    private void filterByExtension(AqlBase.AndClause<AqlApiItem> and, String[] extensions) {
        AqlBase.OrClause<AqlApiItem> or = AqlBase.or();
        Arrays.stream(extensions).iterator().forEachRemaining(extension -> {
                    extension = StringUtils.removeStart(extension, ".");
                    or.append(AqlApiItem.name().matches("*." + extension));
                }
        );
        and.append(or);
    }

    /**
     * This method adds filters to the query in case filename has wildcards.
     * Take into consideration DB performance.
     *
     * @param fileName the file name to search for
     * @param and      the AND clause from the original AQL query
     */
    private void filterByName(AqlBase.AndClause<AqlApiItem> and, String fileName) {
        if (StringUtils.isNotBlank(fileName) && !fileName.equals("*")) {
            if (fileName.contains("*")) {
                and.append(AqlApiItem.name().matches(fileName));
            } else {
                and.append(AqlApiItem.name().equal(fileName));
            }
        }
    }

    private void filterByPath(AqlBase.AndClause<AqlApiItem> and, String path) {
        if (StringUtils.isNotBlank(path) && !path.equals("*")) {
            if (path.contains("*")) {
                and.append(AqlApiItem.path().matches(path));
            } else {
                and.append(AqlApiItem.path().equal(path));
            }
        }
    }

    private void filterByPathEqualOrGreater(AqlBase.AndClause<AqlApiItem> and, String path) {
        if (StringUtils.isEmpty(path) || StringUtils.equals(path, "/")) {
            return;
        }
        path = PathUtils.trimTrailingSlashes(path);
        AqlBase.OrClause<AqlApiItem> or = AqlBase.or();
        or.append(AqlApiItem.path().equal(path));
        or.append(AqlApiItem.path().matches(path + "/*"));
        and.append(or);
    }

    private void filterByPropKeys(AqlBase.AndClause<AqlApiItem> and, String... propKeys) {
        Arrays.stream(propKeys).iterator().forEachRemaining(propKey -> {
                    and.append(property().key().equal(propKey));
                }
        );
    }

    /**
     * This method is adding a filter to the query for the 'nodes.depth' field.
     * We calculate the level of the file within the tree of the given path:
     * meaning for a given path we count the number of '/' in it,
     * then adding +1 in order to search only for the given folder and not its parent.
     * We search by 'depth >= level+1' in order to include any file in sub-directories as well.
     *
     * @param path the path given to search in
     * @param and  the AND clause from the original AQL query
     */
    private void filterByDepth(AqlBase.AndClause and, String path) {
        int length = getPathElements(path).length;
        and.append(AqlApiItem.depth().greaterEquals(length + 1));
        log.debug("Added AND clause to the artifact search query: AND depth>={}", length + 1);
    }

    /**
     * Returns the path elements of the input path.
     * <pre>
     * getPathElements("/a/b/c") = [a, b, c]
     * getPathElements("a/b/c") = [a, b, c]
     * getPathElements("a/b/c/") = [a, b, c]
     * getPathElements("a") = [a]
     * getPathElements("") = []
     * </pre>
     *
     * @param path The path to parse (can be absolute or relative)
     * @return The path's path elements
     */
    public static String[] getPathElements(String path) {
        if (StringUtils.isEmpty(path)) {
            return new String[0];
        }
        if (path.startsWith("/")) {
            // we don't want to return empty string as a path element so remove the leading slash
            path = path.substring(1);
        }
        return path.split("/");
    }

    private Stream<PackageArtifact> executeAqlQueryAndMapResults(AqlApiItem query) {
        log.debug("Executing artifact search query");
        AqlEagerResult<AqlItem> allBlobs = aqlService.executeQueryEager(query);
        log.debug("Artifacts search completed");
        return allBlobs.getResults().stream()
                .map(AqlConverts.toFileInfo)
                .filter(file -> authorizationService.canRead(file.getRepoPath()))
                .map(PackageArtifactMapper::artifact);
    }
}
