package org.artifactory.queue.services;

import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.concurrent.ArtifactoryExecutor;
import org.artifactory.concurrent.ArtifactoryExecutorFactory;
import org.artifactory.concurrent.ThreadPoolConfig;
import org.artifactory.event.queue.QueueConfiguration;
import org.artifactory.event.queue.QueueEventAggregator;
import org.artifactory.event.queue.QueueEventFilter;
import org.artifactory.event.queue.QueueEventImplementor;
import org.artifactory.event.queue.QueueStatus;
import org.artifactory.event.work.NodeEventsErrorsBackOffStrategy;
import org.artifactory.queue.QueueEventImpl;
import org.artifactory.queue.behaviors.QueueReductionBehavioursImpl;
import org.artifactory.queue.error.PersistentQueueBackoffStrategy;
import org.artifactory.queue.error.PersistentQueueErrorService;
import org.artifactory.queue.execution.QueueTriggeringService;
import org.artifactory.queue.model.PersistentQueueGroupContext;
import org.artifactory.queue.model.PersistentQueueRunningContext;
import org.artifactory.queue.model.PersistentQueueRunningResult;
import org.artifactory.storage.db.event.error.model.EventAndErrorInfo;
import org.artifactory.storage.db.event.model.NodeEventCursor;
import org.artifactory.storage.event.EventsService;
import org.jfrog.common.ClockUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

/**
 * The PersistentQueueImpl is the main class where all the work happens.
 *
 * Those are the main steps when triggering a queue.
 * 1. get a lock (DB) handled by workQueue
 * 2. get the current marker (DB)
 * 3. fetch the next events batch + fetch errors (DB)
 * 4. filter out events based on deduplication rules and applicative code (APP)
 * 5. aggregate into groups (APP)
 * 6. call the applicative implementation logic for each group in a different thread (APP)
 * 7. handles errors and update the marker (DB)
 * 8. release the lock (DB) handled by workQueue
 *
 * @author Avishay Halpren
 */
public class PersistentQueueImpl implements PersistentQueue {

    private static final Logger log = LoggerFactory.getLogger(PersistentQueueImpl.class);

    private final String cursorName;
    private final String repoKey;
    private final QueueConfiguration queueConfiguration;
    private final ArtifactoryExecutor jobsExecutor;
    private final EventsService eventsService;
    private final PersistentQueueCursorService persistentQueueCursorService;
    private final PersistentQueueErrorService persistentQueueErrorService;
    private final QueueTriggeringService triggeringService;
    private boolean shutdown;
    private boolean paused;
    private final NodeEventsErrorsBackOffStrategy backoffStrategy;

    public PersistentQueueImpl(QueueConfiguration queueConfiguration, String repoKey) {
        this.queueConfiguration = queueConfiguration;
        this.repoKey = repoKey;
        this.eventsService = ContextHelper.get().beanForType(EventsService.class);
        this.persistentQueueErrorService = ContextHelper.get().beanForType(PersistentQueueErrorService.class);
        this.persistentQueueCursorService = ContextHelper.get().beanForType(PersistentQueueCursorService.class);
        this.triggeringService = ContextHelper.get().beanForType(QueueTriggeringService.class);
        this.backoffStrategy =
                new PersistentQueueBackoffStrategy(queueConfiguration.getBackoffSupplier().getFactor(),
                        queueConfiguration.getBackoffSupplier().getMaxWaitHours());

        //set cursor name
        this.cursorName = new PersistentQueueCursorInfo(repoKey,
                queueConfiguration.getCursorType(),
                queueConfiguration.getVariant()
        ).getCursorName();
        this.jobsExecutor = ArtifactoryExecutorFactory.newClassicThreadPool(ThreadPoolConfig.builder()
                // core is 0, since if no there is no Aggregator the groups will always be 1,
                // When only one group the execution is on the triggering thread
                .maxThreads(queueConfiguration.getThreads())
                .keepAliveTimeout(30L)
                .keepAliveTimeoutTimeUnit(TimeUnit.SECONDS)
                //Limit the blocking queue inorder to make the max threads effective
                //The max jobs we can have is the max events we load + the max errors we can have,
                //Doubling it in order to be on the safe side
                .queue(new LinkedBlockingQueue<>(
                        (queueConfiguration.getMaxBulkSize() + queueConfiguration.getErrorOverflowLimit()) * 2))
                .daemon(true)
                .threadNameFormat("art-" + cursorName + "-%d")
                .build(), ContextHelper.get(), false);
    }

    /**
     * Create cursor marker if not exist.
     */
    public void createMarkerIfNotExists() {
        // Create cursor for marker if not exists
        this.persistentQueueCursorService.createMarkerIfNotExists(this.repoKey, this.cursorName,
                this.queueConfiguration.getCursorType());
    }

    @Override
    public String getName() {
        return cursorName;
    }

    @Override
    public String getRepoKey() {
        return this.repoKey;
    }

    @Override
    public QueueConfiguration getQueueConfiguration() {
        return queueConfiguration;
    }

    @Override
    public TriggerResult trigger() {
        if (shutdown || paused) {
            return TriggerResult.NOT_AVAILABLE;
        }
        TriggerResult triggerResult = null;
        try {
            PersistentQueueRunnerReport.start(cursorName);
            //triggering the queue
            triggerResult = innerExecute();
        } finally {
            PersistentQueueRunnerReport.end(triggerResult);
        }
        return triggerResult;
    }

    protected TriggerResult innerExecute() {
        PersistentQueueRunningContext runningContext = new PersistentQueueRunningContext();
        NodeEventCursor nodeEventCursor = getCurrentMarkerCursorForQueue();
        log.debug("Current event cursor set to: {}", nodeEventCursor.getEventMarker());
        runningContext.setStartingMarker(nodeEventCursor.getEventMarker());
        runningContext.setCursorsLastUpdated(nodeEventCursor.getLastUpdated());
        long lastEvent = persistentQueueCursorService.getCachedLatestEventId();
        //find all relevant DB events
        List<QueueEventImpl> loadedEvents = loadNextEvents(runningContext);
        //no events - nothing to do
        if (loadedEvents.isEmpty()) {
            if (nodeEventCursor.getEventMarker() < lastEvent) {
                log.debug("Next events is empty resetting market to {}", lastEvent);
                updateMarker(lastEvent, runningContext.getCursorsLastUpdated());
            }
            return TriggerResult.NOTHING_TO_DO;
        }
        //filter out identical records
        filterOutEventsBehaviours(loadedEvents, runningContext);

        Collection<PersistentQueueGroupContext> jobs = createJobGroups(runningContext);
        spreadTheWork(jobs);
        return handleResult(runningContext, jobs);
    }

    /**
     * get marker from cursor table by queueName (create if needed for 1st time)
     */
    private NodeEventCursor getCurrentMarkerCursorForQueue() {
        NodeEventCursor nodeEventCursor =
                persistentQueueCursorService.getCursorAndCreateIfNeeded(
                        cursorName,
                        queueConfiguration.getCursorType());
        PersistentQueueRunnerReport.startMarker(nodeEventCursor.getEventMarker());
        return nodeEventCursor;
    }

    /**
     * load the events from the lastMarker (merged with previous errors)
     */
    private List<QueueEventImpl> loadNextEvents(PersistentQueueRunningContext context) {
        //fetch events from main table
        String fileSuffix = queueConfiguration.getFileSuffix();
        int maxBulkSize = queueConfiguration.getMaxBulkSize();
        List<QueueEventImpl> events =
                eventsService.getEventsAfter(context.getStartingMarker(), repoKey, fileSuffix, maxBulkSize).stream()
                        .map(QueueEventImpl::new).collect(Collectors.toList());

        //fetch and separate errors into 2 groups: include and excluded
        List<EventAndErrorInfo> errors = persistentQueueErrorService.getErrors(this);
        markOverFlowLimitErrorsIfNeeded(errors);
        Map<Long, EventAndErrorInfo> included = new LinkedHashMap<>();
        Set<Long> excluded = new HashSet<>();

        long currentTime = ClockUtils.epochMillis();
        for (EventAndErrorInfo error : errors) {
            if (backoffStrategy.isBackoffDelayRequired(error, currentTime)) {
                //we still need the excluded list so we can filter out events from the main query
                excluded.add(error.getEventInfo().getEventId());
            } else {
                //add error event to result
                included.put(error.getEventInfo().getEventId(), error);
            }
        }
        //merging result:
        //result = included + (events - included - excludes)
        List<QueueEventImpl> result = included.values().stream()
                .map(EventAndErrorInfo::getEventInfo)
                .map(QueueEventImpl::new)
                .collect(Collectors.toList());
        events.stream()
                //filter if exist in included (already inside result)
                .filter(e -> !included.containsKey(e.getEventId()))
                //filter if exist in excluded (should not be handled due to time)
                .filter(e -> !excluded.contains(e.getEventId()))
                .forEach(result::add);

        if (!included.isEmpty()) {
            //sort again because included was merged to events
            result.sort(Comparator.comparingLong(QueueEventImpl::getEventId));
        }
        //save to context and return events
        PersistentQueueRunnerReport.loadedEvents(result);
        PersistentQueueRunnerReport.includedErrors(included.keySet());
        context.setExistingErrors(included);
        return result;
    }

    private void filterOutEventsBehaviours(List<QueueEventImpl> events, PersistentQueueRunningContext context) {
        //get the last event marker before starting to filter out.
        long lastMarker = events.get(events.size() - 1).getEventId();
        //apply applicative filter if exist
        QueueEventFilter filter = queueConfiguration.getFilter();
        List<QueueEventImpl> listAfterFilter = filter == null ? events : events.stream()
                .filter((filter::isRelevant))
                .collect(Collectors.toList());
        //apply all reduction behaviors
        log.debug("Events before {} and events after filter {}", events.size(), listAfterFilter.size());
        List<QueueEventImpl> result = QueueReductionBehavioursImpl.get(queueConfiguration.getReductionBehaviours())
                .applyChain(listAfterFilter);
        //report
        log.debug("Events before {} and events after filter chain {}", events.size(), result.size());
        PersistentQueueRunnerReport.addFilterEvents(events, result);
        //save state to context
        context.setEvents(result);
        context.setLastMarker(lastMarker);
    }

    @Nonnull
    private Collection<PersistentQueueGroupContext> createJobGroups(PersistentQueueRunningContext context) {
        QueueEventAggregator aggregator = queueConfiguration.getAggregator();
        LoadingCache<String, PersistentQueueGroupContext> groups = CacheBuilder.newBuilder().build(
                new CacheLoader<>() {
                    @Override
                    public PersistentQueueGroupContext load(@Nonnull String key) {
                        return new PersistentQueueGroupContext(key);
                    }
                });
        for (QueueEventImpl event : context.getEvents()) {
            Set<String> keys = new HashSet<>();
            if (aggregator != null) {
                aggregator.setKeys(event, keys);
            } else {
                //if no aggregator supplied only one group will be created with the repoKey name
                keys.add(repoKey);
            }
            for (String key : keys) {
                PersistentQueueGroupContext persistentQueueJob = groups.getUnchecked(key);
                //adding event to the relevant job
                persistentQueueJob.add(event);
            }
        }
        return groups.asMap().values();
    }


    /**
     * Invoke the contexts in multi threads executor
     * or in the same thread in case of one context
     *
     * Each contexts is set with the Future result.
     */
    private void spreadTheWork(Collection<PersistentQueueGroupContext> contexts) {
        QueueEventImplementor queueImplementor = queueConfiguration.getImplementor();
        if (contexts.size() == 1) {
            PersistentQueueGroupContext context = contexts.iterator().next();
            context.setFutureException(CompletableFuture.completedFuture(execute(queueImplementor, context)));
        } else {
            //for each job - start calculating all items , save all futures
            for (PersistentQueueGroupContext context : contexts) {
                context.setFutureException(jobsExecutor.submit(() -> execute(queueImplementor, context)));
            }
        }
    }

    /**
     * Execute context
     *
     * @return optional exception
     */
    @Nonnull
    private Optional<Exception> execute(QueueEventImplementor queueImplementor, PersistentQueueGroupContext context) {
        try {
            queueImplementor.execute(context.getEvents(), context.getUniqKey());
            return Optional.empty();
        } catch (Exception e) {
            //need to log the error ?
            return Optional.of(e);
        }
    }

    /**
     * sync all jobs, handle failures if exist, update the marker
     */
    @Nonnull
    private TriggerResult handleResult(PersistentQueueRunningContext context,
            Collection<PersistentQueueGroupContext> jobs) {

        Set<PersistentQueueRunningResult> failedResults = jobs.stream()
                //syncing on all failedResults here.
                .map(PersistentQueueGroupContext::getResult)
                //keep only failed result
                .filter(PersistentQueueRunningResult::isFailed)
                .collect(Collectors.toSet());

        context.setFailedResults(failedResults);
        persistentQueueErrorService.processErrors(this, context);

        //Filter out the exhausted errors from the failed result.
        Set<PersistentQueueRunningResult> finalFailedResults = failedResults.stream()
                .filter(failedResult -> !context.getExhaustedErrors().contains(failedResult.getFailedMarker()))
                .collect(Collectors.toSet());

        if (finalFailedResults.isEmpty()) {
            // No errors found we can update to the lastEventMarker - yeh !!
            // Update the marker only if was set (When we loaded events from the db) and > starting marker
            if (context.getLastMarker() > context.getStartingMarker()) {
                updateMarker(context.getLastMarker(), context.getCursorsLastUpdated());
            }
            return TriggerResult.FINISHED_ALL;
        }

        long smallestFailureMarkerFound = Long.MAX_VALUE;
        for (PersistentQueueRunningResult failedResult : finalFailedResults) {
            long failedMarker = failedResult.getFailedMarker();
            if (failedMarker == -1) {
                //in case marker was set to -1, it means we came across unhandled error (e.g NPE...)
                //in this case we do not advance the marker
                return TriggerResult.FINISHED_WITH_UNHANDLED_ERROR;
            }
            smallestFailureMarkerFound = Math.min(smallestFailureMarkerFound, failedMarker);
        }

        //calculate last success, using naive approach
        long lastSuccessfulMarker = smallestFailureMarkerFound;
        if (lastSuccessfulMarker > context.getStartingMarker()) {
            //update marker to the last successful job
            updateMarker(lastSuccessfulMarker, context.getCursorsLastUpdated());
        } else {
            log.debug("marker {} is the same as we started on queue {}", lastSuccessfulMarker, getName());
        }
        //in-case error support is disabled we will return an UNEXPECTED_ERROR instead of ERROR
        //this will prevent the retry loop to go into infinite loop cycles
        return queueConfiguration.isEnableErrorSupport() ?
                TriggerResult.FINISHED_WITH_HANDLED_ERROR :
                TriggerResult.FINISHED_WITH_UNHANDLED_ERROR;
    }

    private void updateMarker(long newMarker, long timestamp) {
        log.debug("updating the marker to {} on queue {}", newMarker, getName());
        PersistentQueueRunnerReport.endMarker(newMarker);
        persistentQueueCursorService.updateMarkerByTimestamp(cursorName, newMarker, timestamp);
    }

    @Override
    public void reset(String repoKey, boolean delete) {
        //removing existing errors for either delete or non delete
        persistentQueueErrorService.deleteErrors(this);
        if (delete) {
            persistentQueueCursorService.deleteMarker(cursorName);
        } else {
            persistentQueueCursorService.resetMarker(repoKey, cursorName);
        }
    }

    @Override
    public void markAsSync() {
        persistentQueueCursorService.updateCursorAsSynced(cursorName);
    }

    @Override
    public void shutdown() {
        //stop accepting new requests
        log.info("shutting down all running jobs for queue {}", cursorName);
        shutdown = true;
        jobsExecutor.shutdown();
        // No need to wait for termination
        // The worst case is that we will repeat some events on next run.
    }

    @Override
    public void pause() {
        paused = true;
        log.info("Queue paused, name: {}, repo: {}", getName(), repoKey);
    }

    @Override
    public void proceed() {
        if (shutdown) {
            log.warn("Should not call proceed on a queue {} which was shutdown ", cursorName);
        }
        if (paused) {
            triggeringService.triggerQueues(repoKey, queueConfiguration.getCursorType());
            paused = false;
            log.info("Queue resumed, name: {}, repo: {}", getName(), repoKey);
        }
    }

    @Override
    public void reset() {
        reset(this.repoKey, false);
    }

    @Override
    public boolean isPaused() {
        return paused;
    }

    @Override
    public int getEventQueueSize() {
        long marker = getCurrentMarkerCursorForQueue().getEventMarker();
        String fileSuffix = queueConfiguration.getFileSuffix();
        return eventsService.countEventsAfter(marker, repoKey, fileSuffix);
    }

    @Override
    public QueueStatus getQueueStatus() {
        int overflowLimit = queueConfiguration.getErrorOverflowLimit();
        if (overflowLimit > 0) {
            int errorCount = persistentQueueErrorService.getErrorCount(this);
            if (errorCount >= overflowLimit) {
                return QueueStatus.ERROR_OVERFLOW;
            }
        }
        return persistentQueueCursorService.getStatus(repoKey, cursorName);
    }

    @Override
    public void resetStatus() {
        markAsSync();
    }

    private void markOverFlowLimitErrorsIfNeeded(@Nonnull List<EventAndErrorInfo> error) {
        if (error.size() > queueConfiguration.getErrorOverflowLimit()
                && queueConfiguration.getErrorOverflowLimit() > 0) {
            persistentQueueCursorService.updateCursorAsOutOfSync(cursorName);
        }
    }
}
