package org.artifactory.servicetrust.cold;

import org.apache.commons.lang.StringUtils;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.retention.ArtifactRetentionAddon;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.security.access.GenericTokenSpec;
import org.artifactory.api.security.access.TokenSpec;
import org.artifactory.descriptor.authentication.AuthenticationToken;
import org.artifactory.servicetrust.ExchangeTokenRequest;
import org.artifactory.servicetrust.ExchangeTokenResponse;
import org.artifactory.servicetrust.ServiceTrust;
import org.artifactory.servicetrust.ServiceTrustException;
import org.artifactory.servicetrust.ServiceTrustUseCase;
import org.artifactory.servicetrust.TokenExtension;
import org.artifactory.servicetrust.TokenSpecInboundService;
import org.artifactory.servicetrust.TokenSpecOutboundService;
import org.artifactory.servicetrust.TokenSpecServiceBaseImpl;
import org.artifactory.util.CollectionUtils;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.jfrog.access.common.ServiceId;
import org.jfrog.access.token.JwtAccessToken;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Optional;
import java.util.stream.Stream;

import static org.artifactory.addon.retention.ArtifactRetentionConstants.INSTANCES_ALREADY_PAIRED;
import static org.artifactory.servicetrust.ServiceTrustService.INTERNAL_SCOPE_PREFIX;

/**
 * @author volodymyrm
 */
public class ArtifactoryColdTokenSpecService extends TokenSpecServiceBaseImpl
        implements TokenSpecOutboundService, TokenSpecInboundService {

    private static final String LIMITED_SCOPE = "scoped";
    public static final ArtifactoryColdTokenSpecService ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE =
            new ArtifactoryColdTokenSpecService();

    @Override
    public TokenSpec<GenericTokenSpec> createPairingTokenSpec(
            @NotNull ServiceId artifactoryServiceId,
            @NotNull ServiceTrust trust,
            @NotNull String subject,
            @NotNull String scope,
            @NotNull String baseUrl,
            @Nullable String pairingUrl,
            @NotNull String exchangeUrl) {

        if (trust.equals(ServiceTrust.TWO_WAY)) {
            throw new IllegalArgumentException("artifactory-cold does not support two-way pairing");
        }
        verifyIsColdStorageEnabled();

        return super.createPairingTokenSpec(artifactoryServiceId, trust, subject, scope,
                baseUrl, pairingUrl, exchangeUrl);
    }

    @NotNull
    @Override
    protected String getUseCasePathForScope(@NotNull String scope) {
        String useCasePath = getUseCaseFromPairScope(Collections.singletonList(scope));
        if (ServiceTrustUseCase.ARTIFACTORY_COLD.matches(useCasePath)) {
            useCasePath = generateIceUseCaseWithNamespace(useCasePath);
        }
        return useCasePath;
    }

    @Override
    public TokenSpec<GenericTokenSpec> createMasterTokenSpec(@NotNull ServiceId artifactoryServiceId,
            @NotNull String useCasePath, @NotNull String pairingScope, @NotNull String baseUrl,
            @NotNull String pairingUrl, @NotNull String exchangeUrl) {

        verifyIsColdStorageEnabled();

        return super.createMasterTokenSpec(artifactoryServiceId, useCasePath, pairingScope,
                baseUrl, pairingUrl, exchangeUrl);
    }

    @Override
    public TokenSpec<GenericTokenSpec> createExchangeTokenSpec(ServiceId artifactoryServiceId, String useCasePath) {
        verifyIsColdStorageEnabled();

        return super.createExchangeTokenSpec(artifactoryServiceId, useCasePath);
    }

    @Override
    public ExchangeTokenRequest createExchangeTokenRequest(String subject, JwtAccessToken jwt) {
        verifyIsWarmStorageEnabled();
        String updatedUseCasePath = getUseCaseFromPairScope(jwt.getScope());
        return new ExchangeTokenRequest(subject, updatedUseCasePath);
    }

    @Override
    public void storeMasterTokenData(ExchangeTokenResponse response, TokenExtension extension) {
        verifyIsWarmStorageEnabled();
        if (!getArtifactRetentionAddon().getColdStoragePairAuthenticationTokens().isEmpty()) {
            throw new ArtifactRetentionBindException(INSTANCES_ALREADY_PAIRED);
        }
        if (StringUtils.isEmpty(response.getScope())) {
            throw new IllegalArgumentException("Master token does not have scope");
        }
        String[] tokens = response.getScope().split(" ");
        String tokenWithNamespace = findFirstRetentionToken(Arrays.stream(tokens));
        if (tokenWithNamespace == null) {
            throw new IllegalArgumentException("Master token scope with the 'artifactory-cold' scope does not exist");
        }
        String namespace = getSimpleScope(tokenWithNamespace)
                .replace(ServiceTrustUseCase.ARTIFACTORY_COLD.getName() + "/", "")
                .replace(ServiceTrustUseCase.ARTIFACTORY_COLD.getName(), "")
                .replace(INTERNAL_SCOPE_PREFIX, "");

        if (StringUtils.isEmpty(namespace)) {
            throw new IllegalArgumentException("Can not parse the master token scope to get a cold namespace: " +
                    "namespace is empty");
        }
        getArtifactRetentionAddon().saveNamespaceAndMasterToken(extension.getBaseUrl(), extension.getExchangeUrl(),
                namespace, response.getTokenId(), response.getAccessToken(), response.getTokenType(),
                response.getExpiresIn(), response.getScope());
    }

    @Override
    public void updateMasterTokenData(ExchangeTokenResponse response) {
        verifyIsWarmStorageEnabled();
        getArtifactRetentionAddon().updateMasterToken(response.getTokenId(), response.getAccessToken(),
                response.getTokenType(), response.getExpiresIn(), response.getScope());
    }

    @Override
    public ExchangeTokenRequest createMasterTokenExchangeRefreshRequest(@NotNull String subject) {
        verifyIsWarmStorageEnabled();
        String namespace = getArtifactRetentionAddon().getJfrogColdStorageArchivingNamespace();
        if (StringUtils.isEmpty(namespace)) {
            throw new IllegalStateException("Can not create exchange request to refresh master token: " +
                    "namespace does not exist");
        }
        return new ExchangeTokenRequest(subject,
                String.join("/", ServiceTrustUseCase.ARTIFACTORY_COLD.getName(), namespace));
    }

    @Override
    public ExchangeTokenRequest createScopedTokenExchangeRequest(@NotNull String subject, String... scopes) {
        verifyIsWarmStorageEnabled();
        String namespace = getArtifactRetentionAddon().getJfrogColdStorageArchivingNamespace();
        if (StringUtils.isEmpty(namespace)) {
            throw new ServiceTrustException("Can not create exchange request to refresh master token: " +
                    "namespace does not exist");
        }
        String target =
                String.format("%s/%s:%s", ServiceTrustUseCase.ARTIFACTORY_COLD.getName(), namespace, LIMITED_SCOPE);
        return new ExchangeTokenRequest(subject, target);
    }

    @Override
    public Optional<AuthenticationToken> getAuthenticationToken(String baseUrl) {
        verifyIsWarmStorageEnabled();
        return getArtifactRetentionAddon().getJfrogColdStorageAuthenticationToken(baseUrl);
    }

    @Override
    public Collection<AuthenticationToken> getPairedInstances() {
        return getArtifactRetentionAddon().getColdStoragePairAuthenticationTokens();
    }

    @Override
    public void deleteAuthenticationToken(String baseUrl) {
        verifyIsWarmStorageEnabled();
        getArtifactRetentionAddon().deleteJfrogColdStorageMasterToken(baseUrl);
    }

    /**
     * Extract a use case with namespace 'artifactory-cold/namespace' from scopes
     */
    @Override
    @NotNull
    protected String getUseCaseFromPairScope(List<String> scope) {
        if (CollectionUtils.isNullOrEmpty(scope)) {
            throw new IllegalArgumentException("Pair's token scope can not be null");
        }
        String retentionScope = findFirstRetentionToken(scope.stream());
        if (retentionScope == null) {
            throw new IllegalArgumentException("Pair's token does not have retention related scope");
        }
        return getSimpleScope(retentionScope);
    }

    private String findFirstRetentionToken(@NotNull Stream<String> tokens) {
        return tokens
                .filter(t -> t.contains(ServiceTrustUseCase.ARTIFACTORY_COLD.getName()))
                .findFirst()
                .orElse(null);
    }

    private void verifyIsColdStorageEnabled() {
        ArtifactRetentionAddon artifactRetentionAddon = getArtifactRetentionAddon();
        if (!artifactRetentionAddon.isColdInstance()) {
            throw new IllegalStateException("Cold Storage mode is disabled");
        }
    }

    private void verifyIsWarmStorageEnabled() {
        ArtifactRetentionAddon artifactRetentionAddon = getArtifactRetentionAddon();
        if (artifactRetentionAddon.isColdInstance()) {
            throw new IllegalStateException("Instance running in the Cold Storage mode");
        }
    }

    private ArtifactRetentionAddon getArtifactRetentionAddon() {
        return ContextHelper.get().beanForType(AddonsManager.class)
                .addonByType(ArtifactRetentionAddon.class);
    }

    /**
     * Generates use case with namespace for the retention feature. It should look like 'artifactory-cold/namespace'
     */
    @NotNull
    private String generateIceUseCaseWithNamespace(@NotNull String useCaseWithoutNamespace) {
        ArtifactRetentionAddon artifactRetentionAddon = getArtifactRetentionAddon();
        String namespace = artifactRetentionAddon.generateJFrogColdStorageNamespace();
        return String.join("/", useCaseWithoutNamespace, namespace);
    }
}