package org.artifactory.ph;

import org.apache.commons.io.IOUtils;
import org.apache.http.HttpStatus;
import org.artifactory.api.request.ArtifactoryHeaders;
import org.artifactory.api.request.ArtifactoryResponse;
import org.artifactory.api.request.DownloadService;
import org.artifactory.api.request.InternalArtifactoryRequest;
import org.artifactory.common.resource.BaseResource;
import org.artifactory.core.api.ph.exception.PackageException;
import org.artifactory.core.api.ph.model.PackageArtifact;
import org.artifactory.core.api.ph.model.PackageDownloadContext;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RemoteRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.descriptor.repo.VirtualRepoDescriptor;
import org.artifactory.fs.FileInfo;
import org.artifactory.fs.RepoResource;
import org.artifactory.model.xstream.fs.PropertiesImpl;
import org.artifactory.repo.LocalCacheRepo;
import org.artifactory.repo.LocalRepo;
import org.artifactory.repo.RemoteRepo;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.repo.virtual.VirtualRepo;
import org.artifactory.request.InternalRequestContext;
import org.artifactory.resource.RemoteRepoResource;
import org.artifactory.rest.exception.RestException;
import org.artifactory.sapi.fs.VfsFile;
import org.artifactory.util.StringInputStream;
import org.jetbrains.annotations.NotNull;
import org.junit.jupiter.api.Assertions;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import javax.ws.rs.core.Response;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.util.List;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.doThrow;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.mockito.MockitoAnnotations.openMocks;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertTrue;
import static org.testng.Assert.fail;

@Test
public class PackageArtifactServiceImplTest {

    @Mock
    private DownloadService downloadService;
    @Mock
    private InternalRepositoryService repositoryService;
    @Mock
    private VirtualRepo virtualRepo;

    private PackageArtifactServiceImpl packageDownloadService;

    @BeforeMethod
    public void setUp() {
        openMocks(this);
        packageDownloadService = new PackageArtifactServiceImpl(downloadService, repositoryService);
    }

    @AfterMethod
    public void tearDown() {
    }

    @Test
    public void testDownload() throws PackageException, IOException {
        PackageDownloadContext context = new PackageDownloadContext("repoKey", "path");
        context.header("key", "val");
        Response response = packageDownloadService.download(context);

        ArgumentCaptor<InternalArtifactoryRequest> captor = ArgumentCaptor.forClass(InternalArtifactoryRequest.class);
        verify(downloadService).process(captor.capture(), any(ArtifactoryResponse.class));
        final List<InternalArtifactoryRequest> params = captor.getAllValues();
        Assertions.assertEquals(200, response.getStatus());
        Assertions.assertEquals("val", params.get(0).getHeaderValues("key").nextElement());
    }

    @Test
    public void testDownloadFormRemoteCache() throws PackageException, IOException {
        LocalCacheRepo localRepo = Mockito.mock(LocalCacheRepo.class);
        when(localRepo.getKey()).thenReturn("repoKey-cache");
        when(repositoryService.localFederatedOrCachedRepositoryByKey("repoKey")).thenReturn(localRepo);
        PackageDownloadContext context = new PackageDownloadContext("repoKey", "path");
        context.header("key", "val");
        Response response = packageDownloadService.downloadFromRemoteCache(context);

        ArgumentCaptor<InternalArtifactoryRequest> captor = ArgumentCaptor.forClass(InternalArtifactoryRequest.class);
        verify(downloadService).process(captor.capture(), any(ArtifactoryResponse.class));
        final List<InternalArtifactoryRequest> params = captor.getAllValues();
        Assertions.assertEquals(200, response.getStatus());
        Assertions.assertEquals("val", params.get(0).getHeaderValues("key").nextElement());
    }

    @Test
    public void testDownloadForbiddenRestException() throws IOException {
        PackageDownloadContext context = new PackageDownloadContext("repoKey", "path");
        context.header("key", "val");
        doThrow(new RestException(HttpStatus.SC_FORBIDDEN, "Forbidden")).when(downloadService).process(any(), any());
        PackageException exception =
                Assertions.assertThrows(PackageException.class, () -> packageDownloadService.download(context));
        Assertions.assertEquals(403, exception.getStatus());
        Assertions.assertEquals("Forbidden", exception.getInternalReason());
        Assertions.assertEquals("Forbidden", exception.getPublicReason());
    }

    @Test
    public void testDownloadFromRemoteCacheRepoNotExist() {
        PackageDownloadContext context = new PackageDownloadContext("repoKey", "path");
        try {
            packageDownloadService.downloadFromRemoteCache(context);
            fail("Expected exception");
        } catch (PackageException e) {
            assertEquals(e.getStatus(), HttpStatus.SC_BAD_REQUEST);
            assertEquals(e.getInternalReason(),
                    "Repo 'repoKey' does not exist or storing artifacts locally is disabled");
        }
    }

    @Test
    public void testDownloadFrommRemoteCacheBadInput() {
        PackageDownloadContext context = new PackageDownloadContext("repoKey-cache", "path");
        try {
            packageDownloadService.downloadFromRemoteCache(context);
            fail("Expected exception");
        } catch (PackageException e) {
            assertEquals(e.getStatus(), HttpStatus.SC_BAD_REQUEST);
            assertEquals(e.getInternalReason(),
                    "Download request from remote-cache repo should not include '-cache' in the repo key");
        }
    }

    @Test
    public void testGetArtifactFromVirtual() throws PackageException {
        String virtualRepoKey = "virtual";
        String localRepoKey = "local";
        String path = "path";
        RepoPath virtualRepoPath = RepoPathFactory.create(virtualRepoKey, path);
        RepoPath localRepoPath = RepoPathFactory.create(localRepoKey, path);
        @SuppressWarnings("unchecked")
        LocalRepo<LocalRepoDescriptor> localRepo = Mockito.mock(LocalRepo.class);
        VirtualRepo virtualRepo = Mockito.mock(VirtualRepo.class);
        Mockito.when(virtualRepo.getKey()).thenReturn(virtualRepoKey);
        RepoResource repoResource = info(localRepoKey, path);
        Mockito.when(virtualRepo.getInfo(any())).thenReturn(repoResource);
        FileInfo fileInfo = Mockito.mock(FileInfo.class);
        VfsFile<?> file = Mockito.mock(VfsFile.class);
        Mockito.when(fileInfo.getRepoKey()).thenReturn(localRepoKey);
        Mockito.when(fileInfo.getRepoPath()).thenReturn(virtualRepoPath);
        Mockito.when(fileInfo.getRelPath()).thenReturn(path);
        Mockito.when(file.getRepoKey()).thenReturn(localRepoKey);
        Mockito.when(file.getRepoPath()).thenReturn(localRepoPath);
        Mockito.when(file.getPath()).thenReturn(path);
        Mockito.when(file.getProperties()).thenReturn(new PropertiesImpl());
        Mockito.when(localRepo.getKey()).thenReturn(localRepoKey);
        Mockito.when(localRepo.getImmutableFile(localRepoPath)).thenReturn(file);
        LocalRepoDescriptor localDescriptor = Mockito.mock(LocalRepoDescriptor.class);
        VirtualRepoDescriptor virtualDescriptor = Mockito.mock(VirtualRepoDescriptor.class);
        Mockito.when(localDescriptor.getType()).thenReturn(RepoType.CRAN);
        Mockito.when(virtualDescriptor.getType()).thenReturn(RepoType.CRAN);
        Mockito.when(localRepo.getDescriptor()).thenReturn(localDescriptor);
        Mockito.when(virtualRepo.getDescriptor()).thenReturn(virtualDescriptor);
        Mockito.when(repositoryService.localFederatedOrCachedRepositoryByKey(localRepoKey)).thenReturn(localRepo);
        Mockito.when(repositoryService.repositoryByKey(localRepoKey)).thenReturn(localRepo);
        Mockito.when(repositoryService.repositoryByKey(virtualRepoKey)).thenReturn(virtualRepo);
        Mockito.when(repositoryService.virtualRepositoryByKey(virtualRepoKey)).thenReturn(virtualRepo);
        Mockito.when(repositoryService.getVirtualFileInfo(virtualRepoPath)).thenReturn(fileInfo);
        PackageArtifact artifact = packageDownloadService.getArtifact(virtualRepoKey, path);
        Assertions.assertEquals(localRepoKey, artifact.getRepoKey());
        Assertions.assertEquals(path, artifact.getPath());
    }

    @Test
    public void testGetArtifactFromLocal() throws PackageException, IOException {
        String repoKey = "repoKey";
        String path = "path";
        RepoPath repoPath = RepoPathFactory.create(repoKey, path);
        LocalRepo<LocalRepoDescriptor> localRepo = Mockito.mock(LocalRepo.class);
        VfsFile<?> file = Mockito.mock(VfsFile.class);
        Mockito.when(file.getStream()).thenReturn(new StringInputStream("Test"));
        Mockito.when(localRepo.getImmutableFile(repoPath)).thenReturn(file);
        Mockito.when(localRepo.isReal()).thenReturn(true);
        Mockito.when(localRepo.getKey()).thenReturn(repoKey);
        Mockito.when(file.getRepoKey()).thenReturn(repoKey);
        Mockito.when(file.getRepoPath()).thenReturn(repoPath);
        Mockito.when(file.getPath()).thenReturn(path);
        Mockito.when(file.getProperties()).thenReturn(new PropertiesImpl());
        Mockito.when(localRepo.getImmutableFile(repoPath)).thenReturn(file);
        LocalRepoDescriptor descriptor = Mockito.mock(LocalRepoDescriptor.class);
        Mockito.when(descriptor.getType()).thenReturn(RepoType.CRAN);
        Mockito.when(localRepo.getDescriptor()).thenReturn(descriptor);
        RepoResource info = info(repoKey, path);
        Mockito.when(localRepo.getInfo(any())).thenReturn(info);
        Mockito.when(repositoryService.repositoryByKey(repoKey)).thenReturn(localRepo);
        Mockito.when(repositoryService.localFederatedOrCachedRepositoryByKey(repoKey)).thenReturn(localRepo);
        Mockito.when(repositoryService.localOrFederatedRepositoryByKey(repoKey)).thenReturn(localRepo);
        PackageArtifact artifact = packageDownloadService.getArtifact(repoKey, path);
        Assertions.assertEquals(repoKey, artifact.getRepoKey());
        Assertions.assertEquals(path, artifact.getPath());
    }

    @NotNull
    private RepoResource info(String repoKey, String path) {
        RepoResource info = Mockito.mock(RepoResource.class);
        Mockito.when(info.getRepoKey()).thenReturn(repoKey);
        Mockito.when(info.getPath()).thenReturn(path);
        Mockito.when(info.isExpired()).thenReturn(false);
        Mockito.when(info.isFound()).thenReturn(true);
        Mockito.when(info.isBlocked()).thenReturn(false);
        Mockito.when(info.getRepoPath()).thenReturn(RepoPathFactory.create(repoKey, path));
        return info;
    }

    @NotNull
    private RepoResource remoteInfo(String repoKey, String path) {
        RemoteRepoResource info = Mockito.mock(RemoteRepoResource.class);
        Mockito.when(info.getRepoKey()).thenReturn(repoKey);
        Mockito.when(info.getPath()).thenReturn(path);
        Mockito.when(info.isExpired()).thenReturn(false);
        Mockito.when(info.isFound()).thenReturn(true);
        Mockito.when(info.isBlocked()).thenReturn(false);
        Mockito.when(info.getRepoPath()).thenReturn(RepoPathFactory.create(repoKey, path));
        return info;
    }


    @Test
    public void testGetArtifactFromRemote() throws PackageException, IOException {
        String remoteRepoKey = "remoteRepoKey";
        String localRepoKey = "localRepoKey";
        String path = "path";
        RepoPath remoteRepoPath = RepoPathFactory.create(remoteRepoKey, path);
        RepoPath localRepoPath = RepoPathFactory.create(localRepoKey, path);
        RemoteRepo<RemoteRepoDescriptor> remoteRepo = Mockito.mock(RemoteRepo.class);
        LocalRepo<LocalRepoDescriptor> localRepo = Mockito.mock(LocalRepo.class);
        VfsFile<?> file = Mockito.mock(VfsFile.class);
        Mockito.when(file.getStream()).thenReturn(new StringInputStream("Test"));
        Mockito.when(localRepo.getImmutableFile(localRepoPath)).thenReturn(file);
        LocalRepoDescriptor localRepoDescriptor = Mockito.mock(LocalRepoDescriptor.class);
        Mockito.when(localRepoDescriptor.getType()).thenReturn(RepoType.CRAN);
        Mockito.when(localRepo.getDescriptor()).thenReturn(localRepoDescriptor);
        Mockito.when(localRepo.getKey()).thenReturn(localRepoKey);
        Mockito.when(remoteRepo.isReal()).thenReturn(true);
        Mockito.when(remoteRepo.getKey()).thenReturn(remoteRepoKey);
        Mockito.when(file.getRepoKey()).thenReturn(remoteRepoKey);
        Mockito.when(file.getRepoPath()).thenReturn(remoteRepoPath);
        Mockito.when(file.getPath()).thenReturn(path);
        Mockito.when(file.getProperties()).thenReturn(new PropertiesImpl());
        RemoteRepoDescriptor descriptor = Mockito.mock(RemoteRepoDescriptor.class);
        Mockito.when(descriptor.getType()).thenReturn(RepoType.CRAN);
        Mockito.when(remoteRepo.getDescriptor()).thenReturn(descriptor);
        RepoResource remote = remoteInfo(remoteRepoKey, path);
        RepoResource local = info(remoteRepoKey, path);
        Mockito.when(remoteRepo.getInfo(any())).thenReturn(remote);
        Mockito.when(localRepo.getInfo(any())).thenReturn(local);
        Mockito.when(repositoryService.repositoryByKey(remoteRepoKey)).thenReturn(remoteRepo);
        Mockito.when(repositoryService.remoteRepositoryByKey(remoteRepoKey)).thenReturn(remoteRepo);
        Mockito.when(repositoryService.repositoryByKey(remoteRepoKey)).thenReturn(remoteRepo);
        Mockito.when(repositoryService.repositoryByKey(localRepoKey)).thenReturn(localRepo);
        Mockito.when(repositoryService.localFederatedOrCachedRepositoryByKey(localRepoKey)).thenReturn(localRepo);
        Mockito.when(repositoryService.localFederatedOrCachedRepositoryByKey(remoteRepoKey)).thenReturn(localRepo);
        PackageArtifact artifact = packageDownloadService.getArtifact(remoteRepoKey, path);
        Assertions.assertEquals(remoteRepoKey, artifact.getRepoKey());
        Assertions.assertEquals(path, artifact.getPath());
    }

    @Test
    public void testGetStreamFromLocal() throws PackageException, IOException {
        String repoKey = "repoKey";
        String path = "path";
        RepoPath repoPath = RepoPathFactory.create(repoKey, path);
        LocalRepo<?> localRepo = Mockito.mock(LocalRepo.class);
        VfsFile<?> file = Mockito.mock(VfsFile.class);
        Mockito.when(file.getStream()).thenReturn(new StringInputStream("Test"));
        Mockito.when(localRepo.getImmutableFile(repoPath)).thenReturn(file);
        Mockito.when(localRepo.isReal()).thenReturn(true);
        Mockito.when(localRepo.getKey()).thenReturn(repoKey);
        RepoResource info = info(repoKey, path);
        Mockito.when(localRepo.getInfo(any())).thenReturn(info);
        Mockito.when(localRepo.getKey()).thenReturn(repoKey);
        Mockito.when(repositoryService.repositoryByKey(repoKey)).thenReturn(localRepo);
        Mockito.when(repositoryService.localFederatedOrCachedRepositoryByKey(repoKey)).thenReturn(localRepo);
        Mockito.when(repositoryService.localOrFederatedRepositoryByKey(repoKey)).thenReturn(localRepo);
        PackageDownloadContext context = new PackageDownloadContext(repoKey, path, "alter");
        context.header("key", "val");
        InputStream stream = packageDownloadService.getStream(context);
        Assertions.assertEquals("Test", IOUtils.toString(stream, StandardCharsets.UTF_8));

        ArgumentCaptor<InternalRequestContext> captor = ArgumentCaptor.forClass(InternalRequestContext.class);
        verify(localRepo).getInfo(captor.capture());
        final List<InternalRequestContext> params = captor.getAllValues();
        Assertions.assertEquals("val", params.get(0).getRequest().getHeaderValues("key").nextElement());
        InternalArtifactoryRequest originRequest = (InternalArtifactoryRequest) params.get(0).getRequest();
        Assertions.assertEquals("alter",
                originRequest.getParameter(ArtifactoryHeaders.PARAM_ALTERNATIVE_REMOTE_DOWNLOAD_URL));
    }

    @Test
    public void testGetResource() throws IOException, PackageException {
        String repoKey = "repoKey";
        String path = "path";
        RepoPath repoPath = RepoPathFactory.create(repoKey, path);
        LocalRepo<?> localRepo = Mockito.mock(LocalRepo.class);
        VfsFile<?> file = Mockito.mock(VfsFile.class);
        Mockito.when(file.getStream()).thenReturn(new StringInputStream("Test"));
        RepoResource resource = Mockito.mock(RepoResource.class);
        Mockito.when(resource.getRepoPath()).thenReturn(repoPath);
        Mockito.when(resource.getRepoKey()).thenReturn(repoKey);
        Mockito.when(resource.getPath()).thenReturn(path);
        Mockito.when(localRepo.getInfo(any())).thenReturn(resource);
        Mockito.when(localRepo.isReal()).thenReturn(true);
        Mockito.when(repositoryService.repositoryByKey(repoKey)).thenReturn(localRepo);
        BaseResource result = packageDownloadService.getResource(repoKey, path);
        Assertions.assertEquals(repoKey, result.getRepoKey());
        Assertions.assertEquals(path, result.getPath());
    }

    @Test
    public void testGetResourceFromVirtualCacheRepoNotFound() {
        try {
            packageDownloadService.getResourceFromVirtualCache("repo1", "a/b/c.json");
            fail("Should get bad request");
        } catch (PackageException e) {
            assertEquals(e.getStatus(), HttpStatus.SC_NOT_FOUND);
        }
    }

    @Test
    public void testGetExpiredResourceFromVirtualCache() throws PackageException {
        when(repositoryService.virtualRepositoryByKey(Mockito.eq("repo1"))).thenReturn(virtualRepo);
        RepoResource resource = Mockito.mock(RepoResource.class);
        when(resource.isFound()).thenReturn(true);
        when(virtualRepo.getInfoFromVirtualCache(any())).thenReturn(resource);
        when(virtualRepo.isCachedResourceExpired(any())).thenReturn(true);
        BaseResource cachedResource = packageDownloadService.getResourceFromVirtualCache("repo1",
                "a/b/c.json");
        assertTrue(cachedResource.isExpired(), "Resource should be expired");

    }

    @Test
    public void testGetResourceFromVirtualCache() throws PackageException {
        when(repositoryService.virtualRepositoryByKey(Mockito.eq("repo1"))).thenReturn(virtualRepo);
        RepoResource resource = Mockito.mock(RepoResource.class);
        when(virtualRepo.getInfoFromVirtualCache(any())).thenReturn(resource);
        BaseResource cachedResource = packageDownloadService.getResourceFromVirtualCache("repo1",
                "a/b/c.json");
        assertNotNull(cachedResource);
    }
}