package org.artifactory.ph;

import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.authentication.ArtifactoryBasicAuthenticationFilter;
import org.artifactory.authentication.Context.AuthenticationContext;
import org.artifactory.core.api.ph.model.PackageAuthenticationRequestContext;
import org.artifactory.core.api.ph.service.PackageAuthenticationAdapter;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.mockito.Mockito;

import javax.annotation.Nullable;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.ws.rs.core.Response;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.stream.Stream;

import static javax.servlet.http.HttpServletResponse.SC_UNAUTHORIZED;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyInt;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

/**
 * @author shahafg
 */
class PackageAuthenticationFilterImplTest {

    private static final String BASIC_AUTH_VALUE = "Basic YWRtaW46cGFzc3dvcmQ=";
    private static final String PRINCIPLE_VALUE = "admin";

    private PackageAuthenticationAdapter packageAuthenticationAdapter;
    private PackageAuthenticationFilterImpl packageFilter;

    @BeforeEach
    public void setUp() {
        this.packageAuthenticationAdapter = Mockito.mock(PackageAuthenticationAdapter.class);
        this.packageFilter = new PackageAuthenticationFilterImpl(packageAuthenticationAdapter);
    }

    @AfterEach
    public void tearDown() {
        this.packageFilter = null;
        this.packageAuthenticationAdapter = null;
    }

    @Test
    void init() throws ServletException {
        FilterConfig filterConfig = Mockito.mock(FilterConfig.class);
        ServletContext servletContext = Mockito.mock(ServletContext.class);
        ArtifactoryContext artifactoryContext = Mockito.mock(ArtifactoryContext.class);
        when(filterConfig.getServletContext()).thenReturn(servletContext);
        when(servletContext.getAttribute(ArtifactoryContext.APPLICATION_CONTEXT_KEY)).thenReturn(artifactoryContext);
        when(artifactoryContext.beanForType(ArtifactoryBasicAuthenticationFilter.class))
                .thenReturn(Mockito.mock(ArtifactoryBasicAuthenticationFilter.class));
        packageFilter.init(filterConfig);
        verify(artifactoryContext, times(1)).beanForType(ArtifactoryBasicAuthenticationFilter.class);
    }

    @ParameterizedTest
    @MethodSource("authenticationFlagsProvider")
    void getAuthenticationContextAuthenticationAllowed(boolean isAuthenticated, boolean isAnonymousAllowed)
            throws ServletException, IOException {
        initFilter(new ArtifactoryBasicAuthenticationFilter());
        HttpServletRequest request = Mockito.mock(HttpServletRequest.class);
        PackageAuthenticationRequestContext context = getRequestContext(isAuthenticated, isAnonymousAllowed);
        AuthenticationContext expected = buildFilterContext(isAuthenticated, isAnonymousAllowed);
        when(packageAuthenticationAdapter.getContext(request)).thenReturn(context);
        if (expected != null) {
            assertArtifactoryFilterContext(expected,
                    packageFilter.getAuthenticationContext(request, Mockito.mock(HttpServletResponse.class)));
        } else {
            HttpServletResponse response = Mockito.mock(HttpServletResponse.class);
            PrintWriter writer = Mockito.mock(PrintWriter.class);
            doNothing().when(response).setStatus(anyInt());
            doNothing().when(response).setHeader(any(), any());
            when(response.getWriter()).thenReturn(writer);
            doNothing().when(writer).write(anyString());
            packageFilter.getAuthenticationContext(request, response);
            verify(response, times(1)).getWriter();
            verify(writer, times(1)).write(anyString());
            verify(response, times(1)).setStatus(anyInt());
            verify(response, times(1)).addHeader(any(), any());
        }
    }

    @Test
    void doFilter() throws ServletException, IOException {
        HttpServletRequest request = Mockito.mock(HttpServletRequest.class);
        ArtifactoryBasicAuthenticationFilter basicFilter = Mockito.mock(ArtifactoryBasicAuthenticationFilter.class);
        doNothing().when(basicFilter).doFilter(any(), any(), any());
        initFilter(basicFilter);
        packageFilter.doFilter(request, Mockito.mock(HttpServletResponse.class), Mockito.mock(FilterChain.class));
        verify(basicFilter, times(1)).doFilter(any(), any(), any());
    }

    private void assertArtifactoryFilterContext(AuthenticationContext expected,
            AuthenticationContext actual) {
        Assertions.assertEquals(expected.isAuthenticated(), actual.isAuthenticated());
        Assertions.assertEquals(expected.isAnonymousAllowed(), actual.isAnonymousAllowed());
        Assertions.assertEquals(expected.getExtractedAuthorizationToken(), actual.getExtractedAuthorizationToken());
        Assertions.assertEquals(expected.getLoginIdentifier(), actual.getLoginIdentifier());
        Assertions.assertEquals(expected.getCacheKey(), actual.getCacheKey());
    }

    private void initFilter(ArtifactoryBasicAuthenticationFilter basicFilter) throws ServletException {
        FilterConfig filterConfig = Mockito.mock(FilterConfig.class);
        ServletContext servletContext = Mockito.mock(ServletContext.class);
        ArtifactoryContext artifactoryContext = Mockito.mock(ArtifactoryContext.class);
        when(filterConfig.getServletContext()).thenReturn(servletContext);
        when(servletContext.getAttribute(ArtifactoryContext.APPLICATION_CONTEXT_KEY)).thenReturn(artifactoryContext);
        when(artifactoryContext.beanForType(ArtifactoryBasicAuthenticationFilter.class)).thenReturn(basicFilter);
        packageFilter.init(filterConfig);
    }

    private PackageAuthenticationRequestContext getRequestContext(boolean isAuthenticated, boolean isAnonymousAllowed) {
        if (isAuthenticated && !isAnonymousAllowed) {
            return PackageAuthenticationRequestContext.builder().authenticationAllowed(true)
                    .anonymousRequestAllowed(false)
                    .authorizationToken(BASIC_AUTH_VALUE).principleUserName(PRINCIPLE_VALUE).build();
        }

        if (!isAuthenticated && isAnonymousAllowed) {
            return PackageAuthenticationRequestContext.builder().anonymousRequestAllowed(true).build();
        }
        return PackageAuthenticationRequestContext.builder().responseError(buildUnAuthorizedResponse()).build();
    }

    @Nullable
    private AuthenticationContext buildFilterContext(boolean isAuthenticated,
            boolean isAnonymousAllowed) {
        if (isAuthenticated && !isAnonymousAllowed) {
            return AuthenticationContext.builder().authenticated(true)
                    .extractedAuthorizationToken(BASIC_AUTH_VALUE)
                    .loginIdentifier(PRINCIPLE_VALUE).cacheKey(BASIC_AUTH_VALUE).build();
        }
        if (!isAuthenticated && isAnonymousAllowed) {
            return AuthenticationContext.builder().anonymousAllowed(true).build();
        }
        return null;
    }

    private static Stream<Arguments> authenticationFlagsProvider() {
        return Stream.of(
                Arguments.of(true, false),
                Arguments.of(false, true),
                Arguments.of(false, false));
    }

    private static Response buildUnAuthorizedResponse() {
        return Response.status(SC_UNAUTHORIZED).entity(new Object())
                .header("WWW-Authenticate", "Basic realm=\"Artifactory Realm\"").build();
    }
}