package org.artifactory.ph;

import org.artifactory.api.security.AuthorizationService;
import org.artifactory.aql.AqlService;
import org.artifactory.aql.api.internal.AqlBase;
import org.artifactory.aql.model.AqlItemTypeEnum;
import org.artifactory.aql.result.AqlEagerResult;
import org.artifactory.aql.result.AqlLazyResult;
import org.artifactory.aql.result.rows.AqlItem;
import org.artifactory.core.api.ph.exception.PackageException;
import org.artifactory.core.api.ph.model.FilteringProperty;
import org.artifactory.core.api.ph.model.PackageArtifact;
import org.artifactory.core.api.ph.model.PackageSearchContext;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;

import java.sql.ResultSet;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.UUID;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.mockito.MockitoAnnotations.openMocks;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertNull;
import static org.testng.Assert.assertTrue;
import static org.testng.Assert.fail;

class PackageSearchArtifactServiceImplTest {


    @Mock
    private AuthorizationService authorizationService;
    @Mock
    private AqlService aqlService;
    private PackageSearchArtifactServiceImpl packageSearchService;

    private final Date date = new Date(System.currentTimeMillis());
    private final String md5 = UUID.randomUUID().toString().replace("-", "");
    private final String sha = UUID.randomUUID().toString().replace("-", "");

    @BeforeEach
    public void setUp() {
        openMocks(this);
        when(authorizationService.canRead(any())).thenReturn(true);
        this.packageSearchService = new PackageSearchArtifactServiceImpl(aqlService, authorizationService);
        AqlEagerResult result = mock(AqlEagerResult.class);
        doAnswer(invocation -> List.of(aqlItem(date, md5, sha + "12345678", sha + sha))).when(result).getResults();
        ResultSet resultSet = mock(ResultSet.class);
        AqlLazyResult resultLazy = mock(AqlLazyResult.class);
        doAnswer(invocation -> resultSet).when(resultLazy).getResultSet();
        when(aqlService.executeQueryEager(any(AqlBase.class))).thenReturn(result);
        when(aqlService.executeQueryLazy(any(AqlBase.class))).thenReturn(resultLazy);
    }

    @ParameterizedTest
    @MethodSource("provideFindArtifactsByExtensionInputs")
    void testFindArtifactsByExtensionInput(String repo, String path, String extension, String expectedError) {
        try {
            packageSearchService.findArtifactsByExtension(repo, path, extension);
            fail("Expected an exception");
        } catch (IllegalArgumentException e) {
            assertEquals(e.getMessage(), expectedError);
        } catch (Exception e) {
            fail("Expected IllegalArgumentException only");
        }
    }

    @Test
    void testFindArtifactsByExtensionAndPath() {
        List<PackageArtifact> arts = packageSearchService.findArtifactsByExtension("repo", "path", "zip", "tar.gz")
                .collect(Collectors.toList());
        ArgumentCaptor<AqlBase> aql = ArgumentCaptor.forClass(AqlBase.class);
        verify(aqlService, times(1)).executeQueryEager(aql.capture());
        assertArts(date, md5, sha, arts);
        String query = aql.getValue().toNative(0);
        assertTrue(query.contains("{\n" +
                "    \"$and\":[\n" +
                "      {\"repo\":{\"$eq\":\"repo\"}},\n" +
                "      {\"$or\":[\n" +
                "        {\"path\":{\"$eq\":\"path\"}},\n" +
                "        {\"path\":{\"$match\":\"path/*\"}}\n" +
                "      ]},\n" +
                "      {\"$or\":[\n" +
                "        {\"name\":{\"$match\":\"*.zip\"}},\n" +
                "        {\"name\":{\"$match\":\"*.tar.gz\"}}\n" +
                "      ]},\n" +
                "      {\"depth\":{\"$gte\":2}},\n" +
                "      {\"type\":{\"$eq\":\"file\"}}\n" +
                "    ]\n" +
                "  }"), query);

    }

    @Test
    void testFindArtifactsByExtensionOnly() {
        List<PackageArtifact> arts = packageSearchService.findArtifactsByExtension("repo", null, "tar.gz")
                .collect(Collectors.toList());
        ArgumentCaptor<AqlBase> aql = ArgumentCaptor.forClass(AqlBase.class);
        verify(aqlService, times(1)).executeQueryEager(aql.capture());
        assertArts(date, md5, sha, arts);
        String query = aql.getValue().toNative(0);
        assertTrue(query.contains("  {\n" +
                "    \"$and\":[\n" +
                "      {\"repo\":{\"$eq\":\"repo\"}},\n" +
                "      {\"$or\":[\n" +
                "        {\"name\":{\"$match\":\"*.tar.gz\"}}\n" +
                "      ]},\n" +
                "      {\"depth\":{\"$gte\":1}},\n" +
                "      {\"type\":{\"$eq\":\"file\"}}\n" +
                "    ]\n" +
                "  }"), query);
    }

    @Test
    void testFindArtifactsChildren() {
        List<PackageArtifact> arts = packageSearchService.findArtifactsChildren("repo", "/path/aaa/bb")
                .collect(Collectors.toList());
        ArgumentCaptor<AqlBase> aql = ArgumentCaptor.forClass(AqlBase.class);
        verify(aqlService, times(1)).executeQueryEager(aql.capture());
        assertArts(date, md5, sha, arts);
        String query = aql.getValue().toNative(0);
        assertTrue(query.contains("{\n" +
                "    \"$and\":[\n" +
                "      {\"repo\":{\"$eq\":\"repo\"}},\n" +
                "      {\"$or\":[\n" +
                "        {\"path\":{\"$eq\":\"/path/aaa/bb\"}},\n" +
                "        {\"path\":{\"$match\":\"/path/aaa/bb/*\"}}\n" +
                "      ]},\n" +
                "      {\"depth\":{\"$gte\":4}},\n" +
                "      {\"type\":{\"$eq\":\"file\"}}\n" +
                "    ]\n" +
                "  }"), query);
    }

    @ParameterizedTest
    @MethodSource("provideFindArtifactsChildrenInputs")
    void testFindArtifactsChildrenInput(String repo, String path, String expectedError) {
        try {
            packageSearchService.findArtifactsChildren(repo, path);
            fail("Expected an exception");
        } catch (IllegalArgumentException e) {
            assertEquals(e.getMessage(), expectedError);
        } catch (Exception e) {
            fail("Expected IllegalArgumentException only");
        }
    }

    @Test
    void testFindArtifactsChildrenStartWithName() {
        List<PackageArtifact> arts =
                packageSearchService.findArtifactsChildrenStartWithName("repo", "/path/aaa/bb", "package_")
                        .collect(Collectors.toList());
        ArgumentCaptor<AqlBase> aql = ArgumentCaptor.forClass(AqlBase.class);
        verify(aqlService, times(1)).executeQueryEager(aql.capture());
        assertArts(date, md5, sha, arts);
        String query = aql.getValue().toNative(0);
        assertTrue(query.contains("  {\n" +
                "    \"$and\":[\n" +
                "      {\"repo\":{\"$eq\":\"repo\"}},\n" +
                "      {\"$or\":[\n" +
                "        {\"path\":{\"$eq\":\"/path/aaa/bb\"}},\n" +
                "        {\"path\":{\"$match\":\"/path/aaa/bb/*\"}}\n" +
                "      ]},\n" +
                "      {\"name\":{\"$match\":\"package_*\"}},\n" +
                "      {\"depth\":{\"$gte\":4}},\n" +
                "      {\"type\":{\"$eq\":\"file\"}}\n" +
                "    ]\n" +
                "  }"), query);
    }

    @ParameterizedTest
    @MethodSource("provideFindArtifactsChildrenStartWithNameInputs")
    void testFindArtifactsChildrenStartWithNameInput(String repo, String path, String name, String expectedError) {
        try {
            packageSearchService.findArtifactsChildrenStartWithName(repo, path, name);
            fail("Expected an exception");
        } catch (IllegalArgumentException e) {
            assertEquals(e.getMessage(), expectedError);
        } catch (Exception e) {
            fail("Expected IllegalArgumentException only");
        }
    }

    @Test
    void testFindArtifactsByExactName() {
        List<PackageArtifact> arts = packageSearchService.findArtifactsByExactName("repo", "PACKAGES")
                .collect(Collectors.toList());
        ArgumentCaptor<AqlBase> aql = ArgumentCaptor.forClass(AqlBase.class);
        verify(aqlService, times(1)).executeQueryEager(aql.capture());
        assertArts(date, md5, sha, arts);
        String query = aql.getValue().toNative(0);
        assertTrue(query.contains("  {\n" +
                "    \"$and\":[\n" +
                "      {\"repo\":{\"$eq\":\"repo\"}},\n" +
                "      {\"name\":{\"$eq\":\"PACKAGES\"}},\n" +
                "      {\"depth\":{\"$gte\":1}},\n" +
                "      {\"type\":{\"$eq\":\"file\"}}\n" +
                "    ]\n" +
                "  }"), query);
    }

    @ParameterizedTest
    @MethodSource("provideFindArtifactsByExactNameInputs")
    void testFindArtifactsByExactNameInput(String repo, String name, String expectedError) {
        try {
            packageSearchService.findArtifactsByExactName(repo, name);
            fail("Expected an exception");
        } catch (IllegalArgumentException e) {
            assertEquals(e.getMessage(), expectedError);
        } catch (Exception e) {
            fail("Expected IllegalArgumentException only");
        }
    }

    @Test
    void testFindArtifactsByExactNameAndPath() {
        Optional<PackageArtifact> artifact =
                packageSearchService.findArtifactByExactPathAndName("repo", "path", "PACKAGES");
        ArgumentCaptor<AqlBase> aql = ArgumentCaptor.forClass(AqlBase.class);
        verify(aqlService, times(1)).executeQueryEager(aql.capture());
        assertTrue(artifact.isPresent());
        assertArts(date, md5, sha, List.of(artifact.get()));
        String query = aql.getValue().toNative(0);
        assertTrue(query.contains("  {\n" +
                "    \"$and\":[\n" +
                "      {\"repo\":{\"$eq\":\"repo\"}},\n" +
                "      {\"path\":{\"$eq\":\"path\"}},\n" +
                "      {\"name\":{\"$eq\":\"PACKAGES\"}},\n" +
                "      {\"depth\":{\"$gte\":2}},\n" +
                "      {\"type\":{\"$eq\":\"file\"}}\n" +
                "    ]\n" +
                "  }"), query);
    }

    @ParameterizedTest
    @MethodSource("provideFindArtifactsByExactNameAndPathInputs")
    void testFindArtifactsByExactNameAndPath(String repo, String path, String name, String expectedError) {
        try {
            packageSearchService.findArtifactByExactPathAndName(repo, path, name);
            fail("Expected an exception");
        } catch (IllegalArgumentException e) {
            assertEquals(e.getMessage(), expectedError);
        } catch (Exception e) {
            fail("Expected IllegalArgumentException only");
        }
    }

    @Test
    void testFindArtifactsByPropKeys() {
        List<PackageArtifact> arts =
                packageSearchService.findArtifactsByPropKeys("repo", "/a/b/c", "deb.component", "deb.distribution")
                        .collect(Collectors.toList());
        ArgumentCaptor<AqlBase> aql = ArgumentCaptor.forClass(AqlBase.class);
        verify(aqlService, times(1)).executeQueryEager(aql.capture());
        assertArts(date, md5, sha, arts);
        String query = aql.getValue().toNative(0);
        assertTrue(query.contains("{\n" +
                "    \"$and\":[\n" +
                "      {\"repo\":{\"$eq\":\"repo\"}},\n" +
                "      {\"path\":{\"$eq\":\"/a/b/c\"}},\n" +
                "      {\"depth\":{\"$gte\":4}},\n" +
                "      {\"property.key\":{\"$eq\":\"deb.component\"}},\n" +
                "      {\"property.key\":{\"$eq\":\"deb.distribution\"}},\n" +
                "      {\"type\":{\"$eq\":\"file\"}}\n" +
                "    ]\n" +
                "  }"), query);
    }

    @ParameterizedTest
    @MethodSource("provideFindArtifactsByPropKeysInputInputs")
    void testFindArtifactsByPropKeysInput(String repo, String path, String propKey, String expectedError) {
        try {
            packageSearchService.findArtifactsByPropKeys(repo, path, propKey);
            fail("Expected an exception");
        } catch (IllegalArgumentException e) {
            assertEquals(e.getMessage(), expectedError);
        } catch (Exception e) {
            fail("Expected IllegalArgumentException only");
        }
    }

    @Test
    void testFindArtifactsByPropKeyAndVal() {
        PackageSearchContext searchContext = new PackageSearchContext("repo", "/a/b/c");
        List<PackageArtifact> arts =
                packageSearchService.findArtifactsByPropKeyAndVal(searchContext, Map.of("deb.component", "a"))
                        .collect(Collectors.toList());
        ArgumentCaptor<AqlBase> aql = ArgumentCaptor.forClass(AqlBase.class);
        verify(aqlService, times(1)).executeQueryEager(aql.capture());
        assertArts(date, md5, sha, arts);
        String query = aql.getValue().toNative(0);
        assertTrue(query.contains("{\n" +
                "    \"$and\":[\n" +
                "      {\"repo\":{\"$eq\":\"repo\"}},\n" +
                "      {\"path\":{\"$eq\":\"/a/b/c\"}},\n" +
                "      {\"depth\":{\"$gte\":4}},\n" +
                "      {\"@deb.component\":{\"$eq\":\"a\"}},\n" +
                "      {\"type\":{\"$eq\":\"file\"}}\n" +
                "    ]\n" +
                "  }"), query);
    }

    @Test
    void testFindArtifactsByPropKeyAndValWithLimitAndOffset() {
        PackageSearchContext searchContext =
                PackageSearchContext.builder().repoKey("repo").path("/a/b/c").offset(1).limit(2).build();
        List<PackageArtifact> arts =
                packageSearchService.findArtifactsByPropKeyAndVal(searchContext, Map.of("deb.component", "a"))
                        .collect(Collectors.toList());
        ArgumentCaptor<AqlBase> aql = ArgumentCaptor.forClass(AqlBase.class);
        verify(aqlService, times(1)).executeQueryEager(aql.capture());
        assertArts(date, md5, sha, arts);
        String query = aql.getValue().toNative(0);
        assertTrue(query.contains("{\n" +
                "    \"$and\":[\n" +
                "      {\"repo\":{\"$eq\":\"repo\"}},\n" +
                "      {\"path\":{\"$eq\":\"/a/b/c\"}},\n" +
                "      {\"depth\":{\"$gte\":4}},\n" +
                "      {\"@deb.component\":{\"$eq\":\"a\"}},\n" +
                "      {\"type\":{\"$eq\":\"file\"}}\n" +
                "    ]\n" +
                "  }"), query);
        assertTrue(query.contains("offset(1).\n" +
                "limit(2)"));
    }

    @ParameterizedTest
    @MethodSource("provideFindArtifactsByPropKeyAndValInputs")
    void testFindArtifactsByPropKeyAndValInput(String repo, String path, String propKey, String propVal,
            String expectedError) {
        PackageSearchContext searchContext = new PackageSearchContext(repo, path);
        try {
            packageSearchService.findArtifactsByPropKeyAndVal(searchContext, Map.of(propKey, propVal));
            fail("Expected an exception");
        } catch (IllegalArgumentException e) {
            assertEquals(e.getMessage(), expectedError);
        } catch (Exception e) {
            fail("Expected IllegalArgumentException only");
        }
    }

    @Test
    void testFindArtifactsWithProperties() throws PackageException {
        List<FilteringProperty> filterByParams = List.of(FilteringProperty.builder()
                .propertyName("nuget.id").propertyValue("helloMyget").isCaseSensitive(false).build());
        packageSearchService.findArtifactsWithProps("repo", null, filterByParams);
        ArgumentCaptor<AqlBase> aql = ArgumentCaptor.forClass(AqlBase.class);
        verify(aqlService, times(1)).executeQueryLazy(aql.capture());
        String query = aql.getValue().toNative(0);
        assertTrue(query.contains("items.find(\n" +
                "  {\n" +
                "    \"$and\":[\n" +
                "      {\"repo\":{\"$eq\":\"repo\"}},\n" +
                "      {\"depth\":{\"$gte\":1}},\n" +
                "      {\"$and\":[\n" +
                "        {\"property.key\":{\"$eq\":\"nuget.id\"}},\n" +
                "        {\"property.value\":{\"$eqic\":\"HELLOMYGET\"}}\n" +
                "      ]}\n" +
                "    ]\n" +
                "  }\n" +
                ")"));
    }


    private AqlItem aqlItem(Date date, String md5, String sha1, String sha2) {
        AqlItem aqlItem = mock(AqlItem.class);
        when(aqlItem.getPath()).thenReturn("path");
        when(aqlItem.getName()).thenReturn("name");
        when(aqlItem.getRepo()).thenReturn("repo");
        when(aqlItem.getSize()).thenReturn(Long.MAX_VALUE);
        when(aqlItem.getSha2()).thenReturn(sha2);
        when(aqlItem.getActualSha1()).thenReturn(sha1);
        when(aqlItem.getNumericCountValue()).thenReturn(Long.MAX_VALUE);
        when(aqlItem.getNodeId()).thenReturn(Long.MAX_VALUE);
        when(aqlItem.getCreated()).thenReturn(date);
        when(aqlItem.getModified()).thenReturn(date);
        when(aqlItem.getUpdated()).thenReturn(date);
        when(aqlItem.getCreatedBy()).thenReturn("admin");
        when(aqlItem.getActualMd5()).thenReturn(md5);
        when(aqlItem.getModifiedBy()).thenReturn("admin");
        when(aqlItem.getOriginalSha1()).thenReturn(sha1);
        when(aqlItem.getType()).thenReturn(AqlItemTypeEnum.file);
        when(aqlItem.getOriginalMd5()).thenReturn(md5);
        when(aqlItem.getDepth()).thenReturn(Integer.MAX_VALUE);
        when(aqlItem.getVirtualRepos()).thenReturn(new String[] {"v1", "v2"});
        return aqlItem;
    }

    private void assertArts(Date date, String md5, String sha, List<PackageArtifact> arts) {
        assertEquals(arts.size(), 1);
        PackageArtifact art = arts.iterator().next();
        assertEquals(art.getRepoKey(), "repo");
        assertEquals(art.getPath(), "path/name");
        assertEquals(art.getCreated(), date.getTime());
        assertEquals(art.getCreatedBy(), "admin");
        assertEquals(art.getLastUpdated(), date.getTime());
        assertEquals(art.getLastModified(), date.getTime());
        assertEquals(art.getId(), Long.MAX_VALUE);
        assertEquals(art.getLength(), Long.MAX_VALUE);
        assertEquals(art.getMd5(), md5);
        assertEquals(art.getModifiedBy(), "admin");
        assertNull(art.getPackageType());
        assertEquals(art.getSha1(), sha + "12345678");
        assertEquals(art.getSha2(), sha + sha);
        assertEquals(art.getName(), "name");
    }

    static Stream<Arguments> provideFindArtifactsByExtensionInputs() {
        return Stream.of(
                Arguments.of("repo", "some/path*", "zip", "Parameter 'some/path*' cannot include '*'"),
                Arguments.of("repo", null, "*.zip", "Parameter '*.zip' cannot include '*'"),
                Arguments.of("repo", null, null, "Extension cannot be empty"),
                Arguments.of("repo", null, "", "Extension cannot be empty"),
                Arguments.of("", "", "", "RepoKey cannot be empty")
        );
    }

    static Stream<Arguments> provideFindArtifactsChildrenInputs() {
        return Stream.of(
                Arguments.of("repo", "some/path*", "Parameter 'some/path*' cannot include '*'"),
                Arguments.of("repo", "/*", "Parameter '/*' cannot include '*'"),
                Arguments.of("", "", "RepoKey cannot be empty")
        );
    }

    static Stream<Arguments> provideFindArtifactsChildrenStartWithNameInputs() {
        return Stream.of(
                Arguments.of("repo", "some/path*", "", "Parameter 'some/path*' cannot include '*'"),
                Arguments.of("repo", "/*", "", "Parameter '/*' cannot include '*'"),
                Arguments.of("repo", "/", "a*", "Parameter 'a*' cannot include '*'"),
                Arguments.of("repo", "/", "", "Input cannot be empty"),
                Arguments.of("", "", "", "RepoKey cannot be empty")
        );
    }

    static Stream<Arguments> provideFindArtifactsByExactNameInputs() {
        return Stream.of(
                Arguments.of("repo", "", "Input cannot be empty"),
                Arguments.of("repo", "*", "Parameter '*' cannot include '*'"),
                Arguments.of("repo", null, "Input cannot be empty"),
                Arguments.of("repo", "someFile.*", "Parameter 'someFile.*' cannot include '*'"),
                Arguments.of("", "", "RepoKey cannot be empty")
        );
    }

    static Stream<Arguments> provideFindArtifactsByExactNameAndPathInputs() {
        return Stream.of(
                Arguments.of("repo", "path", "", "Input cannot be empty"),
                Arguments.of("repo", "path", "*", "Parameter '*' cannot include '*'"),
                Arguments.of("repo", "path/*", "name", "Parameter 'path/*' cannot include '*'"),
                Arguments.of("repo", "", "someFile.*", "Parameter 'someFile.*' cannot include '*'"),
                Arguments.of("", "", "", "RepoKey cannot be empty")
        );
    }

    static Stream<Arguments> provideFindArtifactsByPropKeysInputInputs() {
        return Stream.of(
                Arguments.of("repo", "path", "", "PropKey cannot be empty"),
                Arguments.of("repo", "path", "*", "Parameter '*' cannot include '*'"),
                Arguments.of("repo", "path/*", "name", "Parameter 'path/*' cannot include '*'"),
                Arguments.of("repo", "", "deb.*", "Parameter 'deb.*' cannot include '*'"),
                Arguments.of("repo", "", null, "PropKey cannot be empty"),
                Arguments.of("", "", "", "RepoKey cannot be empty")
        );
    }

    static Stream<Arguments> provideFindArtifactsByPropKeyAndValInputs() {
        return Stream.of(
                Arguments.of("repo", "path", "", "", "PropKey cannot be empty"),
                Arguments.of("repo", "path", "*", "", "Parameter '*' cannot include '*'"),
                Arguments.of("repo", "path/*", "name", "", "Parameter 'path/*' cannot include '*'"),
                Arguments.of("repo", "", "deb.*", "", "Parameter 'deb.*' cannot include '*'"),
                Arguments.of("repo", "", "deb.arch", "*", "Parameter '*' is not a legal property value"),
                Arguments.of("", "", "", "", "RepoKey cannot be empty")
        );
    }
}