package org.artifactory.release.bundle;

import com.google.common.collect.Lists;
import lombok.NonNull;
import org.apache.http.concurrent.BasicFuture;
import org.artifactory.api.common.MoveMultiStatusHolder;
import org.artifactory.api.component.ComponentDetailsFetcher;
import org.artifactory.api.properties.PropertiesService;
import org.artifactory.api.rest.distribution.bundle.models.ReleaseArtifact;
import org.artifactory.api.rest.distribution.bundle.models.ReleaseBundleModel;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.bundle.BundleTransactionStatus;
import org.artifactory.bundle.BundleType;
import org.artifactory.model.common.RepoPathImpl;
import org.artifactory.release.bundle.interceptor.ReleaseBundleInterceptors;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.repo.service.RepositoryServiceImpl;
import org.artifactory.rest.exception.ForbiddenWebAppException;
import org.artifactory.rest.exception.NotFoundException;
import org.artifactory.storage.db.bundle.dao.ArtifactBundlesDao;
import org.artifactory.storage.db.bundle.dao.BundleBlobsDao;
import org.artifactory.storage.db.bundle.model.BundleNode;
import org.artifactory.storage.db.bundle.model.DBArtifactsBundle;
import org.artifactory.storage.db.fs.dao.NodesDao;
import org.artifactory.storage.db.fs.entity.Node;
import org.artifactory.storage.db.fs.entity.NodePath;
import org.artifactory.storage.fs.service.FileService;
import org.jfrog.storage.StorageException;
import org.joda.time.DateTime;
import org.junit.jupiter.api.AfterEach;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;
import org.testng.Assert;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.Future;
import java.util.concurrent.SynchronousQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import static org.artifactory.release.bundle.ReleaseBundleServiceImpl.COULD_NOT_DELETE_RB_ERROR_MSG;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyBoolean;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.doThrow;
import static org.mockito.Mockito.times;

import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoMoreInteractions;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertThrows;
import static org.testng.Assert.assertTrue;

/**
 * @author Eli Skoran 04/05/2020
 */
@Test
public class ReleaseBundleServiceImplTest {

    @Mock
    ArtifactBundlesDao bundlesDao;
    @Mock
    BundleBlobsDao blobsDao;
    @Mock
    NodesDao nodesDao;
    @Mock
    RepositoryServiceImpl repoService;
    @Mock
    PropertiesService propertiesService;
    @Mock
    ComponentDetailsFetcher componentDetailsFetcher;
    @Mock
    AuthorizationService authorizationService;
    @Mock
    InternalReleaseBundleService internalReleaseBundleService;
    @Mock
    FileService fileService;
    @Mock
    ReleaseBundleInterceptors releaseBundleInterceptors;

    private AutoCloseable closable;

    private final String bundleName = "rb_name";
    private final String bundleVersion = "rb_version";

    private ReleaseBundleServiceImpl service;

    @BeforeMethod
    public void setup() throws SQLException {
        closable = MockitoAnnotations.openMocks(this);

        service = new ReleaseBundleServiceImpl(
                bundlesDao,
                blobsDao,
                nodesDao,
                repoService,
                propertiesService,
                componentDetailsFetcher,
                authorizationService,
                internalReleaseBundleService,
                fileService,
                releaseBundleInterceptors
        );

        service.executorService = new SyncedExecutorService();
        service.initialTimeBetweenRetries = 1;
    }

    @AfterEach
    public void tearDown() throws Exception {
        verifyNoMoreInteractions(bundlesDao,
                blobsDao,
                nodesDao,
                repoService,
                propertiesService,
                componentDetailsFetcher,
                authorizationService,
                internalReleaseBundleService,
                fileService,
                releaseBundleInterceptors);
        closable.close();
    }

    public void testCopyBundleArtifactsAsync_basic() throws SQLException {
        when(propertiesService.setProperties(any(), any(), anyBoolean())).thenReturn(true);
        when(nodesDao.getNodeId(any())).thenReturn(111111L);
        when(bundlesDao.create(any(BundleNode.class))).thenReturn(1);
        when(bundlesDao.completeBundle(any(), any(), any())).thenReturn(1);

        service.copyBundleArtifactsAsync(getTestModel(), "dummyRepo", getTestMapping());

        verify(propertiesService, times(2)).setProperties(any(), any(), anyBoolean());
        verify(nodesDao, times(2)).getNodeId(any());
        verify(bundlesDao, times(2)).create(any(BundleNode.class));
        verify(bundlesDao, times(1)).completeBundle(any(), any(), any());

    }

    public void testCopyBundleArtifactsAsync_failStore() throws SQLException {
        when(propertiesService.setProperties(any(), any(), anyBoolean())).thenReturn(true);
        when(nodesDao.getNodeId(any())).thenReturn(111111L);
        when(bundlesDao.create(any(BundleNode.class))).thenThrow(new SQLException(
                "MySQLTransactionRollbackException: Lock wait timeout exceeded; try restarting transaction"));
        when(bundlesDao.setBundleStatus("platform", "7.5.1", BundleType.SOURCE, BundleTransactionStatus.FAILED))
                .thenReturn(1);

        service.copyBundleArtifactsAsync(getTestModel(), "dummyRepo", getTestMapping());

        verify(propertiesService, times(1)).setProperties(any(), any(), anyBoolean());
        verify(nodesDao, times(1)).getNodeId(any());
        verify(bundlesDao, times(1)).create(any(BundleNode.class));
        verify(bundlesDao, times(1))
                .setBundleStatus("platform", "7.5.1", BundleType.SOURCE, BundleTransactionStatus.FAILED);
    }

    @Test (expectedExceptions = StorageException.class)
    public void testIsReleaseBundleExist() throws SQLException {
        when(bundlesDao.isReleaseBundleExist("rb1")).thenReturn(true);
        Assert.assertTrue(service.isReleaseBundleExistByName("rb1"));
        when(bundlesDao.isReleaseBundleExist("rb1")).thenThrow(new SQLException());
        service.isReleaseBundleExistByName("rb1");
    }

    @Test
    public void testCreateBundleNodes() throws SQLException {
        List<RepoPath> repoPaths = Lists.newArrayList(
                new RepoPathImpl("repo1", "my/path/file1"),
                new RepoPathImpl("repo1", "my/path/file2"));
        long bundleNodeId = 555;
        Node node1 = Mockito.mock(Node.class);
        Node node2 = Mockito.mock(Node.class);
        when(node1.getNodeId()).thenReturn(1L);
        when(node2.getNodeId()).thenReturn(2L);
        Map<NodePath, Node> map = new HashMap<>();
        List<NodePath> nodePaths = NodePath.fromRepoPaths(repoPaths);
        map.put(nodePaths.get(0), node1);
        map.put(nodePaths.get(1), node2);
        when(nodesDao.get(NodePath.fromRepoPaths(repoPaths))).thenReturn(map);
        service.createBundleNodes(bundleNodeId, repoPaths);
        List<BundleNode> bundleNodesToCreate = new ArrayList<>();
        for (Map.Entry<NodePath, Node> entry : map.entrySet()) {
            BundleNode bundleNode = new BundleNode();
            bundleNode.setBundleId(bundleNodeId);
            bundleNode.setNodeId(entry.getValue().getNodeId());
            bundleNode.setRepoPath(entry.getKey().toRepoPath().toPath());
            bundleNodesToCreate.add(bundleNode);
        }
        verify(bundlesDao).createBulk(bundleNodesToCreate);
    }

    public void testCopyBundleArtifactsAsync_failStoreWithRetries() throws SQLException {
        when(propertiesService.setProperties(any(), any(), anyBoolean())).thenReturn(true);
        when(nodesDao.getNodeId(any())).thenReturn(111111L);
        when(bundlesDao.create(any(BundleNode.class))).thenThrow(new SQLException(
                "MySQLTransactionRollbackException: Lock wait timeout exceeded; try restarting transaction"));
        // fail for the first few tries, then succeed..
        when(bundlesDao.setBundleStatus("platform", "7.5.1", BundleType.SOURCE, BundleTransactionStatus.FAILED))
                .thenThrow(new SQLException(
                        "MySQLTransactionRollbackException: Lock wait timeout exceeded; try restarting transaction"))
                .thenThrow(new SQLException(
                        "MySQLTransactionRollbackException: Lock wait timeout exceeded; try restarting transaction"))
                .thenThrow(new SQLException(
                        "MySQLTransactionRollbackException: Lock wait timeout exceeded; try restarting transaction"))
                .thenReturn(1);

        service.copyBundleArtifactsAsync(getTestModel(), "dummyRepo", getTestMapping());

        verify(propertiesService, times(1)).setProperties(any(), any(), anyBoolean());
        verify(nodesDao, times(1)).getNodeId(any());
        verify(bundlesDao, times(1)).create(any(BundleNode.class));
        verify(bundlesDao, times(4))
                .setBundleStatus("platform", "7.5.1", BundleType.SOURCE, BundleTransactionStatus.FAILED);
    }

    public void testCleanOrphanArtifactsNoArtifactsToDelete() throws SQLException {
        when(bundlesDao.getOrphanRBVRepoPathsFromIntransit()).thenReturn(new HashSet<>());
        when(bundlesDao.getUnidentifiedRepoPathsOnInTransit()).thenReturn(new HashSet<>());
        service.cleanOrphanAndUnidentifiedArtifacts();
        verify(bundlesDao, times(1)).getOrphanRBVRepoPathsFromIntransit();
        verify(bundlesDao, times(1)).getUnidentifiedRepoPathsOnInTransit();
        verify(repoService, times(0)).delete(any());
    }

    @Test
    public void testSetTargetBundleKeep_happyFlow() throws SQLException {
        // setup
        String bundleName = "b1";
        String bundleVersion = "v1";

        when(bundlesDao.setKeepBundle(bundleName, bundleVersion, true)).thenReturn(1);

        // test
        boolean success = service.setTargetBundleKeep(bundleName, bundleVersion, true);

        // verify
        verify(bundlesDao, times(1)).setKeepBundle(bundleName, bundleVersion, true);

        assertTrue(success);
    }

    @Test(expectedExceptions = StorageException.class)
    public void testSetTargetBundleKeep_errorHandling() throws SQLException {
        // setup
        String bundleName = "b1";
        String bundleVersion = "v1";

        when(bundlesDao.setKeepBundle(bundleName, bundleVersion, true)).thenThrow(
                new SQLException("sql something"));

        // test
        boolean success = service.setTargetBundleKeep(bundleName, bundleVersion, true);

        // verify
        verify(bundlesDao, times(1)).setKeepBundle(bundleName, bundleVersion, true);

        assertTrue(success);
    }

    @Test
    public void testGetTargetBundleKeep_happyFlow() throws SQLException {
        // setup
        String bundleName = "b1";
        String bundleVersion = "v1";

        when(bundlesDao.isKeepBundle(bundleName, bundleVersion)).thenReturn(true);

        // test
        boolean isKeep = service.getTargetBundleKeep(bundleName, bundleVersion);

        // verify
        verify(bundlesDao, times(1)).isKeepBundle(bundleName, bundleVersion);

        assertTrue(isKeep);
    }

    @Test(expectedExceptions = StorageException.class)
    public void testGetTargetBundleKeep_errorHandling() throws SQLException {
        // setup
        String bundleName = "b1";
        String bundleVersion = "v1";

        when(bundlesDao.isKeepBundle(bundleName, bundleVersion)).thenThrow(
                new SQLException("sql something"));

        // test
        boolean isKeep = service.getTargetBundleKeep(bundleName, bundleVersion);

        // verify
        verify(bundlesDao, times(1)).isKeepBundle(bundleName, bundleVersion);

        assertTrue(isKeep);
    }

    @Test (dataProvider = "OrphanArtifactsTestDataProvider")
    public void testCleanOrphanArtifactsWithArtifactsToDelete(boolean hasWarning, boolean hasErrors,
            boolean expectGetMsg) throws SQLException {
        RepoPath repoPathToDelete1 = RepoPathFactory.create("_intransit/path/to/dir/artifact_to_delete_1.ext");
        RepoPath repoPathToDelete2 = RepoPathFactory.create("_intransit/path/to/dir/artifact_to_delete_2.ext");

        Set<RepoPath> orphanRepoPathsToDelete = new HashSet<>();
        orphanRepoPathsToDelete.add(repoPathToDelete1);

        Set<RepoPath> invalidRepoPathsToDelete = new HashSet<>();
        invalidRepoPathsToDelete.add(repoPathToDelete2);

        MoveMultiStatusHolder statusHolder = Mockito.mock(MoveMultiStatusHolder.class);
        when(bundlesDao.getOrphanRBVRepoPathsFromIntransit()).thenReturn(orphanRepoPathsToDelete);
        when(bundlesDao.getUnidentifiedRepoPathsOnInTransit()).thenReturn(invalidRepoPathsToDelete);
        when(repoService.delete(orphanRepoPathsToDelete)).thenReturn(statusHolder);
        when(statusHolder.hasWarnings()).thenReturn(hasWarning);
        when(statusHolder.hasErrors()).thenReturn(hasErrors);

        service.cleanOrphanAndUnidentifiedArtifacts();
        verify(bundlesDao, times(1)).getOrphanRBVRepoPathsFromIntransit();
        verify(bundlesDao, times(1)).getUnidentifiedRepoPathsOnInTransit();

        invalidRepoPathsToDelete.addAll(orphanRepoPathsToDelete);
        verify(repoService, times(1)).delete(invalidRepoPathsToDelete);
        verify(statusHolder, times(expectGetMsg ? 1 : 0)).getStatusMsg();
    }


    @Test
    public void testDeleteNotExistingReleaseBundle() throws SQLException {
        when(bundlesDao.getArtifactsBundle(bundleName, bundleVersion, BundleType.TARGET))
                .thenReturn(null);
        assertThrows(NotFoundException.class,
                () -> service.deleteReleaseBundle(bundleName, bundleVersion, BundleType.TARGET));
        verify(bundlesDao).getArtifactsBundle(bundleName, bundleVersion, BundleType.TARGET);
    }

    @Test
    public void testDeleteReleaseBundleExceptionWhileReadingFromDB() throws SQLException {
        String errorMessage = "Error deleting release bundle";
        when(bundlesDao.getArtifactsBundle(bundleName, bundleVersion, BundleType.TARGET))
                .thenThrow(new SQLException(errorMessage));
        assertThrows(StorageException.class,
                () -> service.deleteReleaseBundle(bundleName, bundleVersion, BundleType.TARGET));
        verify(bundlesDao).getArtifactsBundle(bundleName, bundleVersion, BundleType.TARGET);
    }

    @Test
    public void testDeleteReleaseBundleNoRepoDeletePermission() throws SQLException {
        prepareMocksForReleaseBundleDeletion(false);
        assertThrows(ForbiddenWebAppException.class,
                () -> service.deleteReleaseBundle(bundleName, bundleVersion, BundleType.TARGET));
        verify(bundlesDao).getArtifactsBundle(bundleName, bundleVersion, BundleType.TARGET);
        verify(releaseBundleInterceptors).onDeleteReleaseBundleRequestFailed(any(),
                eq("Delete release bundle " + bundleName + ":" + bundleVersion + " is forbidden"));
    }

    @Test
    public void testDeleteReleaseBundleSuccessfully() throws SQLException {
        prepareMocksForReleaseBundleDeletion(true);
        service.deleteReleaseBundle(bundleName, bundleVersion, BundleType.TARGET);
        verify(bundlesDao).getArtifactsBundle(bundleName, bundleVersion, BundleType.TARGET);
        verify(releaseBundleInterceptors).onDeleteReleaseBundleRequestStarted(any());
        verify(internalReleaseBundleService).deleteReleaseBundleInternal(any(), any(), any());
        verify(releaseBundleInterceptors).onDeleteReleaseBundleRequestCompleted(any());
    }

    @Test
    public void testDeleteReleaseBundleFailedWhileDeleting() throws SQLException {
        prepareMocksForReleaseBundleDeletion(true);
        doThrow(new SQLException()).when(internalReleaseBundleService)
                .deleteReleaseBundleInternal(any(), any(), any());
        assertThrows(StorageException.class,
                () -> service.deleteReleaseBundle(bundleName, bundleVersion, BundleType.TARGET));
        verify(bundlesDao).getArtifactsBundle(bundleName, bundleVersion, BundleType.TARGET);
        verify(releaseBundleInterceptors).onDeleteReleaseBundleRequestStarted(any());
        verify(internalReleaseBundleService).deleteReleaseBundleInternal(any(), any(), any());
        verify(releaseBundleInterceptors).onDeleteReleaseBundleRequestFailed(any(), eq(COULD_NOT_DELETE_RB_ERROR_MSG));
    }

    private void prepareMocksForReleaseBundleDeletion(boolean canDelete)
            throws SQLException {
        DBArtifactsBundle artifactsBundle = getSimpleDbArtifactsBundle();
        when(bundlesDao.getArtifactsBundle(bundleName, bundleVersion, BundleType.TARGET))
                .thenReturn(artifactsBundle);
        when(authorizationService.canRead(any())).thenReturn(true);
        when(authorizationService.canDelete(any())).thenReturn(canDelete);
    }

    @NonNull
    private DBArtifactsBundle getSimpleDbArtifactsBundle() {
        DBArtifactsBundle artifactsBundle = new DBArtifactsBundle();
        artifactsBundle.setName(bundleName);
        artifactsBundle.setVersion(bundleVersion);
        String repoKey = "repo_key";
        artifactsBundle.setStoringRepo(repoKey);
        artifactsBundle.setDateCreated(new DateTime());
        artifactsBundle.setStatus(BundleTransactionStatus.INPROGRESS);
        artifactsBundle.setId(111L);
        return artifactsBundle;
    }

    @DataProvider (name = "OrphanArtifactsTestDataProvider")
    private Object[][] getOrphanArtifactsTestDataProvider() {
        return new Object[][] {
                {true, true, true},
                {true, false, true},
                {false, true, true},
                {false, false, false}};
    }

    private ReleaseBundleModel getTestModel() {
        ReleaseArtifact artifact1 = new ReleaseArtifact();
        artifact1.setRepoPath("dummyRepo/file1.txt");
        artifact1.setChecksum("111aaa");

        ReleaseArtifact artifact2 = new ReleaseArtifact();
        artifact2.setRepoPath("dummyRepo/file2.txt");
        artifact2.setChecksum("222bbb");

        ReleaseBundleModel model = new ReleaseBundleModel();
        model.setName("platform");
        model.setVersion("7.5.1");
        model.setArtifacts(Lists.newArrayList(
                artifact1,
                artifact2
        ));

        return model;
    }

    private Map<String, String> getTestMapping() {
        Map<String, String> mapping = new HashMap<>();
        mapping.put("dummyRepo/file1.txt", "dummyRepo/file1.txt");
        mapping.put("dummyRepo/file2.txt", "dummyRepo/file2.txt");

        return mapping;
    }

    //TODO: extract
    private static class SyncedExecutorService extends ThreadPoolExecutor {

        public SyncedExecutorService() {
            super(0,
                    2,
                    60L,
                    TimeUnit.SECONDS,
                    new SynchronousQueue<>());
        }

        @Override
        public Future<?> submit(Runnable task) {
            task.run();
            return new BasicFuture<>(null);
        }
    }
}
