package org.artifactory.repo.service.mover;

import org.apache.http.HttpStatus;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.replication.ReplicationAddon;
import org.artifactory.api.common.MoveMultiStatusHolder;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.model.common.RepoPathImpl;
import org.artifactory.model.xstream.fs.FolderInfoImpl;
import org.artifactory.repo.LocalRepo;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoRepoPath;
import org.artifactory.repo.StoringRepo;
import org.artifactory.sapi.fs.VfsItem;
import org.artifactory.storage.db.fs.model.DbMutableFile;
import org.artifactory.storage.db.fs.model.DbMutableFolder;
import org.artifactory.storage.db.fs.session.SqlStorageSession;
import org.artifactory.storage.fs.MutableVfsFile;
import org.artifactory.storage.fs.MutableVfsItem;
import org.artifactory.storage.fs.lock.SessionLockEntry;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import static org.artifactory.repo.service.flexible.tools.MoveCopyTestUtils.prepareCopyOrMoveOverrideMocks;
import static org.artifactory.repo.service.flexible.tools.MoveCopyTestUtils.prepareDbMutableFileMock;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.any;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoMoreInteractions;
import static org.mockito.Mockito.when;
import static org.mockito.Mockito.withSettings;
import static org.testng.Assert.assertEquals;

public class BaseRepoPathMoverTest extends ArtifactoryHomeBoundTest {

    @Mock
    RepoPath repoPathMock;
    @Mock
    LocalRepo repoMock;
    @Mock
    MutableVfsItem mutableVfsItemMock;
    @Mock
    SqlStorageSession sqlStorageSessionMock;
    @Mock
    SessionLockEntry sessionLockEntryMock;
    @Mock
    MoveMultiStatusHolder moveMultiStatusHolderMock;
    @Mock
    private ArtifactoryContext context;
    @Mock
    private AddonsManager addonsManager;
    @Mock
    private ReplicationAddon replicationAddon;
    @Mock
    private AuthorizationService authorizationService;

    @BeforeMethod
    public void setup() {
        MockitoAnnotations.openMocks(this);
        when(context.beanForType(AddonsManager.class)).thenReturn(addonsManager);
        when(context.getAuthorizationService()).thenReturn(authorizationService);
        when(addonsManager.addonByType(ReplicationAddon.class)).thenReturn(replicationAddon);
        ArtifactoryContextThreadBinder.bind(context);
    }

    @AfterMethod
    public void tearDown() {
        ArtifactoryContextThreadBinder.unbind();
    }

    @Test
    public void testMoveFile() {
        BaseRepoPathMover baseRepoPathMove = mock(BaseRepoPathMover.class, Mockito.CALLS_REAL_METHODS);
        DbMutableFile dbMutableFile = prepareDbMutableFileMock();
        RepoRepoPath<StoringRepo> repoRepoPath =
                prepareCopyOrMoveOverrideMocks(repoPathMock, repoMock, mutableVfsItemMock, sessionLockEntryMock,
                        sqlStorageSessionMock, moveMultiStatusHolderMock, true);

        baseRepoPathMove.moveFile(dbMutableFile, repoRepoPath, moveMultiStatusHolderMock, true);

        verify(repoMock, times(1)).createOrGetFile(any());
        verify(sqlStorageSessionMock, times(1)).save();
        verify(moveMultiStatusHolderMock, times(1)).artifactMoved();
    }

    /**
     * Depending on the unixStyleBehaviour target dir will contain `source-dir` as a child.
     */
    @SuppressWarnings("unchecked")
    @Test(dataProvider = "unixStyleBehaviourTest")
    public void testMoveDirUnixStyle(boolean unixStyleBehaviour, String targetDirExpectedPath) {
        RepoPathImpl fromRepoPath = new RepoPathImpl("source-repo", "source-dir");
        RepoPathImpl targetRepoPath = new RepoPathImpl("source-repo", ".index");

        FolderInfoImpl srcFolderInfo = new FolderInfoImpl(fromRepoPath);
        DbMutableFolder srcDir = new DbMutableFolder(
                mock(org.artifactory.storage.fs.repo.StoringRepo.class), 52437842367L, srcFolderInfo);

        FolderInfoImpl destFolderInfo = new FolderInfoImpl(targetRepoPath);
        DbMutableFolder destinationDir = new DbMutableFolder(
                mock(org.artifactory.storage.fs.repo.StoringRepo.class), 35865650921L, destFolderInfo);

        final StoringRepo<LocalRepoDescriptor> repo = mock(StoringRepo.class);
        when(repo.createOrGetFile(any())).thenReturn(mock(MutableVfsFile.class));
        RepoRepoPath<org.artifactory.repo.StoringRepo> targetRepoRepoPath = new RepoRepoPath<>(repo, targetRepoPath);

        when(repo.getImmutableFsItem(targetRepoPath)).thenReturn(destinationDir);

        BaseRepoPathMover baseRepoPathMove = spy(new DefaultRepoPathMover(new MoveMultiStatusHolder(),
                new MoverConfigBuilder(fromRepoPath, targetRepoPath)
                        .dryRun(true)
                        .unixStyleBehavior(unixStyleBehaviour).build()));
        doReturn(true).when(baseRepoPathMove).canMove(eq(srcDir), any());
        final ArgumentCaptor<VfsItem> sourceArg = ArgumentCaptor.forClass(VfsItem.class);
        final ArgumentCaptor<RepoRepoPath> targetArg = ArgumentCaptor.forClass(RepoRepoPath.class);
        doNothing().when(baseRepoPathMove).moveCopyRecursive(sourceArg.capture(), targetArg.capture());
        baseRepoPathMove.moveOrCopy(srcDir, targetRepoRepoPath);

        assertEquals(sourceArg.getValue(), srcDir);
        assertEquals(targetArg.getValue().getRepoPath().getPath(), targetDirExpectedPath);
    }

    @DataProvider(name = "unixStyleBehaviourTest")
    public static Object[][] unixStyleBehaviourTest() {
        return new Object[][] {{true, ".index/source-dir"}, {false, ".index"}};
    }

    @Test
    public void testCanMoveWithoutReadPermissionToSourceRepository() {
        RepoPath sourceRepoPath = mock(RepoPath.class);
        VfsItem item = mock(VfsItem.class);
        when(item.getRepoPath()).thenReturn(sourceRepoPath);
        AuthorizationService authorizationService = mock(AuthorizationService.class);
        when(authorizationService.canRead(sourceRepoPath)).thenReturn(false);
        when(context.getAuthorizationService()).thenReturn(authorizationService);
        AddonsManager addonsManager = mock(AddonsManager.class);
        when(addonsManager.addonByType(ReplicationAddon.class)).thenReturn(mock(ReplicationAddon.class));
        when(context.beanForType(AddonsManager.class)).thenReturn(addonsManager);
        BaseRepoPathMover baseRepoPathMove = mock(BaseRepoPathMover.class,
                withSettings().useConstructor(new MoveMultiStatusHolder(), mock(MoverConfig.class)));
        when(baseRepoPathMove.canMove(any(), any())).thenCallRealMethod();
        Assert.assertFalse(baseRepoPathMove.canMove(item, null));
        Assert.assertEquals(baseRepoPathMove.status.getStatusCode(), HttpStatus.SC_FORBIDDEN);
        Assert.assertTrue(baseRepoPathMove.status.getStatusMsg().contains("User doesn't have permissions to read"));
        verify(item, times(1)).getRepoPath();
        verify(authorizationService, times(1)).canRead(sourceRepoPath);
        verify(context, times(1)).getAuthorizationService();
        verifyNoMoreInteractions(item, authorizationService);
    }

}