package org.artifactory.security;

import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.common.ConstantValues;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.config.CentralConfigDescriptorImpl;
import org.artifactory.descriptor.repo.FederatedRepoDescriptor;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.repo.RepoDetailsType;
import org.artifactory.spring.ArtifactoryApplicationContext;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import static org.artifactory.repo.config.RepoConfigDefaultValues.EXAMPLE_REPO_KEY;
import static org.artifactory.repo.system.SystemRepos.BUILDINFO_REPO_KEY;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertTrue;

/**
 * @author Mor Merhav
 */
public class FirstStartupHelperTest extends ArtifactoryHomeBoundTest {

    @Mock
    private ArtifactoryApplicationContext context;
    @Mock
    private RepositoryService repositoryService;

    @BeforeClass
    public void beforeClass() {
        MockitoAnnotations.openMocks(this);
    }

    @BeforeMethod
    public void setupMethod() {
        ArtifactoryContextThreadBinder.bind(context);
        when(context.getRepositoryService()).thenReturn(repositoryService);
    }

    @AfterMethod
    public void afterMethod() {
        ArtifactoryContextThreadBinder.unbind();
    }

    @Test
    public void testOnlyDefaultReposExistFilterPipelineRepoReturnTrue() {
        CentralConfigDescriptor centralConfigDescriptor = new CentralConfigDescriptorImpl();
        LocalRepoDescriptor exampleRepo = new LocalRepoDescriptor();
        exampleRepo.setKey(EXAMPLE_REPO_KEY);
        LocalRepoDescriptor buildInfoRepo = new LocalRepoDescriptor();
        buildInfoRepo.setKey(BUILDINFO_REPO_KEY);
        LocalRepoDescriptor pipelinesRepo = new LocalRepoDescriptor();
        buildInfoRepo.setKey(ConstantValues.defaultPipelinesRepoKey.getString());

        centralConfigDescriptor.getLocalRepositoriesMap().put(EXAMPLE_REPO_KEY, exampleRepo);
        centralConfigDescriptor.getLocalRepositoriesMap().put(BUILDINFO_REPO_KEY, buildInfoRepo);
        centralConfigDescriptor.getLocalRepositoriesMap()
                .put(ConstantValues.defaultPipelinesRepoKey.getString(), pipelinesRepo);

        when(repositoryService.isRepoContainsFiles(any())).thenReturn(false);

        boolean onlyDefaultReposExist = FirstStartupHelper.onlyDefaultReposExist(centralConfigDescriptor);
        assertTrue(onlyDefaultReposExist);
    }

    @Test
    public void testNotOnlyDefaultReposExistReturnFalse() {
        CentralConfigDescriptor centralConfigDescriptor = new CentralConfigDescriptorImpl();
        LocalRepoDescriptor exampleRepo = new LocalRepoDescriptor();
        exampleRepo.setKey(EXAMPLE_REPO_KEY);
        LocalRepoDescriptor buildInfoRepo = new LocalRepoDescriptor();
        buildInfoRepo.setKey(BUILDINFO_REPO_KEY);
        LocalRepoDescriptor nonDefaultRepo = new LocalRepoDescriptor();
        nonDefaultRepo.setKey("nonDefaultRepo");

        centralConfigDescriptor.getLocalRepositoriesMap().put(EXAMPLE_REPO_KEY, exampleRepo);
        centralConfigDescriptor.getLocalRepositoriesMap().put(BUILDINFO_REPO_KEY, buildInfoRepo);
        centralConfigDescriptor.getLocalRepositoriesMap().put("nonDefaultRepo", nonDefaultRepo);

        when(repositoryService.isRepoContainsFiles(any())).thenReturn(false);

        boolean onlyDefaultReposExist = FirstStartupHelper.onlyDefaultReposExist(centralConfigDescriptor);
        assertFalse(onlyDefaultReposExist);
    }

    @Test
    public void testDefaultReposNotEmptyReturnFalse() {
        CentralConfigDescriptor centralConfigDescriptor = new CentralConfigDescriptorImpl();
        LocalRepoDescriptor exampleRepo = new LocalRepoDescriptor();
        exampleRepo.setKey(EXAMPLE_REPO_KEY);
        LocalRepoDescriptor buildInfoRepo = new LocalRepoDescriptor();
        buildInfoRepo.setKey(BUILDINFO_REPO_KEY);

        centralConfigDescriptor.getLocalRepositoriesMap().put(EXAMPLE_REPO_KEY, exampleRepo);
        centralConfigDescriptor.getLocalRepositoriesMap().put(BUILDINFO_REPO_KEY, buildInfoRepo);

        when(repositoryService.isRepoContainsFiles(any())).thenReturn(true);

        boolean onlyDefaultReposExist = FirstStartupHelper.onlyDefaultReposExist(centralConfigDescriptor);
        assertFalse(onlyDefaultReposExist);
    }

    @Test
    public void testDefaultReposNotEmptyWithFederationReturnFalse() {
        CentralConfigDescriptor centralConfigDescriptor = new CentralConfigDescriptorImpl();
        LocalRepoDescriptor exampleRepo = new LocalRepoDescriptor();
        exampleRepo.setKey(EXAMPLE_REPO_KEY);
        LocalRepoDescriptor buildInfoRepo = new LocalRepoDescriptor();
        buildInfoRepo.setKey(BUILDINFO_REPO_KEY);
        FederatedRepoDescriptor federatedRepoDescriptor = new FederatedRepoDescriptor();
        centralConfigDescriptor.getFederatedRepositoriesMap()
                .put(RepoDetailsType.FEDERATED_REPO, federatedRepoDescriptor);


        centralConfigDescriptor.getLocalRepositoriesMap().put(EXAMPLE_REPO_KEY, exampleRepo);
        centralConfigDescriptor.getLocalRepositoriesMap().put(BUILDINFO_REPO_KEY, buildInfoRepo);

        boolean onlyDefaultReposExist = FirstStartupHelper.onlyDefaultReposExist(centralConfigDescriptor);
        assertFalse(onlyDefaultReposExist);
    }
}