package org.artifactory.servicetrust;

import lombok.SneakyThrows;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.federated.FederatedRepoAddon;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.rest.exception.BadRequestException;
import org.artifactory.security.access.AccessService;
import org.artifactory.spring.InternalArtifactoryContext;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.jfrog.access.common.ServiceId;
import org.jfrog.access.token.JwtAccessToken;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.springframework.security.core.context.SecurityContextHolder;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeTest;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.io.IOException;
import java.util.List;
import java.util.Optional;

import static org.artifactory.servicetrust.ServiceTrustServiceImpl.ERROR_PAIRING_INSTANCE_WITH_ITSELF;
import static org.artifactory.servicetrust.ServiceTrustServiceImpl.ERROR_PAIRING_TOKEN_EXPIRED;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.when;

public class ServiceTrustServiceTest extends ArtifactoryHomeBoundTest {

    @Mock
    private InternalArtifactoryContext context;

    @Mock
    private AddonsManager addonsManager;

    @Mock
    private FederatedRepoAddon federatedRepoAddon;

    private AutoCloseable closeableMock;

    @BeforeTest
    private void setup() {
        closeableMock = MockitoAnnotations.openMocks(this);
        ArtifactoryContextThreadBinder.bind(context);
        when(addonsManager.addonByType(FederatedRepoAddon.class)).thenReturn(federatedRepoAddon);
        when(context.beanForType(AddonsManager.class)).thenReturn(addonsManager);
    }

    @AfterMethod
    private void cleanup() throws Exception {
        ArtifactoryContextThreadBinder.unbind();
        SecurityContextHolder.clearContext();
        closeableMock.close();
    }

    @Test(expectedExceptions = BadRequestException.class,
            expectedExceptionsMessageRegExp = ERROR_PAIRING_TOKEN_EXPIRED)
    public void testPairingShouldFailInCasePairingTokenExpired() throws IOException {
        ServiceTrustServiceImpl service = spy(ServiceTrustServiceImpl.class);

        var jwt = mock(JwtAccessToken.class);
        when(jwt.getExpirationTime()).thenReturn(System.currentTimeMillis() - 10000);

        doReturn(jwt).when(service).parseTokenValue(anyString());

        service.pairToken(ServiceTrustUseCase.FEDERATED_REPO, "token");
    }

    @SneakyThrows
    @Test(expectedExceptions = BadRequestException.class,
            expectedExceptionsMessageRegExp = ERROR_PAIRING_INSTANCE_WITH_ITSELF)
    public void testPairingShouldFailInCasePairingTokenAudienceIsThisInstance() {
        ServiceTrustServiceImpl service = spy(ServiceTrustServiceImpl.class);
        AccessService accessService = mock(AccessService.class);
        ServiceId serviceId = new ServiceId("jfrt", "th1sServ2ce1d");
        when(accessService.getArtifactoryServiceId()).thenReturn(serviceId);
        service.setAccessService(accessService);

        var jwt = mock(JwtAccessToken.class);
        when(jwt.getExpirationTime()).thenReturn(System.currentTimeMillis() + 10000);
        List<String> audiences = List.of(accessService.getArtifactoryServiceId().getFormattedName());
        when(jwt.getAudience()).thenReturn(audiences);

        doReturn(jwt).when(service).parseTokenValue(anyString());

        service.pairToken(ServiceTrustUseCase.FEDERATED_REPO, "token");
    }

    @Test(dataProvider = "validUseCaseProvider")
    public void findUsecaseSucessTest(String usecase) {
        ServiceTrustService service = new ServiceTrustServiceImpl();
        Optional<ServiceTrustUseCase> useCase = service.findUseCase(usecase);
        Assert.assertNotNull(useCase);
    }

    @Test(dataProvider = "invalidUsecaseProvider",
            expectedExceptions = {ServiceTrustException.class})
    public void findUsecaseFailTest(String usecase) {
        ServiceTrustService service = new ServiceTrustServiceImpl();
        service.findUseCase(usecase);

    }

    @DataProvider
    private static Object[][] validUseCaseProvider() {
        return new Object[][] {
                {"federated-repo/fedrepo-1"},
                {"federated-repo/*"},
                {"internal:federated-repo/*"},
                {"artifact:*:r"}
        };
    }

    @DataProvider
    private static Object[][] invalidUsecaseProvider() {
        return new Object[][] {
                {"federated-repo/fed-repo-1 "},
                {"  federated-repo/fed-repo-1"},
                {"federated-repo/  fed-repo-1"},
                {"   federated-repo/*"},
        };
    }


}
