package org.artifactory.config;

import com.google.common.collect.Maps;
import lombok.extern.slf4j.Slf4j;

import java.util.HashMap;
import java.util.Map;
import java.util.stream.Collectors;

@Slf4j
public class AccessConfigHolder {

    public static final String ANONYMOUS_ACCESS_ENABLED = "anonymous-access-enabled";
    public static final String ATTEMPTS = "attempts";
    public static final String PASSWORD_EXPIRY_DAYS = "password-expiry-days";
    public static final String SECURITY = "security";
    public static final String USER_LOCK_POLICY = "user-lock-policy";
    private final Object lockObject = new Object();
    private final Map<String, Object> configMap = new HashMap<>();

    private AccessConfigHolder() {
    }

    public Map<String, Object> drainConfig() {
        synchronized (lockObject) {
            Map<String, Object> mapCopy = configMap.entrySet().stream()
                    .collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue));
            configMap.clear();
            return mapCopy;
        }
    }

    public void addLoginAttempts(int loginAttempts) {
        addUserLockPolicy(ATTEMPTS, loginAttempts);
    }

    private static final class InstanceHolder {
        private static final AccessConfigHolder INSTANCE = new AccessConfigHolder();
    }

    public static AccessConfigHolder getInstance() {
        return InstanceHolder.INSTANCE;
    }

    public boolean isEmpty() {
        return configMap.isEmpty();
    }

    public void addPasswordExpiryDays(int passwordExpiryDays) {
        addUserLockPolicy(PASSWORD_EXPIRY_DAYS, passwordExpiryDays);
    }

    public void addAnonAccessEnabled(Boolean anonAccessEnabled) {
        synchronized (lockObject) {
            addSecurityMap(ANONYMOUS_ACCESS_ENABLED, anonAccessEnabled, configMap);
        }
    }

    private void addUserLockPolicy(String key, Integer value) {
        synchronized (lockObject) {
            addUserLockPolicyMap(key, value, configMap);
        }
    }

    public static void addSecurityMap(String key, Object value, Map<String, Object> map) {
        Map<String, Object> securityMap;
        if (map.containsKey(SECURITY)) {
            securityMap = (Map<String, Object>) map.get(SECURITY);
        } else {
            securityMap = new HashMap<>();
        }
        securityMap.put(key, value);
        map.put(SECURITY, securityMap);
    }

    public static void addUserLockPolicyMap(String key, Integer value, Map<String, Object> map) {
        Map<String, Object> userLockPolicy;
        if (map.containsKey(SECURITY)) {
            Map<String, Object> securityMap = (Map<String, Object>) map.get(SECURITY);
            if (securityMap.containsKey(USER_LOCK_POLICY)) {
                userLockPolicy = (Map<String, Object>) securityMap.get(USER_LOCK_POLICY);
                userLockPolicy.put(key, value);
            } else {
                userLockPolicy = Maps.newHashMap();
                userLockPolicy.put(key, value);
                addSecurityMap(USER_LOCK_POLICY, userLockPolicy, map);
            }
        } else {
            userLockPolicy = Maps.newHashMap();
            userLockPolicy.put(key, value);
            addSecurityMap(USER_LOCK_POLICY, userLockPolicy, map);
        }
    }

    public static void addLoginAttempts(Integer value, Map<String, Object> map) {
        addUserLockPolicyMap(ATTEMPTS, value, map);
    }

    public static void addAnonAccessEnabled(Boolean anonAccessEnabled, Map<String, Object> map) {
        addSecurityMap(ANONYMOUS_ACCESS_ENABLED, anonAccessEnabled, map);
    }
}