package org.artifactory.core.api.ph.model;

import com.google.common.collect.Maps;
import lombok.EqualsAndHashCode;
import lombok.Getter;
import lombok.Setter;
import org.apache.commons.lang3.StringUtils;
import org.artifactory.core.api.ph.virtual.resolution.PackageRequestContext;

import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;
import javax.servlet.http.HttpServletRequest;
import javax.ws.rs.core.HttpHeaders;
import javax.ws.rs.core.MultivaluedMap;
import java.util.Enumeration;
import java.util.List;
import java.util.Map;

@Immutable
@Getter
@EqualsAndHashCode
public class PackageDownloadContext implements PackageRequestContext {

    /**
     * Repository key to download from
     */
    private final String repoKey;
    /**
     * Path to file
     */
    private final String path;
    /**
     * FOR REMOTE ONLY
     * Set this value when your download url is not relative to the {@link this.path}
     */
    @Nullable
    private final String alternativeUrl;
    /**
     * FOR REMOTE ONLY
     * Set this value when your remote download base URL is not the remote repo base URL
     */
    @Setter
    @Nullable
    private String alternateRemoteBaseUrl;
    /**
     * Set this value when you need to download a specific path from a zip archive
     */
    @Nullable
    private final String zipResourcePath;
    /**
     * Additional headers to the internal download request
     */
    private final Map<String, String> requestHeaders = Maps.newHashMap();

    /**
     * Headers that can be delegated to the remote host.
     *
     * This is white-list filtering that used for preventing from
     * unauthorized headers leaking to the external request
     * towards destination server.
     */
    private final Map<String, String> delegationAllowedHeaders = Maps.newHashMap();
    /**
     * Will replace the HEAD request in RetrieveInfo with a GET request
     */
    @Setter
    private boolean replaceHeadRequestWithGet = false;
    /**
     * Use it to override the default execution method
     * E.g. use HEAD only
     */
    @Setter
    @Nullable
    private String httpMethod;

    @Setter
    private String servletContextUrl = "";

    /**
     * Set this field when the request is not a download request (e.g findPackageById, getLatestVersion etc..)
     */
    @Setter
    private String requestContextInfo;

    /**
     * Enable\disable redirect on remote downloads (Default is disabled)
     */
    private boolean disableRedirect = true;

    /**
     * Skip URL encoding before execution
     */
    @Setter
    private boolean skipEncoding = false;

    @Setter
    private boolean replicationOriginatedDownloadRequest = false;

    @Getter
    private String originalRepoKey = null; // in case that the original request was to virtual repository

    public PackageDownloadContext(String repoKey, String path) {
        this(repoKey, path, null);
    }

    public PackageDownloadContext(String repoKey, String path, @Nullable String alternativeUrl) {
        this(repoKey, path, alternativeUrl, null);
    }

    public PackageDownloadContext(String repoKey, String path, @Nullable String alternativeUrl,
            @Nullable String zipResourcePath) {
        this.repoKey = repoKey;
        this.path = path;
        this.alternativeUrl = alternativeUrl;
        this.zipResourcePath = zipResourcePath;
    }

    /**
     * Add additional headers to the internal download request
     */
    public PackageDownloadContext(PackageRequestContext packageRequestContext) {
        this(packageRequestContext.getRepoKey(), packageRequestContext.getPath(),
                packageRequestContext.getAlternativeUrl());
        servletContextUrl = packageRequestContext.getServletContextUrl();
        requestHeaders.putAll(packageRequestContext.getRequestHeaders());
    }

    public PackageDownloadContext originalRepoKey(String virtualRepoKey) {
        this.originalRepoKey = virtualRepoKey;
        return this;
    }

    public PackageDownloadContext header(String key, String value) {
        if (key != null && value != null) {
            this.requestHeaders.put(key, value);
        }
        return this;
    }

    /**
     * Add additional headers to the internal download request
     */
    public PackageDownloadContext headers(HttpHeaders httpHeaders) {
        if (httpHeaders != null) {
            MultivaluedMap<String, String> multiMap = httpHeaders.getRequestHeaders();
            for (Map.Entry<String, List<String>> entry : multiMap.entrySet()) {
                List<String> requestHeader = entry.getValue();
                if (!requestHeader.isEmpty()) {
                    header(entry.getKey(), requestHeader.get(0));
                }
            }
        }
        return this;
    }

    /**
     * Add additional headers that can be delegated to the remote host (In addition to the {@link this.requestHeaders})
     */
    public PackageDownloadContext addHeaderEligibleForDelegation(String key, String value) {
        if (key != null && value != null) {
            this.delegationAllowedHeaders.put(key, value);
        }
        return this;
    }

    @Override
    public String getRequestContextInfo() {
        if (StringUtils.isNotBlank(path)) {
            return path;
        }
        return requestContextInfo;
    }

    public void enableRedirect() {
        this.disableRedirect = false;
    }

    public void extractHeadersFromHttpServletRequest(HttpServletRequest httpServletRequest) {
        if (httpServletRequest != null) {
            Enumeration<String> headerNames = httpServletRequest.getHeaderNames();
            if (headerNames != null) {
                while (headerNames.hasMoreElements()) {
                    String headerName = headerNames.nextElement();
                    header(headerName, httpServletRequest.getHeader(headerName));
                }
            }
        }
    }
}
