package org.artifactory.core.api.ph.service;

import com.google.common.collect.Multimap;
import lombok.NonNull;
import org.artifactory.common.repo.PackageType;
import org.artifactory.common.repo.RepoInfo;
import org.artifactory.common.repo.RepoType;
import org.artifactory.core.api.ph.exception.PackageException;
import org.artifactory.core.api.ph.exception.PackageNotFoundException;
import org.artifactory.core.api.ph.virtual.resolution.PackageRepo;
import org.artifactory.core.api.ph.virtual.resolution.PackageRequestContext;

import javax.annotation.Nullable;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;

public interface PackageRepoService {

    /**
     * checks if repository is virtual
     *
     * @param repoKey repository to check
     * @return true the repository is virtual repository
     */
    boolean isVirtual(String repoKey);

    /**
     * checks if repository is remote
     *
     * @param repoKey repository to check
     * @return true the repository is remote repository
     */
    boolean isRemote(String repoKey);

    /**
     * checks if repository is local
     *
     * @param repoKey repository to check
     * @return true the repository is local repository
     */
    boolean isLocal(String repoKey);

    /**
     * Moves repository path (pointing to a folder) to another absolute target. The move will only move paths the user
     * has permissions to move and paths that are accepted by the target repository.
     *
     * @param sourceRepoKey The non-cached source local repository name.
     * @param targetRepoKey The non-cached target local repository name.
     * @param targetPath    The target path to move the path to.
     * @throws PackageException in that the source path doesn't exist or internal error
     */
    void move(String sourceRepoKey, String targetRepoKey, String sourcePath, String targetPath) throws PackageException;

    /**
     * Resolve the package type from repo key
     *
     * @param repoKey the repository to check
     * @return the above repository package type
     */
    @Nullable
    PackageType getPackageType(String repoKey);

    /**
     * Resolve the repo info package type (cran, maven, debian...) and repo type(local, virtual and remote)
     *
     * @param repoKey the repository to check
     * @return RepoInfo which is pair of package type and repo type
     * @throws PackageException in any case of not found or internal error
     */
    @NonNull
    RepoInfo getRepoInfo(String repoKey) throws PackageException;

    /**
     * Set all attribute key values on the given path.
     * Existing prop-keys on the path will get overwritten.
     * Existing prop-keys on the artifact that not included in the given properties - will remain as they are
     * @param path       The path for which to set the attribute values
     * @param properties Multimap for keys with multiple values
     */
    void setAttributes(String repoKey, String path, Multimap<String, String> properties) throws PackageException;

    /**
     * Get all attributes (properties) on the given path.
     * Existing prop-keys on the path will get overwritten.
     *
     * @param path The path for which to get the attribute values
     * @throws PackageException if the repoKey is not real (local, fed, remote with caching) or does not exist.
     */
    @NonNull
    Map<String, Collection<String>> getAttributes(String repoKey, String path) throws PackageException;

    /**
     * Remove the given values for an attribute on the given path, or the entire property
     * Note: If provided a virtual repository, the default-deployment repo wil be used
     *
     * @param path           the path for which to remove the attribute value
     * @param propKey        prop key to remove value from
     * @param valuesToRemove Set for values to remove, or null in order to remove the entire property
     */
    void removeAttributes(String repoKey, String path, String propKey, @Nullable Set<String> valuesToRemove)
            throws PackageException;

    /**
     * Checks if an artifact exists locally
     *
     * @param repoKey the artifact repo
     * @param path    the artifact path
     * @return true if artifact exist
     */
    boolean existLocally(String repoKey, String path);

    /**
     * Undeploy an artifact
     *
     * @param repoKey the artifact repository key
     * @param path    the artifact path
     */
    void undeployArtifact(String repoKey, String path);

    /**
     * Undeploy multiple artifacts atomically.
     *
     * @param repoKey repo key to undeploy from.
     * @param paths   paths to undeploy.
     */
    void undeployArtifactsAtomically(String repoKey, Set<String> paths) throws PackageException;

    /**
     * Return virtual repository sub repositories, the repositories are sorted according to the virtual repository logic
     * Which is prioritized (local->remote-cache->remote->virtual) prioritized (local->remote-cache->remote->virtual)
     *
     * @param packageRequestContext              the virtual repository to resolve from.
     * @param filterCacheRepositoriesDuplication if true, return a list without cache repositories in case that the
     *                                           corresponding remote repository is part of the result
     * @return List of sorted sub repositories
     * @throws PackageException in any case of not found or internal error
     */
    List<? extends PackageRepo> getSubRepositories(PackageRequestContext packageRequestContext,
            boolean filterCacheRepositoriesDuplication) throws PackageException;

    /**
     * In case of a Virtual repoKey, return the default deployment local repoKey - if there is.
     * In case of an exist local repoKey, return the local repoKey.
     * Any other case, will get an ExposedException.
     *
     * @return A local repoKey.
     * @throws PackageException in any case of not found or internal error
     */
    String getDefaultDeploymentRepoKey(String repoKey) throws PackageException;

    /**
     * This function cleaning the given  cache path of a virtual repository related to the given aggregated local
     * repository.
     * Use this function as a system
     *
     * @param localRepoKey            - repokey of the local repository
     * @param virtualRepoMetadataRoot - cache path to clean
     */
    void undeployVirtualMetadataCache(String localRepoKey, String virtualRepoMetadataRoot);

    /**
     * @return The remote repository base URL.
     * @throws PackageException when repo is not remote or does not exist.
     */
    String getRemoteRepoBaseUrl(String remoteRepoKey) throws PackageException;

    /**
     * @param packageType - package type to filter by
     * @param repoType    - repo type to filter by
     * @return List of packageRepo from the given repoType and packageType
     */
    List<PackageRepo> getAllReposByType(@NonNull PackageType packageType, @NonNull RepoType repoType);

    /**
     * @param repoKey - Remote repo key
     * @return True if the remote repo exist and unavailable for remote network operations.
     * @throws PackageNotFoundException - If a remote repo in this name was not found
     */
    boolean isRemoteRepoOffline(@NonNull String repoKey) throws PackageNotFoundException;

    /**
     * @param remoteRepoKey - Remote repo key
     * @return remote-cache repoKey
     * @throws PackageNotFoundException - If a remote repo in this name was not found or storing local artifacts is
     *                                  disabled.
     */
    @NonNull
    String getRemoteCacheRepoKey(@NonNull String remoteRepoKey) throws PackageNotFoundException;
}