package org.artifactory.core.api.ph.virtual;

import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import org.apache.commons.io.IOUtils;
import org.artifactory.common.resource.BaseResource;
import org.artifactory.confprop.ConfProps;
import org.artifactory.core.api.ph.exception.PackageException;
import org.artifactory.core.api.ph.resource.PackageUnfoundedResource;
import org.artifactory.core.api.ph.service.PackageArtifactService;
import org.artifactory.core.api.ph.service.PackageRepoService;
import org.artifactory.core.api.ph.service.PackageUploadService;
import org.artifactory.core.api.ph.virtual.resolution.PackageMetadataRow;
import org.artifactory.core.api.ph.virtual.resolution.PackageMetadataVersionsSet;
import org.artifactory.core.api.ph.virtual.resolution.PackageRepo;
import org.artifactory.core.api.ph.virtual.resolution.PackageRequestContext;
import org.artifactory.core.api.ph.virtual.resolution.PackageVirtualController;
import org.artifactory.core.api.ph.virtual.resolution.VirtualResolutionStrategyType;
import org.jetbrains.annotations.NotNull;
import org.jfrog.common.concurrency.ConflictsGuard;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.testng.annotations.BeforeMethod;

import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.concurrent.TimeUnit;
import java.util.function.BiFunction;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyLong;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.CALLS_REAL_METHODS;
import static org.mockito.MockitoAnnotations.openMocks;

public abstract class VirtualRepoTraversalBaseTest {

    protected static final String virtualRepoKey = "my_repo";
    protected static final String localRepoKey1 = "local_repo1";
    protected static final String localRepoKey2 = "local_repo2";
    protected static final String path = "a/b/c.test";
    protected static final String path1 = "a/b/c1.test";
    protected static final String path2 = "a/b/c2.test";

    @Mock
    protected ConflictsGuard<String> guard;

    @Mock
    protected PackageArtifactService downloadService;
    @Mock
    protected PackageUploadService uploadService;
    @Mock
    protected PackageRepoService repoService;
    protected VirtualRepoTraversal virtualRepoTraversal;
    @Mock
    protected ConfProps confProps;

    @BeforeMethod
    public void setup() throws InterruptedException, PackageException {
        openMocks(this);
        Mockito.when(guard.tryToLock(anyString(), anyLong(), any(TimeUnit.class))).thenReturn(true);
        virtualRepoTraversal = new VirtualRepoTraversal(repoService, downloadService, uploadService,
                guard, confProps);
    }

    @NotNull
    protected List<? extends PackageRepo> subRepositories(String... repoKeys) {
        List<PackageRepo> subRepositories = Lists.newArrayList();
        for (String repoKey : repoKeys) {
            subRepositories.add(repository(repoKey, false));
        }
        return subRepositories;
    }

    protected static class TestingPackageMetadataRow implements PackageMetadataRow {
        private final String packageName;

        public TestingPackageMetadataRow(String packageName) {
            this.packageName = packageName;
        }

        @Override
        public String getPackageIdentifier() {
            return packageName;
        }

        @Override
        public boolean equals(Object other) {
            if (this == other) { return true; }
            if (other == null || getClass() != other.getClass()) { return false; }
            TestingPackageMetadataRow that = (TestingPackageMetadataRow) other;
            return this.getPackageIdentifier().equals(that.getPackageIdentifier());
        }

        @Override
        public int hashCode() {
            return Objects.hash(packageName);
        }

        public String getPackageName() {
            return packageName;
        }
    }

    protected static abstract class BaseResourceForTesting implements BaseResource {
        public abstract InputStream getInputStream();
    }

    @SuppressWarnings("rawtypes")
    protected PackageVirtualController controller(VirtualResolutionStrategyType strategy) throws PackageException {
        abstract class PackageVirtualControllerImpl implements PackageVirtualController {
        }

        PackageVirtualController<?> controller = Mockito.mock(PackageVirtualControllerImpl.class, CALLS_REAL_METHODS);
        Mockito.when(controller.getVirtualResolutionStrategy(any())).thenReturn(strategy);
        Mockito.doAnswer(invocation -> {
            BaseResource res = invocation.getArgument(0, BaseResource.class);
            Map<String, List<PackageMetadataRow>> map = Maps.newHashMap();
            String data = "[" + res.getRepoKey() + "/" + res.getPath() + "]";
            List<PackageMetadataRow> list = Lists.newArrayList(new TestingPackageMetadataRow(data));
            map.put(data, list);
            return map;
        }).when(controller).extract(any(), any(), any(), any());
        Mockito.doAnswer(invocation -> {
            StringBuilder stringBuilder = new StringBuilder();
            Map<String, PackageMetadataVersionsSet<TestingPackageMetadataRow>> metadata = invocation.getArgument(0);
            for (PackageMetadataVersionsSet<TestingPackageMetadataRow> value : metadata.values()) {
                value.getVersions().forEach(row -> stringBuilder.append(row.getPackageName()));
            }
            return IOUtils.toInputStream(stringBuilder, StandardCharsets.UTF_8);
        }).when(controller).toInputStream(any(), any());
        return controller;
    }

    protected PackageRepo repository(String repoKey, boolean priority) {
        PackageRepo repo = Mockito.mock(PackageRepo.class);
        Mockito.when(repo.getKey()).thenReturn(repoKey);
        Mockito.when(repo.getFileContent(anyString())).thenReturn(null);
        Mockito.when(repo.isLocal()).thenReturn(true);
        Mockito.when(repo.isPriorityResolution()).thenReturn(priority);
        return repo;
    }

    protected BaseResource resource(String repoKey, String path, InputStream inputStream) {
        BaseResourceForTesting resource = Mockito.mock(BaseResourceForTesting.class);
        Mockito.when(resource.getRepoKey()).thenReturn(repoKey);
        Mockito.when(resource.getPath()).thenReturn(path);
        Mockito.when(resource.isFound()).thenReturn(true);
        Mockito.when(resource.getLastModified()).thenReturn(0L);
        Mockito.when(resource.isExactQueryMatch()).thenReturn(true);
        Mockito.when(resource.isExpired()).thenReturn(false);
        Mockito.when(resource.isBlocked()).thenReturn(false);
        Mockito.when(resource.getInputStream()).thenReturn(inputStream);
        return resource;
    }

    protected BiFunction<PackageRequestContext, PackageRepo, BaseResource> getResourceFunc(
            PackageVirtualController<?> controller) {
        return (packageRequestContext, packageRepo) -> {
            PackageRequestContext translated = null;
            try {
                translated = controller.translateSubRepositoryContext(packageRepo, packageRequestContext);
                return downloadService.getResource(translated);
            } catch (PackageException e) {
                if (translated == null) {
                    translated = packageRequestContext;
                }
                return new PackageUnfoundedResource(translated.getRepoKey(), translated.getPath(),
                        e.getStatus(),
                        e.getPublicReason());
            }
        };
    }
}
