package org.artifactory.core.api.ph.virtual;

import org.apache.commons.io.IOUtils;
import org.artifactory.common.repo.PackageType;
import org.artifactory.common.resource.BaseResource;
import org.artifactory.core.api.ph.exception.PackageException;
import org.artifactory.core.api.ph.model.PackageDownloadContext;
import org.artifactory.core.api.ph.virtual.resolution.PackageRepo;
import org.artifactory.core.api.ph.virtual.resolution.PackageRequestContext;
import org.artifactory.core.api.ph.virtual.resolution.PackageVirtualController;
import org.artifactory.core.api.ph.virtual.resolution.VirtualResolutionStrategyType;
import org.junit.jupiter.api.Assertions;
import org.mockito.ArgumentCaptor;
import org.mockito.stubbing.Answer;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.util.List;
import java.util.Map;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyInt;
import static org.mockito.ArgumentMatchers.anyLong;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.ArgumentMatchers.argThat;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.ArgumentMatchers.refEq;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.never;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

@Test
public class VirtualRepoTraversalMergerTest extends VirtualRepoTraversalBaseTest {
    private static final String metadata = "a/b/metadata.test";

    /**
     * Test virtual repo resolution
     * 1. upload two different metadata files into two different repositories that belongs to virtual repo
     * 2. try to download metadata and expect to receive merged file.
     */
    @Test
    public void testMerge() throws PackageException, IOException {
        List<? extends PackageRepo> subRepos = subRepositories(localRepoKey1, localRepoKey2);
        doAnswer(invocation -> subRepos).when(repoService).getSubRepositories(any(), eq(true));
        BaseResource result1 = resource(localRepoKey1, metadata, null);
        BaseResource result2 = resource(localRepoKey2, metadata, null);
        when(downloadService.getResource(refEq(new PackageDownloadContext(localRepoKey1, metadata))))
                .thenReturn(result1);
        when(downloadService.getResource(refEq(new PackageDownloadContext(localRepoKey2, metadata))))
                .thenReturn(result2);

        doAnswer((Answer<?>) invocationOnMock -> {
            String repoKey = invocationOnMock.getArgument(0);
            String resourcePath = invocationOnMock.getArgument(1);
            InputStream inputStream = invocationOnMock.getArgument(2);
            return resource(repoKey, resourcePath, inputStream);
        }).when(uploadService).uploadToVirtualOrRemoteCache(any(), any(), any());

        PackageVirtualController<?> controller = controller(VirtualResolutionStrategyType.MERGER);
        BaseResource resource = virtualRepoTraversal.resolve(new PackageDownloadContext(virtualRepoKey, metadata)
                , controller, getResourceFunc(controller));
        Assertions.assertEquals(virtualRepoKey, resource.getRepoKey());
        String data = IOUtils.toString(((BaseResourceForTesting) resource).getInputStream(), StandardCharsets.UTF_8);
        Assertions.assertTrue(data.contains("[local_repo2/a/b/metadata.test]"));
        Assertions.assertTrue(data.contains("[local_repo1/a/b/metadata.test]"));
    }

    /**
     * Test virtual repo resolution
     * 1. upload one package into repos that belongs to virtual repo
     * 2. try to download metadata and expect to receive merged file with one package.
     */
    @Test
    public void testMergeSingleFile() throws PackageException, IOException {
        List<? extends PackageRepo> subRepos = subRepositories(localRepoKey1, localRepoKey2);
        doAnswer(invocation -> subRepos).when(repoService).getSubRepositories(any(), eq(true));
        BaseResource result1 = resource(localRepoKey1, metadata, null);
        when(downloadService.getResource(refEq(new PackageDownloadContext(localRepoKey1, metadata))))
                .thenReturn(result1);

        doAnswer((Answer<?>) invocationOnMock -> {
            String repoKey = invocationOnMock.getArgument(0);
            String resourcePath = invocationOnMock.getArgument(1);
            InputStream inputStream = invocationOnMock.getArgument(2);
            return resource(repoKey, resourcePath, inputStream);
        }).when(uploadService).uploadToVirtualOrRemoteCache(any(), any(), any());

        PackageVirtualController<?> controller = controller(VirtualResolutionStrategyType.MERGER);
        BaseResource resource = virtualRepoTraversal.resolve(new PackageDownloadContext(virtualRepoKey, metadata)
                , controller, getResourceFunc(controller));
        Assertions.assertEquals(virtualRepoKey, resource.getRepoKey());
        String data = IOUtils.toString(((BaseResourceForTesting) resource).getInputStream(), StandardCharsets.UTF_8);
        Assertions.assertFalse(data.contains("[local_repo2/a/b/metadata.test]"));
        Assertions.assertTrue(data.contains("[local_repo1/a/b/metadata.test]"));
    }


    /**
     * Test virtual repo merging
     * 1. Make sure repositories are empty
     * 2. try to download metadata and expect and expect not found
     */
    @Test
    public void testMergeNotExist() throws PackageException {
        List<? extends PackageRepo> subRepos = subRepositories(localRepoKey1, localRepoKey2);
        doAnswer(invocation -> subRepos).when(repoService).getSubRepositories(any(), eq(true));
        doAnswer((Answer<?>) invocationOnMock -> {
            String repoKey = invocationOnMock.getArgument(0);
            String resourcePath = invocationOnMock.getArgument(1);
            InputStream inputStream = invocationOnMock.getArgument(2);
            return resource(repoKey, resourcePath, inputStream);
        }).when(uploadService).uploadToVirtualOrRemoteCache(any(), any(), any());

        PackageVirtualController<?> controller = controller(VirtualResolutionStrategyType.MERGER);
        BaseResource resource = virtualRepoTraversal.resolve(new PackageDownloadContext(virtualRepoKey, metadata)
                , controller, getResourceFunc(controller));
        Assertions.assertEquals(virtualRepoKey, resource.getRepoKey());
        Assertions.assertTrue(resource.isFound());
    }

    /**
     * The merger when cache resource is available, the cache entry should return, no merging is done.
     */
    @Test
    public void testMergerVirtualCachedAvailable() throws PackageException {
        List<? extends PackageRepo> subRepos = subRepositories(localRepoKey1, localRepoKey2);
        doAnswer(invocation -> subRepos).when(repoService).getSubRepositories(any(), eq(true));
        BaseResource result1 = resource(localRepoKey1, path1, null);
        BaseResource result2 = resource(localRepoKey2, path2, null);
        when(downloadService.getResource(localRepoKey1, path)).thenReturn(result1);
        when(downloadService.getResource(refEq(new PackageDownloadContext(localRepoKey1, path)))).thenReturn(result1);
        when(downloadService.getResource(refEq(new PackageDownloadContext(localRepoKey2, path)))).thenReturn(result2);
        BaseResource availableCachedResource = resource(virtualRepoKey, path, null);
        when(downloadService.getResourceFromVirtualCache(
                eq(virtualRepoKey), eq(path))).thenReturn(
                availableCachedResource);
        PackageVirtualController<?> controller = controller(VirtualResolutionStrategyType.MERGER);
        PackageDownloadContext packageDownloadContext = new PackageDownloadContext(virtualRepoKey, path);
        BaseResource resource =
                virtualRepoTraversal.resolve(packageDownloadContext, controller,
                        getResourceFunc(controller));
        Assertions.assertEquals(virtualRepoKey, resource.getRepoKey());
        Assertions.assertEquals(path, resource.getPath());
        verify(repoService, never()).getSubRepositories(refEq(packageDownloadContext), eq(true));
    }

    @DataProvider
    public Object[][] testMergerVirtualCachedNotAvailableDataProvider() {
        //on of found, expired, block

        BaseResource notFound = resource(virtualRepoKey, path, null);
        when(notFound.isFound()).thenReturn(false);
        when(notFound.isExpired()).thenReturn(false);
        when(notFound.isBlocked()).thenReturn(false);

        BaseResource expired = resource(virtualRepoKey, path, null);
        when(expired.isFound()).thenReturn(true);
        when(expired.isExpired()).thenReturn(true);
        when(expired.isBlocked()).thenReturn(false);

        BaseResource blocked = resource(virtualRepoKey, path, null);
        when(blocked.isFound()).thenReturn(true);
        when(blocked.isExpired()).thenReturn(false);
        when(blocked.isBlocked()).thenReturn(true);

        return new Object[][] {{null}, {notFound}, {expired}, {blocked}};
    }


    /**
     * Test a merge when cache entry is not available
     */
    @Test(dataProvider = "testMergerVirtualCachedNotAvailableDataProvider")
    public void testMergerVirtualCachedNotAvailable(BaseResource unAvailable) throws PackageException {
        List<? extends PackageRepo> subRepos = subRepositories(localRepoKey1, localRepoKey2);
        doAnswer(invocation -> subRepos).when(repoService).getSubRepositories(any(), eq(true));
        BaseResource result1 = resource(localRepoKey1, path, null);
        BaseResource result2 = resource(localRepoKey2, path, null);
        when(downloadService.getResource(refEq(new PackageDownloadContext(localRepoKey1, path)))).thenReturn(result1);
        when(downloadService.getResource(refEq(new PackageDownloadContext(localRepoKey2, path)))).thenReturn(result2);

        when(downloadService.getResourceFromVirtualCache(
                eq(virtualRepoKey), eq(path))).thenReturn(unAvailable);

        BaseResource uploaded = resource(virtualRepoKey, path, null);
        when(uploaded.isFound()).thenReturn(true);
        when(uploaded.isExpired()).thenReturn(false);
        when(uploaded.isBlocked()).thenReturn(false);

        when(uploadService.uploadToVirtualOrRemoteCache(eq(virtualRepoKey), eq(path), any())).thenReturn(uploaded);
        PackageVirtualController<?> controller = controller(VirtualResolutionStrategyType.MERGER);

        PackageDownloadContext packageDownloadContext = new PackageDownloadContext(virtualRepoKey, path);
        BaseResource resource =
                virtualRepoTraversal.resolve(packageDownloadContext, controller,
                        getResourceFunc(controller));

        Assertions.assertEquals(virtualRepoKey, resource.getRepoKey());
        Assertions.assertEquals(path, resource.getPath());

        verify(repoService, times(1)).getSubRepositories(refEq(packageDownloadContext), eq(true));
        verify(uploadService, times(1)).uploadToVirtualOrRemoteCache(eq(virtualRepoKey),
                eq(path), any());

        Assertions.assertEquals(uploaded, resource);
    }

    /**
     * Test that when a searchable resource does not found, we clean the virtual cache
     */
    @Test
    public void testMergerVirtualCachedNotAvailableAndSearchableNotFoundCleanCache()
            throws PackageException {
        List<? extends PackageRepo> subRepos = subRepositories(localRepoKey1, localRepoKey2);
        doAnswer(invocation -> subRepos).when(repoService).getSubRepositories(any(), eq(true));
        BaseResource result1 = resource(localRepoKey1, path, null);
        BaseResource result2 = resource(localRepoKey2, path, null);
        when(downloadService.getResource(refEq(new PackageDownloadContext(localRepoKey1, path)))).thenReturn(result1);
        when(downloadService.getResource(refEq(new PackageDownloadContext(localRepoKey2, path)))).thenReturn(result2);

        BaseResource unAvailable = resource(virtualRepoKey, path, null);
        when(unAvailable.isFound()).thenReturn(false);
        when(unAvailable.isExpired()).thenReturn(true);
        when(unAvailable.isBlocked()).thenReturn(true);
        when(downloadService.getResourceFromVirtualCache(
                eq(virtualRepoKey), eq(path))).thenReturn(unAvailable);

        BaseResource searchableResource = resource(virtualRepoKey, path, null);
        when(searchableResource.isFound()).thenReturn(false);
        when(uploadService.uploadToVirtualOrRemoteCache(eq(virtualRepoKey), eq(path), any())).thenReturn(
                searchableResource);

        PackageVirtualController<?> controller = controller(VirtualResolutionStrategyType.MERGER);

        PackageDownloadContext packageDownloadContext = new PackageDownloadContext(virtualRepoKey, path);
        BaseResource resource =
                virtualRepoTraversal.resolve(packageDownloadContext, controller,
                        getResourceFunc(controller));

        Assertions.assertEquals(virtualRepoKey, resource.getRepoKey());
        Assertions.assertEquals(path, resource.getPath());

        verify(repoService, times(1)).getSubRepositories(refEq(packageDownloadContext), eq(true));
        verify(uploadService, times(1)).uploadToVirtualOrRemoteCache(eq(virtualRepoKey), eq(path), any());
        Assertions.assertEquals(searchableResource, resource);
        verify(repoService, times(1)).undeployArtifact(eq(virtualRepoKey), eq(path));
    }


    /**
     * Test that the override (translated) path is used for accessing the virtual cache
     */
    @Test
    public void testMergerVirtualCachedNotAvailableOverridePath() throws PackageException {
        List<? extends PackageRepo> subRepos = subRepositories(localRepoKey1, localRepoKey2);
        doAnswer(invocation -> subRepos).when(repoService).getSubRepositories(any(), eq(true));
        BaseResource result1 = resource(localRepoKey1, path, null);
        BaseResource result2 = resource(localRepoKey2, path, null);
        when(downloadService.getResource(refEq(new PackageDownloadContext(localRepoKey1, path)))).thenReturn(result1);
        when(downloadService.getResource(refEq(new PackageDownloadContext(localRepoKey2, path)))).thenReturn(result2);

        BaseResource unAvailable = resource(virtualRepoKey, path, null);
        when(unAvailable.isExpired()).thenReturn(true);
        when(downloadService.getResourceFromVirtualCache(
                eq(virtualRepoKey), eq("override/" + path))).thenReturn(
                unAvailable);

        BaseResource uploaded = resource(virtualRepoKey, path, null);
        when(uploaded.isFound()).thenReturn(true);
        when(uploaded.isExpired()).thenReturn(false);
        when(uploaded.isBlocked()).thenReturn(false);

        when(uploadService.uploadToVirtualOrRemoteCache(
                eq(virtualRepoKey), eq("override/" + path), any())).thenReturn(
                uploaded);

        PackageDownloadContext request = new PackageDownloadContext(virtualRepoKey, path);

        PackageVirtualController<?> controller = controller(VirtualResolutionStrategyType.MERGER);
        PackageRequestContext
                overrideReqContext = new PackageDownloadContext(virtualRepoKey, "override/" + path);
        when(controller.getCachePath(any())).thenReturn(overrideReqContext);

        BaseResource resource = virtualRepoTraversal.resolve(request, controller, getResourceFunc(controller));

        Assertions.assertEquals(virtualRepoKey, resource.getRepoKey());
        Assertions.assertEquals(path, resource.getPath());

        verify(controller, times(1)).toInputStream(any(), eq(request));
        verify(downloadService, times(2)).getResourceFromVirtualCache(eq(virtualRepoKey), eq("override/" + path));
        verify(repoService, times(1)).getSubRepositories(
                argThat(packageRequestContext -> packageRequestContext.getRepoKey().equals(virtualRepoKey)), eq(true));
        verify(uploadService, times(1)).uploadToVirtualOrRemoteCache(eq(virtualRepoKey), eq("override/" + path), any());
        verify(uploadService, times(1)).uploadToVirtualOrRemoteCache(eq(virtualRepoKey), eq("override/" + path), any());

        Assertions.assertEquals(uploaded, resource);
    }

    /**
     * Test the case that merge lock was not acquired, in that case we should return what we have in the cache.
     */
    @Test
    public void testMergerLockNotAcquired() throws PackageException, InterruptedException {
        when(guard.tryToLock(anyString(), anyLong(), any())).thenReturn(false);
        List<? extends PackageRepo> subRepos = subRepositories(localRepoKey1, localRepoKey2);
        doAnswer(invocation -> subRepos).when(repoService).getSubRepositories(any(), eq(true));
        BaseResource result1 = resource(localRepoKey1, path, null);
        when(downloadService.getResource(refEq(new PackageDownloadContext(localRepoKey1, path)))).thenReturn(result1);

        BaseResource unAvailable = resource(virtualRepoKey, path, null);
        when(unAvailable.isExpired()).thenReturn(true);
        when(downloadService.getResourceFromVirtualCache(
                eq(virtualRepoKey), eq(path))).thenReturn(unAvailable);

        PackageVirtualController<?> controller = controller(VirtualResolutionStrategyType.MERGER);
        PackageRequestContext overrideContext = new PackageDownloadContext(virtualRepoKey, path);
        when(controller.getCachePath(any(PackageRequestContext.class))).thenReturn(overrideContext);
        PackageDownloadContext packageDownloadContext = new PackageDownloadContext(virtualRepoKey, path);
        BaseResource resource =
                virtualRepoTraversal.resolve(packageDownloadContext, controller,
                        getResourceFunc(controller));

        Assertions.assertEquals(virtualRepoKey, resource.getRepoKey());
        Assertions.assertEquals(path, resource.getPath());

        verify(downloadService, times(2)).getResourceFromVirtualCache(eq(virtualRepoKey), eq(path));
        verify(repoService, times(1)).getSubRepositories(refEq(packageDownloadContext), eq(true));
        verify(uploadService, never()).uploadToVirtualOrRemoteCache(eq(virtualRepoKey), eq(path), any());
        verify(uploadService, never()).uploadToVirtualOrRemoteCache(eq(virtualRepoKey), eq(path), any());
        Assertions.assertEquals(unAvailable, resource);
    }

    @DataProvider
    public Object[][] testMergerResolveLockTimeoutDataProvider() {
        return new Object[][] {
                {Map.of(), VirtualRepoTraversal.DEFAULT_MERGE_LOCK_TIMEOUT_SEC},
                {Map.of(VirtualRepoTraversal.MERGE_LOCK_TIMEOUT_PROP_GLOBAL_KEY, 1), 1},
                {Map.of(VirtualRepoTraversal.MERGE_LOCK_TIMEOUT_PROP_GLOBAL_KEY + ".CRAN", 2), 2},
                {Map.of(VirtualRepoTraversal.MERGE_LOCK_TIMEOUT_PROP_GLOBAL_KEY + ".CRAN." + virtualRepoKey, 3), 3}
        };
    }

    /**
     * Test Resolving the merge lock timeout
     * The order is:
     * repo prop timeout
     * package prop timeout
     * global prop timeout
     * default hard coded
     */
    @Test(dataProvider = "testMergerResolveLockTimeoutDataProvider")
    public void testMergerResolveLockTimeout(Map<String, Integer> props, int expectedTimeout)
            throws PackageException, InterruptedException {
        doAnswer(invocation -> {
            Integer value = props.get(invocation.getArgument(0, String.class));
            if (value == null) {
                value = invocation.getArgument(1, Integer.class);
            }
            return value;
        }).when(confProps).getIntValue(anyString(), any());

        when(guard.tryToLock(anyString(), anyInt(), any())).thenReturn(true);
        when(repoService.getPackageType(eq(virtualRepoKey))).thenReturn(PackageType.CRAN);

        List<? extends PackageRepo> subRepos = subRepositories(localRepoKey1, localRepoKey2);
        doAnswer(invocation -> subRepos).when(repoService).getSubRepositories(any(), eq(true));
        BaseResource result1 = resource(localRepoKey1, path, null);
        when(downloadService.getResource(refEq(new PackageDownloadContext(localRepoKey1, path)))).thenReturn(result1);

        BaseResource unAvailable = resource(virtualRepoKey, path, null);
        when(unAvailable.isExpired()).thenReturn(true);
        when(downloadService.getResourceFromVirtualCache(
                eq(virtualRepoKey), eq(path))).thenReturn(unAvailable);

        BaseResource uploaded = resource(virtualRepoKey, path, null);
        when(uploadService.uploadToVirtualOrRemoteCache(eq(virtualRepoKey), eq(path), any())).thenReturn(uploaded);

        PackageVirtualController<?> controller = controller(VirtualResolutionStrategyType.MERGER);
        PackageDownloadContext packageDownloadContext = new PackageDownloadContext(virtualRepoKey, path);
        BaseResource resource =
                virtualRepoTraversal.resolve(packageDownloadContext, controller,
                        getResourceFunc(controller));

        Assertions.assertEquals(virtualRepoKey, resource.getRepoKey());
        Assertions.assertEquals(path, resource.getPath());

        verify(downloadService, times(2)).getResourceFromVirtualCache(eq(virtualRepoKey), eq(path));
        verify(repoService, times(1)).getSubRepositories(refEq(packageDownloadContext), eq(true));
        verify(uploadService, times(1)).uploadToVirtualOrRemoteCache(eq(virtualRepoKey), eq(path), any());
        Assertions.assertEquals(uploaded, resource);

        ArgumentCaptor<Long> actualTimeout = ArgumentCaptor.forClass(Long.class);
        verify(guard, times(1)).tryToLock(eq(virtualRepoKey + "/" + path), actualTimeout.capture(), any());
        Assertions.assertEquals(expectedTimeout, actualTimeout.getValue().intValue());
    }

}
