package org.artifactory.webapp.main;

import com.google.common.base.Charsets;
import com.jfrog.sysconf.test.SystemYamlBuilder;
import com.jfrog.sysconf.test.YamlNodeProps;
import com.jfrog.sysconf.test.YamlRouterEntrypoints;
import com.jfrog.sysconf.test.YamlServiceSysConfig;
import com.jfrog.sysconf.test.YamlSysConfig;
import org.apache.catalina.startup.Tomcat;
import org.apache.commons.io.FileUtils;
import org.artifactory.webapp.WebappUtils;
import org.jfrog.common.YamlUtils;

import java.io.File;
import java.io.IOException;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;
import java.util.UUID;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.ThreadPoolExecutor;

import static org.artifactory.webapp.WebappUtils.copyNewerDevResources;
import static org.artifactory.webapp.WebappUtils.updateMimetypes;
import static org.artifactory.webapp.main.StartArtifactoryDev.setSystemProperties;
import static org.artifactory.webapp.main.StartArtifactoryDev.startAccessProcess;
import static org.artifactory.webapp.main.StartArtifactoryDev.startFrontendProcess;
import static org.artifactory.webapp.main.StartArtifactoryDev.startMetadataProcess;
import static org.artifactory.webapp.main.StartArtifactoryDev.startRouterProcess;
import static org.jfrog.config.home.JFrogHome.ETC_FOLDER;
import static org.jfrog.config.home.JFrogHome.SYSTEM_YAML_FILENAME;
import static org.jfrog.config.home.JFrogHome.VAR_FOLDER;

/**
 * @author Tamir Hadad
 */
public class StartArtifactoryDevWithServices {

    public static void main(String... args) throws Exception {
        File devArtHome = StartArtifactoryDev.getArtifactoryDevHome(args);
        File devEtcDir = WebappUtils.getDevEtcFolder();
        updateMimetypes(devEtcDir);
        copyNewerDevResources(devEtcDir, devArtHome, true);
        setSystemProperties();
        createDevSystemYaml(devArtHome);
        ThreadPoolExecutor executor = (ThreadPoolExecutor) Executors.newFixedThreadPool(2);
        List<Future<File>> resultList = new ArrayList<>();
        resultList.add((Future<File>) executor.submit(() -> startFrontendProcess(devArtHome)));
        resultList.add((Future<File>) executor.submit(() -> startMetadataProcess(devArtHome)));
        AccessProcess accessProcess = startAccessProcess(8040, false, devArtHome, 8046);
        startRouterProcess(accessProcess, devArtHome);
        Tomcat tomcat = StartArtifactoryDev.startTomcatWithRouter(devArtHome);
        for (Future<File> future : resultList) {
            future.get();
        }
        executor.shutdown();
        tomcat.getServer().await();
    }

    @SuppressWarnings("ConstantConditions")
    static void createDevSystemYaml(File home) {
        File systemYamlFile =
                new File(Paths.get(home.getAbsolutePath(), VAR_FOLDER, ETC_FOLDER, SYSTEM_YAML_FILENAME).toUri());
        try {
            // preserve old node.id if present
            YamlServiceSysConfig.Builder shared;
            if (systemYamlFile.exists()) {
                YamlSysConfig yamlSysConfig = YamlUtils.getInstance().readValue(systemYamlFile, YamlSysConfig.class);
                shared = yamlSysConfig.getShared() != null ? yamlSysConfig.getShared().toBuilder() :
                        YamlServiceSysConfig.builder();
            } else {
                shared = YamlServiceSysConfig.builder()
                        .node(YamlNodeProps.builder().id(UUID.randomUUID().toString()).build());
            }
            YamlServiceSysConfig router = YamlServiceSysConfig.builder()
                    .enabled(true)
                    .entrypoints(YamlRouterEntrypoints.builder().internalPort(8046).build())
                    .build();
            YamlSysConfig systemYamlConfig = YamlSysConfig.builder()
                    .shared(shared.build())
                    .router(router)
                    .build();
            FileUtils.deleteQuietly(systemYamlFile);
            FileUtils.writeStringToFile(systemYamlFile, SystemYamlBuilder.toYaml(systemYamlConfig), Charsets.UTF_8,
                    false);
        } catch (IOException e) {
            throw new IllegalStateException("Unable to create/append to system.yaml.", e);
        }
    }
}
