package org.artifactory.webapp.servlet.model;

import org.artifactory.authentication.ArtifactoryAuthenticationContextFilter;
import org.artifactory.authentication.Context.AuthenticationContext;
import org.artifactory.model.xstream.security.UserImpl;
import org.artifactory.ph.PackageAuthenticationFilterImpl;
import org.artifactory.security.SimpleUser;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mockito;
import org.springframework.security.core.Authentication;

import javax.servlet.FilterChain;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.when;
import static org.mockito.Mockito.times;


/**
 * @author shahafg
 */
class ArtifactoryFilterContextHolderImplTest {

    private static final String DUMMY_CACHE_KEY = "Dummy cache key";
    private static final String DUMMY_LOGIN_IDENTIFIER = "Dummy login Identifier";

    private ArtifactoryFilterContextHolderImpl filterHolder;
    private ArtifactoryAuthenticationContextFilter filter;
    private AuthenticationContext authenticationContext;

    @BeforeEach
    void setup() {
        filter = Mockito.mock(PackageAuthenticationFilterImpl.class);
        authenticationContext =
                AuthenticationContext.builder().authenticated(true).cacheKey(DUMMY_CACHE_KEY)
                        .extractedAuthorizationToken(DUMMY_CACHE_KEY)
                        .loginIdentifier(DUMMY_LOGIN_IDENTIFIER)
                        .build();
        filterHolder = new ArtifactoryFilterContextHolderImpl(filter, authenticationContext);
    }

    @Test
    void getCacheKey() {
        Assertions.assertEquals(DUMMY_CACHE_KEY, filterHolder.getCacheKey(Mockito.mock(HttpServletRequest.class)));
    }

    @Test
    void getLoginIdentifier() {
        Assertions.assertEquals(DUMMY_LOGIN_IDENTIFIER,
                filterHolder.getLoginIdentifier(Mockito.mock(HttpServletRequest.class)));
    }

    @Test
    void isAcceptedFilter() {
        Assertions.assertTrue(filterHolder.isAcceptedFilter(Mockito.mock(HttpServletRequest.class)));
    }

    @Test
    void isAnonymousAllowed() {
        Assertions.assertFalse(filterHolder.isAnonymousAllowed());
        authenticationContext = AuthenticationContext.builder().authenticated(true).cacheKey(DUMMY_CACHE_KEY)
                .loginIdentifier(DUMMY_LOGIN_IDENTIFIER).build();

    }

    @Test
    void doFilter() throws IOException, ServletException {
        doNothing().when(filter).doFilter(any(), any(), any());
        filterHolder.doFilter(Mockito.mock(HttpServletRequest.class), Mockito.mock(HttpServletResponse.class),
                Mockito.mock(FilterChain.class));
        Mockito.verify(filter, times(1)).doFilter(any(), any(), any());
    }

    @Test
    void doesRequireReAuthentication() {
        Authentication authentication = Mockito.mock(Authentication.class);
        SimpleUser user = new SimpleUser(new UserImpl(authenticationContext.getLoginIdentifier()));
        when(authentication.getPrincipal()).thenReturn(user);
        Assertions.assertFalse(
                filterHolder.doesRequireReAuthentication(Mockito.mock(HttpServletRequest.class), authentication));
        user = new SimpleUser(new UserImpl("Dummy"));
        when(authentication.getPrincipal()).thenReturn(user);
        Assertions.assertTrue(
                filterHolder.doesRequireReAuthentication(Mockito.mock(HttpServletRequest.class), authentication));
    }

    @Test
    void getAuthenticationContext() {
        Assertions.assertNotNull(filterHolder.getAuthenticationContext());
    }

    @Test
    void getFilter() {
        Assertions.assertNotNull(filterHolder.getFilter());
    }
}