package org.artifactory.rest.common.exception.mapper;

import com.jfrog.bh.acl.error.base.BundleException;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.artifactory.rest.ErrorResponse;
import org.springframework.stereotype.Component;

import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import javax.ws.rs.ext.ExceptionMapper;
import javax.ws.rs.ext.Provider;
import java.util.Objects;

import static org.apache.http.HttpStatus.SC_INTERNAL_SERVER_ERROR;

/**
 * Acts as a controller advice for all Bundle-Handler REST resources.
 * <p>
 * All instances of {@link BundleException}, which are thrown up to
 * the REST controllers are intercepted at this point.
 * <p>
 * HTTP clients get only a status code and a public message from the exception.
 * <p>
 * The log message varies based on the exception subtype and the log level enabled.
 */
@Slf4j
@Provider
@Component
public class BundleExceptionMapper implements ExceptionMapper<BundleException> {

    @NonNull
    @Override
    public Response toResponse(@NonNull BundleException e) {
        logBundleException(e);
        return buildErrorResponse(e);
    }

    /**
     * The majority of the anticipated exceptions are from the 4xx range.
     * Do not log them until the debug/trace level is enabled for troubleshooting.
     * Instead, log the ones, which are from the 5xx range (server faults).
     */
    private void logBundleException(@NonNull BundleException e) {
        if (e.getStatus() >= SC_INTERNAL_SERVER_ERROR) {
            log.error(buildLogMessage(e), e);
        } else if (log.isTraceEnabled()) {
            log.trace(buildLogMessage(e), e);
        } else if (log.isDebugEnabled()) {
            log.debug(buildLogMessage(e));
        }
    }

    @NonNull
    private String buildLogMessage(@NonNull BundleException e) {
        StringBuilder message = new StringBuilder("Bundle-handler:");
        message.append(" API={").append(e.getPublicMessage()).append("}");
        if (!Objects.equals(e.getPublicMessage(), e.getInternalMessage())) {
            message.append(" EXT={").append(e.getInternalMessage()).append("}");
        }
        return message.toString();
    }

    @NonNull
    private Response buildErrorResponse(@NonNull BundleException e) {
        return Response.status(e.getStatus())
                .type(MediaType.APPLICATION_JSON)
                .entity(new ErrorResponse(e.getStatus(), e.getPublicMessage()))
                .build();
    }
}
