/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.ui.rest.service.admin.configuration.general;

import org.apache.commons.lang3.StringUtils;
import org.apache.http.HttpStatus;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.release.bundle.ReleaseBundleAddon;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.rest.release.ReleaseBundlesConfigModel;
import org.artifactory.descriptor.Descriptor;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.descriptor.download.DownloadRedirectConfigDescriptor;
import org.artifactory.descriptor.download.FolderDownloadConfigDescriptor;
import org.artifactory.descriptor.message.SystemMessageDescriptor;
import org.artifactory.descriptor.replication.GlobalReplicationsConfigDescriptor;
import org.artifactory.descriptor.signature.SignedUrlConfig;
import org.artifactory.descriptor.subscription.SubscriptionConfig;
import org.artifactory.descriptor.sumologic.SumoLogicConfigDescriptor;
import org.artifactory.descriptor.trashcan.TrashcanConfigDescriptor;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.rest.common.service.RestService;
import org.artifactory.service.ConfigValidationService;
import org.artifactory.ui.rest.model.admin.configuration.generalconfig.GeneralConfig;
import org.artifactory.util.CollectionUtils;
import org.jfrog.common.validator.NameValidator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.regex.Pattern;

/**
 * @author Chen Keinan
 */
@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
public class UpdateGeneralConfigService implements RestService {
    private static final Logger log = LoggerFactory.getLogger(UpdateGeneralConfigService.class);

    private final CentralConfigService centralConfigService;
    private final AddonsManager addonsManager;
    private final ConfigValidationService configValidationService;
    private final Pattern logoUrlPattern =
            Pattern.compile("^(http)?s?:?(//[^\"']*\\.(?:jpg|jpeg|gif|png|apng|avif|webp|svg))$");

    @Autowired
    public UpdateGeneralConfigService(CentralConfigService centralConfigService, AddonsManager addonsManager,
            ConfigValidationService configValidationService) {
        this.centralConfigService = centralConfigService;
        this.addonsManager = addonsManager;
        this.configValidationService = configValidationService;
    }

    @Override
    public void execute(ArtifactoryRestRequest request, RestResponse response) {
        try {
            GeneralConfig generalConfig = (GeneralConfig) request.getImodel();
            updateDescriptorAndSave(generalConfig);
            response.info("Successfully updated settings");
        } catch (IllegalArgumentException e) {
            String message = e.getMessage();
            log.error(message);
            log.debug(message, e);

            response.error(message);
            response.responseCode(HttpStatus.SC_BAD_REQUEST);
        }
    }

    /**
     * update config descriptor with general config setting and save
     *
     * @param generalConfig - general setting sent from client
     */
    void updateDescriptorAndSave(GeneralConfig generalConfig) {
        assertGeneralConfigValidValues(generalConfig);
        MutableCentralConfigDescriptor mutableDescriptor = centralConfigService.getMutableDescriptor();
        mutableDescriptor.setServerName(generalConfig.getServerName());
        mutableDescriptor.setDateFormat(generalConfig.getDateFormat());
        mutableDescriptor.setUrlBase(generalConfig.getCustomUrlBase());
        mutableDescriptor.setFederatedRepoUrlBase(generalConfig.getFederatedRepoUrlBase());
        mutableDescriptor.setFileUploadMaxSizeMb(generalConfig.getFileUploadMaxSize());
        mutableDescriptor.setOfflineMode(generalConfig.isGlobalOfflineMode());
        mutableDescriptor.setArchiveIndexEnabled(generalConfig.isArchiveIndexEnabled());
        mutableDescriptor.getAddons().setShowAddonsInfo(generalConfig.isShowAddonSettings());
        mutableDescriptor.setLogo(generalConfig.getLogoUrl());
        mutableDescriptor.setHelpLinksEnabled(generalConfig.isHelpLinksEnabled());
        List<Descriptor> descriptorList = new ArrayList<>();

        //System message config
        descriptorList.add(updateSystemMessageConfig(generalConfig, mutableDescriptor));
        //Folder download config
        descriptorList.add(updateFolderDownloadConfig(generalConfig, mutableDescriptor));
        //Trashcan config
        descriptorList.add(updateTrashcanConfig(generalConfig, mutableDescriptor));
        //Update global replication config
        descriptorList.add(updateGlobalReplicationConfig(generalConfig, mutableDescriptor));
        // Update Sumo Logic config
        descriptorList.add(updateSumoLogicConfig(generalConfig, mutableDescriptor));
        // Update Release Bundles config
        ReleaseBundlesConfigModel model = updateReleaseBundlesConfig(generalConfig, mutableDescriptor);
        //Update Signed Url config
        descriptorList.add(updateSignedUrlConfig(generalConfig, mutableDescriptor));
        // Update Cloud storage download redirect
        descriptorList.add(updateDownloadRedirectConfig(generalConfig, mutableDescriptor));
        // Update Subscription config
        updateSubscriptionConfig(generalConfig, mutableDescriptor);
        centralConfigService.saveEditedDescriptorAndReload(mutableDescriptor, generalConfig, descriptorList, model);
    }

    private void assertGeneralConfigValidValues(GeneralConfig generalConfig) {
        assertGeneralConfigServerName(generalConfig.getServerName());
        assertGeneralConfigLogoUrl(generalConfig.getLogoUrl());
    }

    private void assertGeneralConfigLogoUrl(String logoUrl) {
        if (!StringUtils.isBlank(logoUrl) &&
                !logoUrlPattern.matcher(logoUrl).matches()) {
            throw new IllegalArgumentException("Logo URL is not valid");
        }
    }

    private void assertGeneralConfigServerName(String serverName) {
        if (!StringUtils.isBlank(serverName)) {
            Optional<String> serverNamePossibleError = NameValidator.validate(serverName);
            if (serverNamePossibleError.isPresent()) {
                throw new IllegalArgumentException(serverNamePossibleError.get());
            }
        }
    }

    private void updateSubscriptionConfig(GeneralConfig generalConfig,
            MutableCentralConfigDescriptor mutableDescriptor) {
        if (generalConfig.getSubscription() != null
                && CollectionUtils.notNullOrEmpty(generalConfig.getSubscription().getEmails())) {

            SubscriptionConfig subscriptionConfig = new SubscriptionConfig();
            subscriptionConfig.setEmails(generalConfig.getSubscription().getEmails());
            // We validate here to avoid complicating the code by returning optionals
            this.configValidationService.validate(subscriptionConfig);
            mutableDescriptor.setSubscriptionConfig(subscriptionConfig);
        }
    }

    //Does not override defaults if UI sent empty model.
    private SystemMessageDescriptor updateSystemMessageConfig(GeneralConfig generalConfig,
            MutableCentralConfigDescriptor descriptor) {
        SystemMessageDescriptor systemMessageDescriptor =
                Optional.ofNullable(descriptor.getSystemMessageConfig()).orElse(new SystemMessageDescriptor());
        systemMessageDescriptor.setEnabled(Optional.ofNullable(
                generalConfig.isSystemMessageEnabled()).orElse(systemMessageDescriptor.isEnabled()));
        systemMessageDescriptor.setTitle(Optional.ofNullable(
                generalConfig.getSystemMessageTitle()).orElse(systemMessageDescriptor.getTitle()));
        systemMessageDescriptor.setTitleColor(Optional.ofNullable(
                generalConfig.getSystemMessageTitleColor()).orElse(systemMessageDescriptor.getTitleColor()));
        systemMessageDescriptor.setMessage(Optional.ofNullable(
                generalConfig.getSystemMessage()).orElse(systemMessageDescriptor.getMessage()));
        systemMessageDescriptor.setShowOnAllPages(Optional.ofNullable(
                generalConfig.isShowSystemMessageOnAllPages()).orElse(systemMessageDescriptor.isShowOnAllPages()));
        descriptor.setSystemMessageConfig(systemMessageDescriptor);
        return systemMessageDescriptor;
    }

    //Does not override defaults if UI sent empty model.
    private FolderDownloadConfigDescriptor updateFolderDownloadConfig(GeneralConfig generalConfig,
            MutableCentralConfigDescriptor descriptor) {
        FolderDownloadConfigDescriptor folderDownloadConfig = descriptor.getFolderDownloadConfig();
        folderDownloadConfig.setEnabled(
                Optional.ofNullable(generalConfig.isFolderDownloadEnabled()).orElse(folderDownloadConfig.isEnabled()));
        folderDownloadConfig.setEnabledForAnonymous(
                Optional.ofNullable(generalConfig.isFolderDownloadEnabledForAnonymous())
                        .orElse(folderDownloadConfig.isEnabledForAnonymous()));
        folderDownloadConfig.setMaxConcurrentRequests(Optional.ofNullable(
                        generalConfig.getFolderDownloadMaxConcurrentRequests())
                .orElse(folderDownloadConfig.getMaxConcurrentRequests()));
        folderDownloadConfig.setMaxDownloadSizeMb(Optional.ofNullable(
                generalConfig.getFolderDownloadMaxSizeMb()).orElse(folderDownloadConfig.getMaxDownloadSizeMb()));
        folderDownloadConfig.setMaxFiles(Optional.ofNullable(
                generalConfig.getMaxFolderDownloadFilesLimit()).orElse(folderDownloadConfig.getMaxFiles()));
        folderDownloadConfig.setEnabledEmptyDirectories(
                Optional.ofNullable(generalConfig.isFolderDownloadEnabledEmptyDirectories())
                        .orElse(folderDownloadConfig.isEnabledEmptyDirectories()));
        descriptor.setFolderDownloadConfig(folderDownloadConfig);
        return folderDownloadConfig;
    }

    //Does not override defaults if UI sent empty model.
    private TrashcanConfigDescriptor updateTrashcanConfig(GeneralConfig generalConfig,
            MutableCentralConfigDescriptor descriptor) {
        TrashcanConfigDescriptor trashcanConfig = descriptor.getTrashcanConfig();
        trashcanConfig.setEnabled(Optional.ofNullable(
                generalConfig.getTrashcanEnabled()).orElse(trashcanConfig.isEnabled()));
        trashcanConfig.setAllowPermDeletes(Optional.ofNullable(
                generalConfig.getAllowPermDeletes()).orElse(trashcanConfig.isAllowPermDeletes()));
        trashcanConfig.setRetentionPeriodDays(Optional.ofNullable(
                generalConfig.getTrashcanRetentionPeriodDays()).orElse(trashcanConfig.getRetentionPeriodDays()));
        descriptor.setTrashcanConfig(trashcanConfig);
        return trashcanConfig;
    }

    //Does not override defaults if UI sent empty model.
    private GlobalReplicationsConfigDescriptor updateGlobalReplicationConfig(GeneralConfig generalConfig,
            MutableCentralConfigDescriptor descriptor) {
        GlobalReplicationsConfigDescriptor replicationsConfig = descriptor.getReplicationsConfig();
        replicationsConfig.setBlockPullReplications(Optional.ofNullable(
                generalConfig.getBlockPullReplications()).orElse(replicationsConfig.isBlockPullReplications()));
        replicationsConfig.setBlockPushReplications(Optional.ofNullable(
                generalConfig.getBlockPushReplications()).orElse(replicationsConfig.isBlockPushReplications()));
        return replicationsConfig;

    }

    //Does not override defaults if UI sent empty model.
    private SumoLogicConfigDescriptor updateSumoLogicConfig(GeneralConfig generalConfig,
            MutableCentralConfigDescriptor descriptor) {
        SumoLogicConfigDescriptor sumoLogicConfig = descriptor.getSumoLogicConfig();
        sumoLogicConfig.setEnabled(
                Optional.ofNullable(generalConfig.getSumoLogicEnabled()).orElse(sumoLogicConfig.isEnabled()));
        sumoLogicConfig.setCollectorUrl(Optional.ofNullable(
                generalConfig.getSumoLogicCollectorUrl()).orElse(sumoLogicConfig.getCollectorUrl()));
        return sumoLogicConfig;
    }

    //Does not override defaults if UI sent empty model.
    private ReleaseBundlesConfigModel updateReleaseBundlesConfig(GeneralConfig generalConfig,
            MutableCentralConfigDescriptor descriptor) {
        ReleaseBundlesConfigModel model = new ReleaseBundlesConfigModel(
                Optional.ofNullable(generalConfig.getReleaseBundlesCleanup())
                        .orElse(descriptor.getReleaseBundlesConfig().getIncompleteCleanupPeriodHours()));
        addonsManager.addonByType(ReleaseBundleAddon.class).setReleaseBundlesConfig(model);
        return model;
    }

    //Does not override defaults if UI sent empty model.
    private SignedUrlConfig updateSignedUrlConfig(GeneralConfig generalConfig,
            MutableCentralConfigDescriptor descriptor) {
        SignedUrlConfig signedUrlConfig = descriptor.getSignedUrlConfig();
        signedUrlConfig.setMaxValidForSeconds(Optional.ofNullable(generalConfig.getSignedUrlMaxValidForSecs())
                .orElse(signedUrlConfig.getMaxValidForSeconds()));
        return signedUrlConfig;
    }

    //Does not override defaults if UI sent empty model.
    private DownloadRedirectConfigDescriptor updateDownloadRedirectConfig(GeneralConfig generalConfig,
            MutableCentralConfigDescriptor descriptor) {
        DownloadRedirectConfigDescriptor downloadRedirectConfig = descriptor.getDownloadRedirectConfig();
        downloadRedirectConfig.setFileMinimumSize(Optional.ofNullable(
                        generalConfig.getDownloadRedirectFileMinimumSize())
                .orElse(downloadRedirectConfig.getFileMinimumSize()));
        return downloadRedirectConfig;
    }

}
