package org.artifactory.ui.utils;

import com.google.common.collect.Maps;
import org.artifactory.aol.AolService;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.descriptor.repo.FederatedRepoDescriptor;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.rest.common.exception.RepoConfigException;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.artifactory.ui.rest.model.admin.configuration.repository.RepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.local.LocalRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.remote.RemoteRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.typespecific.GenericTypeSpecificConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.typespecific.MavenTypeSpecificConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.typespecific.TypeSpecificConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.virtual.VirtualRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.virtual.VirtualSelectedRepository;
import org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.RepositoryType;
import org.artifactory.ui.rest.model.onboarding.QuickRepositoryModel;
import org.artifactory.ui.rest.service.admin.configuration.repositories.CreateRepositoryConfigService;
import org.artifactory.ui.rest.service.onboarding.CreateDefaultReposResponseModel;
import org.mockito.ArgumentCaptor;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.util.List;
import java.util.Map;
import java.util.Set;

import static org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.RepositoryType.LOCAL;
import static org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.RepositoryType.REMOTE;
import static org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.RepositoryType.VIRTUAL;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertNull;
import static org.testng.Assert.assertTrue;

/**
 * @author Alexei Vainshtein
 */
public class DefaultRepoCreatorTest extends ArtifactoryHomeBoundTest {

    public static String REPO_NAME = "repo";

    @BeforeMethod
    public void setUp() {
        ArtifactoryContext context = mock(ArtifactoryContext.class);
        ArtifactoryContextThreadBinder.bind(context);
        MutableCentralConfigDescriptor descriptor = mock(MutableCentralConfigDescriptor.class);
        Map<String, LocalRepoDescriptor> reposMap = Maps.newLinkedHashMap();
        reposMap.put(REPO_NAME, new FederatedRepoDescriptor());
        when(descriptor.getLocalAndFederatedRepositoriesMap()).thenReturn(reposMap);
        CentralConfigService service = mock(CentralConfigService.class);
        when(service.getMutableDescriptor()).thenReturn(descriptor);
        when(context.getCentralConfig()).thenReturn(service);
        AolService aolService = mock(AolService.class);
        when(aolService.isAol()).thenReturn(false);
        when(context.beanForType(AolService.class)).thenReturn(aolService);
    }

    @AfterMethod
    public void tearDown() {
        ArtifactoryContextThreadBinder.unbind();
    }

    @Test
    public void testIsDefaultDeploymentRepoValid() {
        DefaultRepoCreator repoCreator = new DefaultRepoCreator();
        QuickRepositoryModel model = new QuickRepositoryModel();
        model.setDefaultDeploymentRepo(REPO_NAME);
        assertTrue(repoCreator.isDefaultDeploymentRepoValid(model));
    }

    @Test
    public void testLocalQuickModel() throws RepoConfigException {
        DefaultRepoCreator repoCreator = new DefaultRepoCreator();
        Set<String> envs = Set.of("prod");
        QuickRepositoryModel model = createModel(LOCAL, RepoType.Generic, "jfrog1", envs);
        CreateDefaultReposResponseModel responseModel = new CreateDefaultReposResponseModel();
        CreateRepositoryConfigService createRepoService = mock(CreateRepositoryConfigService.class);
        RestResponse restResponse = mock(RestResponse.class);
        var repoConfigCaptor = ArgumentCaptor.forClass(RepositoryConfigModel.class);
        when(createRepoService.createRepo(eq(restResponse), repoConfigCaptor.capture(), any())).thenReturn(
                mock(MutableCentralConfigDescriptor.class));

        repoCreator.createQuickRepos(createRepoService, restResponse, model, responseModel, false);

        assertTrue(repoConfigCaptor.getValue() instanceof LocalRepositoryConfigModel);
        var localModel = (LocalRepositoryConfigModel) repoConfigCaptor.getValue();
        var general = localModel.getGeneral();
        assertEquals(general.getRepoKey(), "repo2");
        assertEquals(envs, general.getEnvironments());
        assertEquals(general.getProjectKey(), "jfrog1");
        var basic = localModel.getBasic();
        assertEquals(basic.getLayout(), "simple-default");
        assertNull(basic.getXrayConfig());
        assertNull(basic.getKeypair());
        assertNull(basic.getPrimaryKeyPair());
        assertNull(basic.getSecondaryKeyPair());
        TypeSpecificConfigModel typeSpecific = localModel.getTypeSpecific();
        assertTrue(typeSpecific instanceof GenericTypeSpecificConfigModel);
        assertTrue(((GenericTypeSpecificConfigModel) typeSpecific).isListRemoteFolderItems());
    }
    @Test
    public void testRemoteQuickModel() throws RepoConfigException {
        DefaultRepoCreator repoCreator = new DefaultRepoCreator();
        Set<String> envs = Set.of("prod");
        QuickRepositoryModel model = createModel(REMOTE, RepoType.Maven, "jfrog1", envs);
        CreateDefaultReposResponseModel responseModel = new CreateDefaultReposResponseModel();
        CreateRepositoryConfigService createRepoService = mock(CreateRepositoryConfigService.class);
        RestResponse restResponse = mock(RestResponse.class);
        var repoConfigCaptor = ArgumentCaptor.forClass(RepositoryConfigModel.class);
        when(createRepoService.createRepo(eq(restResponse), repoConfigCaptor.capture(), any())).thenReturn(
                mock(MutableCentralConfigDescriptor.class));

        repoCreator.createQuickRepos(createRepoService, restResponse, model, responseModel, true);

        assertTrue(repoConfigCaptor.getValue() instanceof RemoteRepositoryConfigModel);
        var remoteModel = (RemoteRepositoryConfigModel) repoConfigCaptor.getValue();
        var general = remoteModel.getGeneral();
        assertEquals(general.getRepoKey(), "repo2");
        assertEquals(envs, general.getEnvironments());
        assertEquals(general.getProjectKey(), "jfrog1");
        var basic = remoteModel.getBasic();
        assertEquals(basic.getLayout(), "maven-2-default");
        assertEquals(basic.getUrl(), "https://repo1.maven.org/maven2/");
        assertNull(basic.getXrayConfig());
        assertNull(basic.getKeypair());
        assertNull(basic.getPrimaryKeyPair());
        assertNull(basic.getSecondaryKeyPair());
        TypeSpecificConfigModel typeSpecific = remoteModel.getTypeSpecific();
        assertTrue(typeSpecific instanceof MavenTypeSpecificConfigModel);
        assertTrue(((MavenTypeSpecificConfigModel) typeSpecific).isListRemoteFolderItems());
    }

    @Test
    public void testVirtualQuickModel() throws RepoConfigException {
        DefaultRepoCreator repoCreator = new DefaultRepoCreator();
        Set<String> envs = Set.of("prod");
        QuickRepositoryModel model = createModel(VIRTUAL, RepoType.Maven, "jfrog1", envs);
        model.setIncludedLocalRepositories(List.of(REPO_NAME));
        CreateDefaultReposResponseModel responseModel = new CreateDefaultReposResponseModel();
        CreateRepositoryConfigService createRepoService = mock(CreateRepositoryConfigService.class);
        RestResponse restResponse = mock(RestResponse.class);
        var repoConfigCaptor = ArgumentCaptor.forClass(RepositoryConfigModel.class);
        when(createRepoService.createRepo(eq(restResponse), repoConfigCaptor.capture(), any())).thenReturn(
                mock(MutableCentralConfigDescriptor.class));

        repoCreator.createQuickRepos(createRepoService, restResponse, model, responseModel, true);

        assertTrue(repoConfigCaptor.getValue() instanceof VirtualRepositoryConfigModel);
        var virtualModel = (VirtualRepositoryConfigModel) repoConfigCaptor.getValue();
        var general = virtualModel.getGeneral();
        assertEquals(general.getRepoKey(), "repo2");
        assertEquals(envs, general.getEnvironments());
        assertEquals(general.getProjectKey(), "jfrog1");
        var basic = virtualModel.getBasic();
        assertEquals(basic.getLayout(), "maven-2-default");
        assertEquals(basic.getSelectedRepositories().size(), 1);
        VirtualSelectedRepository selectedRepo = basic.getSelectedRepositories().get(0);
        assertEquals(selectedRepo.getRepoName(), REPO_NAME);
        assertEquals(selectedRepo.getType(), "local");
        assertNull(basic.getXrayConfig());
        assertNull(basic.getKeypair());
        assertNull(basic.getPrimaryKeyPair());
        assertNull(basic.getSecondaryKeyPair());
        TypeSpecificConfigModel typeSpecific = virtualModel.getTypeSpecific();
        assertTrue(typeSpecific instanceof MavenTypeSpecificConfigModel);
        assertTrue(((MavenTypeSpecificConfigModel) typeSpecific).isListRemoteFolderItems());
    }

    private static QuickRepositoryModel createModel(RepositoryType type, RepoType packageType, String project,
            Set<String> envs) {
        QuickRepositoryModel model = new QuickRepositoryModel();
        model.setRepoName("repo2");
        model.setRepoType(type);
        model.setPackageType(packageType);
        model.setProject(project);
        model.setEnvs(envs);
        return model;
    }

}